/*********************************************************************
*                   (c) SEGGER Microcontroller GmbH                  *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 2003 - 2019     SEGGER Microcontroller GmbH              *
*                                                                    *
*       www.segger.com     Support: www.segger.com/ticket            *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host * USB Host stack for embedded applications        *
*                                                                    *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product.                          *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       emUSB-Host version: V2.16                                    *
*                                                                    *
**********************************************************************
-------------------------- END-OF-HEADER -----------------------------

File    : USBH_MIDI_PlayMIDI.c
Purpose : This sample is designed to present emUSBH's capability to
          integrate with emLib-MIDI and play a track from a MIDI
          file along with the lyrics for that track when visualization
          via emWin is enabled.

Additional information:
  In case emWin for visualization shall be used, please set in the
  preprocessor project setting USE_GUI=1 .
  This is configured for the LCD on an STM32F746G-Discovery board.
  For other LCD sizes, you can configure dimensions accordingly.

*/

/*********************************************************************
*
*       #include section
*
**********************************************************************
*/


#include "RTOS.h"
#include "BSP.h"

#include "USBH_MIDI.h"
#include "USBH_Util.h"
#include "MIDI.h"
#include "SEGGER.h"


#ifndef USE_GUI
  #define USE_GUI              0
#endif
#if USE_GUI
 #include "GUI.h"
#endif

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/

/*********************************************************************
*
*       Event decomposition
*/
#define MIDI_CABLE(E)    ((E) >> 28)
#define MIDI_STATUS(E)   (((E) >> 16) & 0xFF)
#define MIDI_CHANNEL(E)  ((((E) >> 16) & 0xF) + 1)
#define MIDI_DATA2(E)    (unsigned)(((E) >> 8) & 0x7Fu)
#define MIDI_DATA3(E)    (unsigned)((E) & 0x7Fu)
#define MIDI_DATAX(E)    ((MIDI_DATA3(E) << 7) + MIDI_DATA2(E))

/*********************************************************************
*
*       MIDI note numbers
*/
#define MIDI_NOTE_B7          107
#define MIDI_NOTE_A7_SHARP    106
#define MIDI_NOTE_A7          105
#define MIDI_NOTE_G7_SHARP    104
#define MIDI_NOTE_G7          103
#define MIDI_NOTE_F7_SHARP    102
#define MIDI_NOTE_F7          101
#define MIDI_NOTE_E7          100
#define MIDI_NOTE_D7_SHARP     99
#define MIDI_NOTE_D7           98
#define MIDI_NOTE_C7_SHARP     97
#define MIDI_NOTE_C7           96
#define MIDI_NOTE_B6           95
#define MIDI_NOTE_A6_SHARP     94
#define MIDI_NOTE_A6           93
#define MIDI_NOTE_G6_SHARP     92
#define MIDI_NOTE_G6           91
#define MIDI_NOTE_F6_SHARP     90
#define MIDI_NOTE_F6           89
#define MIDI_NOTE_E6           88
#define MIDI_NOTE_D6_SHARP     87
#define MIDI_NOTE_D6           86
#define MIDI_NOTE_C6_SHARP     85
#define MIDI_NOTE_C6           84
#define MIDI_NOTE_B5           83
#define MIDI_NOTE_A5_SHARP     82
#define MIDI_NOTE_A5           81
#define MIDI_NOTE_G5_SHARP     80
#define MIDI_NOTE_G5           79
#define MIDI_NOTE_F5_SHARP     78
#define MIDI_NOTE_F5           77
#define MIDI_NOTE_E5           76
#define MIDI_NOTE_D5_SHARP     75
#define MIDI_NOTE_D5           74
#define MIDI_NOTE_C5_SHARP     73
#define MIDI_NOTE_C5           72
#define MIDI_NOTE_B4           71
#define MIDI_NOTE_A4_SHARP     70
#define MIDI_NOTE_A4           69
#define MIDI_NOTE_G4_SHARP     68
#define MIDI_NOTE_G4           67
#define MIDI_NOTE_F4_SHARP     66
#define MIDI_NOTE_F4           65
#define MIDI_NOTE_E4           64
#define MIDI_NOTE_D4_SHARP     63
#define MIDI_NOTE_D4           62
#define MIDI_NOTE_C4_SHARP     61
#define MIDI_NOTE_C4           60
#define MIDI_NOTE_B3           59
#define MIDI_NOTE_A3_SHARP     58
#define MIDI_NOTE_A3           57
#define MIDI_NOTE_G3_SHARP     56
#define MIDI_NOTE_G3           55
#define MIDI_NOTE_F3_SHARP     54
#define MIDI_NOTE_F3           53
#define MIDI_NOTE_E3           52
#define MIDI_NOTE_D3_SHARP     51
#define MIDI_NOTE_D3           50
#define MIDI_NOTE_C3_SHARP     49
#define MIDI_NOTE_C3           48
#define MIDI_NOTE_B2           47
#define MIDI_NOTE_A2_SHARP     46
#define MIDI_NOTE_A2           45
#define MIDI_NOTE_G2_SHARP     44
#define MIDI_NOTE_G2           43
#define MIDI_NOTE_F2_SHARP     42
#define MIDI_NOTE_F2           41
#define MIDI_NOTE_E2           40
#define MIDI_NOTE_D2_SHARP     39
#define MIDI_NOTE_D2           38
#define MIDI_NOTE_C2_SHARP     37
#define MIDI_NOTE_C2           36
#define MIDI_NOTE_B1           35
#define MIDI_NOTE_A1_SHARP     34
#define MIDI_NOTE_A1           33
#define MIDI_NOTE_G1_SHARP     32
#define MIDI_NOTE_G1           31
#define MIDI_NOTE_F1_SHARP     30
#define MIDI_NOTE_F1           29
#define MIDI_NOTE_E1           28
#define MIDI_NOTE_D1_SHARP     27
#define MIDI_NOTE_D1           26
#define MIDI_NOTE_C1_SHARP     25
#define MIDI_NOTE_C1           24
#define MIDI_NOTE_B0           23
#define MIDI_NOTE_A0_SHARP     22
#define MIDI_NOTE_A0           21

/*********************************************************************
*
*       Defines, configurable
*
**********************************************************************
*/

/*********************************************************************
*
*       Varies per MIDI file
*/
#define LYRIC_TRACK  0     // Track containing lyrics
#define MELODY_TRACK 1     // Track containing playable melody

/*********************************************************************
*
*       Virtual keyboard
*/
#define VKB_MIDI_CHANNEL        1     // MIDI output channel, 1...16
#define VKB_NOTE_ON_VELOCITY   64     // Velocity for Note On, 1...127
#define VKB_NOTE_OFF_VELOCITY  64     // Velocity for Note Off, 0...127

/*********************************************************************
*
*       Palette
*/
#define COLOR_BACKGROUND     GUI_MAKE_COLOR(0x801010)
#define COLOR_KEY_BORDER     GUI_MAKE_COLOR(0x473E34)
#define COLOR_KEY_DOWN       GUI_MAKE_COLOR(0x30C030)
#define COLOR_KEY_ERROR      GUI_MAKE_COLOR(0x3030D0)

/*********************************************************************
*
*       Key dimensions
*/
#define WHITE_KEY_WIDTH      19
#define WHITE_KEY_LENGTH     100
#define BLACK_KEY_LENGTH     ((int)(WHITE_KEY_LENGTH * 0.6))
#define BLACK_KEY_WIDTH      ((int)(WHITE_KEY_WIDTH * (13.7/23.5)))

/*********************************************************************
*
*       Keybed dimensions
*/
#define KEYBED_LOW_NOTE      MIDI_NOTE_C2
#define KEYBED_HIGH_NOTE     MIDI_NOTE_F5

/*********************************************************************
*
*       Keybed position
*/
#define MARGIN_LEFT          12
#define MARGIN_TOP           80

/*********************************************************************
*
*       Local types
*
**********************************************************************
*/

enum {
  TASK_PRIO_KB_DISP = 150,
  TASK_PRIO_KB_RECV,
  TASK_PRIO_APP,
  TASK_PRIO_KB_PLAY,
  TASK_PRIO_USBH_MAIN,
  TASK_PRIO_USBH_ISR
};

typedef struct {
  unsigned PlayHead;
  unsigned CurTrackId;
  unsigned PlayTrackId;
  U32      Tempo;
  unsigned Division;
  unsigned LyricPos;
  char     aLyric[128];
} PLAY_CONTEXT;

#if USE_GUI
typedef struct {
  U32 KeyColor;
  U32 KeyBorder;
  U32 LegendColor;
} KEY_PALETTE;

typedef struct {
  enum { WHITE, BLACK } Color;
  unsigned              Index;
} KEY_DESC;
#endif

/*********************************************************************
*
*       Prototypes
*
**********************************************************************
*/

static void _PlayAPI_Header     (void *pContext, unsigned Format, unsigned TrackCnt, unsigned Division);
static void _PlayAPI_BeginTrack (void *pContext, unsigned Id);
static void _PlayAPI_Delta      (void *pContext, U32 Delta);
static void _PlayAPI_Lyric      (void *pContext, const U8 *pText, unsigned TextLen);
static void _PlayAPI_SetTempo   (void *pContext, U32 Tempo);
static void _PlayAPI_MidiEvent  (void *pContext, const U8 *pData, unsigned DataLen);

/*********************************************************************
*
*       Static const data
*
**********************************************************************
*/
#if USE_GUI
static const KEY_PALETTE WhiteKeyOff    = { GUI_WHITE,       COLOR_KEY_BORDER, GUI_GRAY_60 };
static const KEY_PALETTE WhiteKeyOn     = { COLOR_KEY_DOWN,  COLOR_KEY_BORDER, GUI_WHITE   };
static const KEY_PALETTE BlackKeyOff    = { GUI_BLACK,       COLOR_KEY_BORDER, GUI_BLACK   };
static const KEY_PALETTE BlackKeyOn     = { COLOR_KEY_DOWN,  COLOR_KEY_BORDER, GUI_WHITE   };
static const KEY_DESC _aKeyMapping[] = {
  { WHITE, 0 },  // A
  { BLACK, 0 },  // A#
  { WHITE, 1 },  // B
  { WHITE, 2 },  // C
  { BLACK, 2 },  // C#
  { WHITE, 3 },  // D
  { BLACK, 3 },  // D#
  { WHITE, 4 },  // E
  { WHITE, 5 },  // F
  { BLACK, 5 },  // F#
  { WHITE, 6 },  // G
  { BLACK, 6 },  // G#
};
#endif

static const MIDI_FILE_API _PlayAPI = {
  /*pfError           */  NULL,
  /*pfWarning         */  NULL,
  /*pfHeader          */  _PlayAPI_Header,
  /*pfBeginTrack      */  _PlayAPI_BeginTrack,
  /*pfDelta           */  _PlayAPI_Delta,
  /*pfKeySignature    */  NULL,
  /*pfTimeSignature   */  NULL,
  /*pfText            */  NULL,
  /*pfCopyrightNotice */  NULL,
  /*pfTrackName       */  NULL,
  /*pfInstrumentName  */  NULL,
  /*pfLyric           */  NULL,
  /*pfMarker          */  NULL,
  /*pfCuePoint        */  NULL,
  /*pfProgramName     */  NULL,
  /*pfDeviceName      */  NULL,
  /*pfSetTempo        */  _PlayAPI_SetTempo,
  /*pfSequenceNum     */  NULL,
  /*pfChannelPrefix   */  NULL,
  /*pfPortPrefix      */  NULL,
  /*pfMidiEvent       */  _PlayAPI_MidiEvent,
  /*pfBeginSysex      */  NULL,
  /*pfSysexData       */  NULL,
  /*pfEndSysex        */  NULL,
  /*pfEndTrack        */  NULL
};

static const MIDI_FILE_API _LyricAPI = {
  /*pfError           */  NULL,
  /*pfWarning         */  NULL,
  /*pfHeader          */  _PlayAPI_Header,
  /*pfBeginTrack      */  _PlayAPI_BeginTrack,
  /*pfDelta           */  _PlayAPI_Delta,
  /*pfKeySignature    */  NULL,
  /*pfTimeSignature   */  NULL,
  /*pfText            */  NULL,
  /*pfCopyrightNotice */  NULL,
  /*pfTrackName       */  NULL,
  /*pfInstrumentName  */  NULL,
  /*pfLyric           */  _PlayAPI_Lyric,
  /*pfMarker          */  NULL,
  /*pfCuePoint        */  NULL,
  /*pfProgramName     */  NULL,
  /*pfDeviceName      */  NULL,
  /*pfSetTempo        */  _PlayAPI_SetTempo,
  /*pfSequenceNum     */  NULL,
  /*pfChannelPrefix   */  NULL,
  /*pfPortPrefix      */  NULL,
  /*pfMidiEvent       */  NULL,
  /*pfBeginSysex      */  NULL,
  /*pfSysexData       */  NULL,
  /*pfEndSysex        */  NULL,
  /*pfEndTrack        */  NULL
};

const unsigned char _aMidiFile[21781] = {
 0x4D,  0x54,  0x68,  0x64,  0x00,  0x00,  0x00,  0x06,  0x00,  0x01,  0x00,  0x02,  0x01,  0xE0,  0x4D,  0x54,  0x72,  0x6B,  0x00,  0x00,  0x10,  0xFC,  0x00,  0xFF,  0x58,  0x04,  0x04,  0x02,  0x18,  0x08,  0x00,  0xFF,  0x51,  0x03,  0x0F,  0x42,  0x40,  0x00,  0xFF,  0x59,  0x02,  0x02,  0x00,  0xCB,  0x00,  0xFF,  0x05,  0x02,  0x2E,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x2E,  0x2E,  0x20,  0x00,  0xFF,
 0x05,  0x01,  0x0D,  0x83,  0x60,  0xFF,  0x05,  0x04,  0x2E,  0x2E,  0x2E,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x83,  0x60,  0xFF,  0x05,  0x05,  0x2E,  0x2E,  0x2E,  0x2E,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x83,  0x60,  0xFF,  0x05,  0x04,  0x53,  0x77,  0x69,  0x6E,  0x78,  0xFF,  0x05,  0x05,  0x67,  0x69,  0x6E,  0x67,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x69,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x68,  0x65,  0x20,
 0x78,  0xFF,  0x05,  0x04,  0x62,  0x61,  0x63,  0x6B,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x79,  0x61,  0x72,  0x64,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x70,  0x75,  0x6C,  0x6C,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x75,  0x70,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x69,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x79,  0x6F,  0x75,  0x72,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x66,  0x61,  0x73,  0x74,  0x20,
 0x81,  0x70,  0xFF,  0x05,  0x04,  0x63,  0x61,  0x72,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x77,  0x68,  0x69,  0x73,  0x78,  0xFF,  0x05,  0x02,  0x74,  0x6C,  0x78,  0xFF,  0x05,  0x04,  0x69,  0x6E,  0x67,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x6D,  0x79,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x6E,  0x61,  0x6D,  0x65,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x00,  0xFF,  0x05,  0x01,  0x0A,  0x8B,  0x20,
 0xFF,  0x05,  0x01,  0x4F,  0x78,  0xFF,  0x05,  0x04,  0x70,  0x65,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x75,  0x70,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x61,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x62,  0x65,  0x65,  0x72,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x61,  0x6E,  0x64,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x73,  0x61,  0x79,  0x20,  0x78,
 0xFF,  0x05,  0x04,  0x67,  0x65,  0x74,  0x20,  0x78,  0xFF,  0x05,  0x01,  0x6F,  0x78,  0xFF,  0x05,  0x04,  0x76,  0x65,  0x72,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x68,  0x65,  0x72,  0x65,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x49,  0x27,  0x6D,  0x20,  0x78,  0xFF,  0x05,  0x0A,  0x70,  0x6C,  0x61,  0x79,  0x69,  0x6E,  0x67,  0x5F,  0x61,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x76,  0x69,  0x78,  0xFF,
 0x05,  0x02,  0x64,  0x65,  0x78,  0xFF,  0x05,  0x02,  0x6F,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x67,  0x61,  0x6D,  0x65,  0x20,  0x82,  0x68,  0xFF,  0x05,  0x01,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x00,  0xFF,  0x05,  0x01,  0x0A,  0x87,  0x40,  0xFF,  0x05,  0x04,  0x49,  0x27,  0x6D,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x69,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x68,  0x69,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x66,
 0x61,  0x78,  0xFF,  0x05,  0x07,  0x76,  0x6F,  0x72,  0x69,  0x74,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x73,  0x75,  0x6E,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x06,  0x64,  0x72,  0x65,  0x73,  0x73,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x03,  0x77,  0x61,  0x74,  0x78,  0xFF,  0x05,  0x06,  0x63,  0x68,  0x69,  0x6E,  0x67,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x6D,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x04,
 0x67,  0x65,  0x74,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x75,  0x6E,  0x81,  0x70,  0xFF,  0x05,  0x08,  0x64,  0x72,  0x65,  0x73,  0x73,  0x65,  0x64,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x74,  0x61,  0x6B,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x74,  0x68,  0x61,  0x74,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x62,  0x6F,  0x78,  0xFF,  0x05,  0x03,  0x64,  0x79,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x64,
 0x6F,  0x77,  0x6E,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x74,  0x6F,  0x77,  0x6E,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x00,  0xFF,  0x05,  0x01,  0x0A,  0x89,  0x30,  0xFF,  0x05,  0x02,  0x49,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x73,  0x61,  0x79,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x68,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x62,  0x65,  0x73,  0x81,  0x70,  0xFF,  0x05,
 0x05,  0x74,  0x65,  0x73,  0x74,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x6C,  0x65,  0x61,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x69,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x66,  0x6F,  0x72,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x61,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x62,  0x69,  0x67,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x6B,  0x69,  0x73,  0x73,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,
 0x81,  0x70,  0xFF,  0x05,  0x04,  0x70,  0x75,  0x74,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x68,  0x69,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x66,  0x61,  0x78,  0xFF,  0x05,  0x07,  0x76,  0x6F,  0x72,  0x69,  0x74,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x70,  0x65,  0x72,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x66,  0x75,  0x6D,  0x65,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x03,  0x6F,  0x6E,  0x20,  0x81,  0x16,  0xFF,  0x05,  0x01,  0x20,
 0x00,  0xFF,  0x05,  0x01,  0x0D,  0x00,  0xFF,  0x05,  0x01,  0x0A,  0x83,  0x42,  0xFF,  0x05,  0x03,  0x47,  0x6F,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x70,  0x6C,  0x61,  0x79,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x61,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x76,  0x69,  0x78,  0xFF,  0x05,  0x02,  0x64,  0x65,  0x78,  0xFF,  0x05,  0x02,  0x6F,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x67,  0x61,  0x6D,  0x65,  0x20,  0x00,  0xFF,  0x05,  0x01,
 0x0D,  0x00,  0xFF,  0x05,  0x01,  0x0A,  0x8A,  0x28,  0xFF,  0x05,  0x05,  0x49,  0x74,  0x27,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x82,  0x68,  0xFF,  0x05,  0x05,  0x49,  0x74,  0x27,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x82,  0x68,  0xFF,  0x05,  0x05,  0x49,  0x74,  0x27,  0x73,  0x20,  0x78,  0xFF,  0x05,
 0x04,  0x61,  0x6C,  0x6C,  0x20,  0x82,  0x68,  0xFF,  0x05,  0x04,  0x66,  0x6F,  0x72,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x45,  0x76,  0x65,  0x82,  0x68,  0xFF,  0x05,  0x02,  0x72,  0x79,  0x78,  0xFF,  0x05,  0x06,  0x74,  0x68,  0x69,  0x6E,  0x67,  0x20,  0x82,  0x68,  0xFF,  0x05,  0x02,  0x49,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x64,  0x6F,
 0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x82,  0x68,  0xFF,  0x05,  0x02,  0x49,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x74,  0x65,  0x6C,  0x6C,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x61,  0x6C,  0x6C,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x68,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x74,  0x69,  0x6D,  0x65,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x83,  0x60,  0xFF,  0x05,  0x03,
 0x48,  0x65,  0x61,  0x78,  0xFF,  0x05,  0x04,  0x76,  0x65,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x69,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x61,  0x20,  0x78,  0xFF,  0x05,  0x06,  0x70,  0x6C,  0x61,  0x63,  0x65,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x03,  0x6F,  0x6E,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x06,  0x65,  0x61,  0x72,  0x74,  0x68,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x77,  0x69,  0x74,  0x68,  0x20,  0x78,  0xFF,
 0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x84,  0x58,  0xFF,  0x05,  0x05,  0x54,  0x65,  0x6C,  0x6C,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x6D,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x61,  0x6C,  0x6C,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x68,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x07,  0x74,  0x68,  0x69,  0x6E,  0x67,  0x73,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x81,
 0x70,  0xFF,  0x05,  0x05,  0x77,  0x61,  0x6E,  0x74,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x03,  0x74,  0x6F,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x64,  0x6F,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x01,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x82,  0x68,  0xFF,  0x05,  0x08,  0x49,  0x5F,  0x68,  0x65,  0x61,  0x72,  0x64,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x74,  0x68,  0x61,  0x74,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,
 0x20,  0x78,  0xFF,  0x05,  0x05,  0x6C,  0x69,  0x6B,  0x65,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x74,  0x68,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x62,  0x61,  0x64,  0x20,  0x82,  0x68,  0xFF,  0x05,  0x06,  0x67,  0x69,  0x72,  0x6C,  0x73,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x02,  0x48,  0x6F,  0x83,  0x60,  0xFF,  0x05,  0x04,  0x6E,  0x65,  0x79,  0x20,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x69,
 0x73,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x74,  0x68,  0x61,  0x74,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x74,  0x72,  0x75,  0x65,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x01,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x84,  0x58,  0xFF,  0x05,  0x05,  0x49,  0x74,  0x27,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x62,  0x65,  0x74,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x65,  0x72,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x74,  0x68,  0x61,
 0x6E,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x49,  0x20,  0x78,  0xFF,  0x05,  0x01,  0x65,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x76,  0x65,  0x72,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x01,  0x65,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x76,  0x65,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x6B,  0x6E,  0x65,  0x77,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x84,  0x58,  0xFF,  0x05,  0x05,  0x54,  0x68,  0x65,  0x79,  0x20,  0x78,  0xFF,  0x05,  0x04,
 0x73,  0x61,  0x79,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x74,  0x68,  0x61,  0x74,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x68,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x06,  0x77,  0x6F,  0x72,  0x6C,  0x64,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x77,  0x61,  0x73,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x06,  0x62,  0x75,  0x69,  0x6C,  0x74,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x66,  0x6F,  0x72,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x74,
 0x77,  0x6F,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x01,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x82,  0x68,  0xFF,  0x05,  0x02,  0x4F,  0x6E,  0x81,  0x70,  0xFF,  0x05,  0x03,  0x6C,  0x79,  0x20,  0x78,  0xFF,  0x05,  0x06,  0x77,  0x6F,  0x72,  0x74,  0x68,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x6C,  0x69,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x76,  0x69,  0x6E,  0x67,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x03,  0x69,  0x66,  0x20,  0x81,  0x70,
 0xFF,  0x05,  0x04,  0x73,  0x6F,  0x6D,  0x65,  0x81,  0x70,  0xFF,  0x05,  0x02,  0x62,  0x6F,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x64,  0x79,  0x20,  0x85,  0x50,  0xFF,  0x05,  0x03,  0x69,  0x73,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x02,  0x6C,  0x6F,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x76,  0x69,  0x6E,  0x67,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x78,  0xFF,  0x05,  0x01,  0x20,  0x00,  0xFF,  0x05,  0x01,
 0x0D,  0x88,  0x38,  0xFF,  0x05,  0x02,  0x42,  0x61,  0x78,  0xFF,  0x05,  0x03,  0x62,  0x79,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x6E,  0x6F,  0x77,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x64,  0x6F,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x01,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x00,  0xFF,  0x05,  0x01,  0x0A,  0x9D,  0x08,  0xFF,  0x05,  0x03,  0x53,  0x69,  0x6E,  0x78,  0xFF,
 0x05,  0x05,  0x67,  0x69,  0x6E,  0x67,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x69,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x68,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x6F,  0x6C,  0x64,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x62,  0x61,  0x72,  0x73,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x53,  0x77,  0x69,  0x6E,  0x78,  0xFF,  0x05,  0x05,  0x67,  0x69,  0x6E,  0x67,  0x20,  0x78,  0xFF,
 0x05,  0x05,  0x77,  0x69,  0x74,  0x68,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x68,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x6F,  0x6C,  0x64,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x06,  0x73,  0x74,  0x61,  0x72,  0x73,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x02,  0x4C,  0x69,  0x78,  0xFF,  0x05,  0x05,  0x76,  0x69,  0x6E,  0x67,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x66,  0x6F,  0x72,  0x20,  0x78,  0xFF,
 0x05,  0x04,  0x74,  0x68,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x66,  0x61,  0x6D,  0x65,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x00,  0xFF,  0x05,  0x01,  0x0A,  0x8B,  0x20,  0xFF,  0x05,  0x03,  0x4B,  0x69,  0x73,  0x78,  0xFF,  0x05,  0x05,  0x73,  0x69,  0x6E,  0x67,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x69,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x68,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x62,  0x6C,  0x75,  0x65,
 0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x64,  0x61,  0x72,  0x6B,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x03,  0x50,  0x6C,  0x61,  0x78,  0xFF,  0x05,  0x05,  0x79,  0x69,  0x6E,  0x67,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x70,  0x6F,  0x6F,  0x6C,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x61,  0x6E,  0x64,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x77,  0x69,  0x6C,  0x64,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x06,  0x64,
 0x61,  0x72,  0x74,  0x73,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x02,  0x56,  0x69,  0x78,  0xFF,  0x05,  0x02,  0x64,  0x65,  0x78,  0xFF,  0x05,  0x02,  0x6F,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x06,  0x67,  0x61,  0x6D,  0x65,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x01,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x00,  0xFF,  0x05,  0x01,  0x0A,  0x89,  0x30,  0xFF,  0x05,  0x03,  0x48,  0x65,  0x20,  0x78,  0xFF,
 0x05,  0x06,  0x68,  0x6F,  0x6C,  0x64,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x6D,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x69,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x68,  0x69,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x62,  0x69,  0x67,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x61,  0x72,  0x6D,  0x73,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x06,  0x64,  0x72,  0x75,  0x6E,  0x6B,  0x20,  0x78,
 0xFF,  0x05,  0x04,  0x61,  0x6E,  0x64,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x49,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x61,  0x6D,  0x20,  0x78,  0xFF,  0x05,  0x07,  0x73,  0x65,  0x65,  0x69,  0x6E,  0x67,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x06,  0x73,  0x74,  0x61,  0x72,  0x73,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x74,  0x68,  0x69,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x69,  0x73,  0x20,  0x78,
 0xFF,  0x05,  0x04,  0x61,  0x6C,  0x6C,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x49,  0x20,  0x78,  0xFF,  0x05,  0x06,  0x74,  0x68,  0x69,  0x6E,  0x6B,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x03,  0x6F,  0x66,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x00,  0xFF,  0x05,  0x01,  0x0A,  0x89,  0x30,  0xFF,  0x05,  0x03,  0x57,  0x61,  0x74,  0x78,  0xFF,  0x05,  0x06,  0x63,  0x68,  0x69,  0x6E,  0x67,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x61,  0x6C,
 0x6C,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x6F,  0x75,  0x72,  0x20,  0x78,  0xFF,  0x05,  0x08,  0x66,  0x72,  0x69,  0x65,  0x6E,  0x64,  0x73,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x66,  0x61,  0x6C,  0x6C,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x03,  0x69,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x61,  0x6E,  0x64,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x6F,  0x75,  0x74,  0x20,  0x78,  0xFF,  0x05,  0x03,
 0x6F,  0x66,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x6F,  0x6C,  0x64,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x07,  0x70,  0x61,  0x75,  0x6C,  0x27,  0x73,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x74,  0x68,  0x69,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x69,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x6D,  0x79,  0x20,  0x78,  0xFF,  0x05,  0x01,  0x69,  0x78,  0xFF,  0x05,  0x04,  0x64,  0x65,  0x61,  0x20,
 0x81,  0x70,  0xFF,  0x05,  0x03,  0x6F,  0x66,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x66,  0x75,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x01,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x84,  0x58,  0xFF,  0x05,  0x04,  0x50,  0x6C,  0x61,  0x79,  0x78,  0xFF,  0x05,  0x04,  0x69,  0x6E,  0x67,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x76,  0x69,  0x78,  0xFF,  0x05,  0x02,  0x64,  0x65,  0x78,  0xFF,  0x05,  0x02,  0x6F,  0x20,  0x81,  0x70,  0xFF,
 0x05,  0x06,  0x67,  0x61,  0x6D,  0x65,  0x73,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x00,  0xFF,  0x05,  0x01,  0x0A,  0x8A,  0x28,  0xFF,  0x05,  0x05,  0x49,  0x74,  0x27,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x82,  0x68,  0xFF,  0x05,  0x05,  0x49,  0x74,  0x27,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x82,
 0x68,  0xFF,  0x05,  0x05,  0x49,  0x74,  0x27,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x61,  0x6C,  0x6C,  0x20,  0x82,  0x68,  0xFF,  0x05,  0x04,  0x66,  0x6F,  0x72,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x45,  0x76,  0x65,  0x82,  0x68,  0xFF,  0x05,  0x02,  0x72,  0x79,  0x78,  0xFF,  0x05,  0x06,  0x74,  0x68,  0x69,  0x6E,  0x67,  0x20,  0x82,  0x68,  0xFF,  0x05,  0x02,  0x49,
 0x20,  0x78,  0xFF,  0x05,  0x03,  0x64,  0x6F,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x82,  0x68,  0xFF,  0x05,  0x02,  0x49,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x74,  0x65,  0x6C,  0x6C,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x61,  0x6C,  0x6C,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x68,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x74,  0x69,  0x6D,  0x65,  0x20,  0x00,  0xFF,  0x05,
 0x01,  0x0D,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x48,  0x65,  0x61,  0x78,  0xFF,  0x05,  0x04,  0x76,  0x65,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x69,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x61,  0x20,  0x78,  0xFF,  0x05,  0x06,  0x70,  0x6C,  0x61,  0x63,  0x65,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x03,  0x6F,  0x6E,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x06,  0x65,  0x61,  0x72,  0x74,  0x68,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,
 0x77,  0x69,  0x74,  0x68,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x84,  0x58,  0xFF,  0x05,  0x05,  0x54,  0x65,  0x6C,  0x6C,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x6D,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x61,  0x6C,  0x6C,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x68,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x07,  0x74,  0x68,  0x69,  0x6E,  0x67,  0x73,  0x20,  0x81,  0x70,  0xFF,
 0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x77,  0x61,  0x6E,  0x74,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x03,  0x74,  0x6F,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x64,  0x6F,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x01,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x82,  0x68,  0xFF,  0x05,  0x08,  0x49,  0x5F,  0x68,  0x65,  0x61,  0x72,  0x64,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x74,  0x68,  0x61,  0x74,  0x20,
 0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x6C,  0x69,  0x6B,  0x65,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x74,  0x68,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x62,  0x61,  0x64,  0x20,  0x82,  0x68,  0xFF,  0x05,  0x06,  0x67,  0x69,  0x72,  0x6C,  0x73,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x02,  0x48,  0x6F,  0x83,  0x60,  0xFF,  0x05,  0x04,  0x6E,  0x65,  0x79,
 0x20,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x69,  0x73,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x74,  0x68,  0x61,  0x74,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x74,  0x72,  0x75,  0x65,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x01,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x84,  0x58,  0xFF,  0x05,  0x05,  0x49,  0x74,  0x27,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x62,  0x65,  0x74,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x65,  0x72,  0x20,
 0x78,  0xFF,  0x05,  0x05,  0x74,  0x68,  0x61,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x49,  0x20,  0x78,  0xFF,  0x05,  0x01,  0x65,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x76,  0x65,  0x72,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x01,  0x65,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x76,  0x65,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x6B,  0x6E,  0x65,  0x77,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x84,  0x58,  0xFF,  0x05,  0x05,  0x54,  0x68,
 0x65,  0x79,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x73,  0x61,  0x79,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x74,  0x68,  0x61,  0x74,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x68,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x06,  0x77,  0x6F,  0x72,  0x6C,  0x64,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x77,  0x61,  0x73,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x06,  0x62,  0x75,  0x69,  0x6C,  0x74,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x66,  0x6F,
 0x72,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x77,  0x6F,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x01,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x82,  0x68,  0xFF,  0x05,  0x02,  0x4F,  0x6E,  0x81,  0x70,  0xFF,  0x05,  0x03,  0x6C,  0x79,  0x20,  0x78,  0xFF,  0x05,  0x06,  0x77,  0x6F,  0x72,  0x74,  0x68,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x6C,  0x69,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x76,  0x69,  0x6E,  0x67,  0x20,  0x81,  0x70,  0xFF,
 0x05,  0x03,  0x69,  0x66,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x73,  0x6F,  0x6D,  0x65,  0x81,  0x70,  0xFF,  0x05,  0x02,  0x62,  0x6F,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x64,  0x79,  0x20,  0x85,  0x50,  0xFF,  0x05,  0x03,  0x69,  0x73,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x02,  0x6C,  0x6F,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x76,  0x69,  0x6E,  0x67,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x78,  0xFF,
 0x05,  0x01,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x88,  0x38,  0xFF,  0x05,  0x02,  0x42,  0x61,  0x78,  0xFF,  0x05,  0x03,  0x62,  0x79,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x6E,  0x6F,  0x77,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x64,  0x6F,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x01,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x00,  0xFF,  0x05,  0x01,  0x0A,  0x82,  0x68,  0xFF,
 0x05,  0x05,  0x28,  0x6E,  0x6F,  0x77,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x64,  0x6F,  0x20,  0x84,  0x58,  0xFF,  0x05,  0x04,  0x6E,  0x6F,  0x77,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x64,  0x6F,  0x20,  0x84,  0x58,  0xFF,  0x05,  0x04,  0x6E,  0x6F,  0x77,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,
 0x20,  0x78,  0xFF,  0x05,  0x04,  0x64,  0x6F,  0x29,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x8C,  0x18,  0xFF,  0x05,  0x05,  0x28,  0x6E,  0x6F,  0x77,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x64,  0x6F,  0x20,  0x84,  0x58,  0xFF,  0x05,  0x04,  0x6E,  0x6F,  0x77,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x64,  0x6F,  0x20,
 0x84,  0x58,  0xFF,  0x05,  0x04,  0x6E,  0x6F,  0x77,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x64,  0x6F,  0x29,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x00,  0xFF,  0x05,  0x01,  0x0A,  0x00,  0xFF,  0x05,  0x01,  0x0A,  0x87,  0x40,  0xFF,  0x05,  0x05,  0x49,  0x74,  0x27,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x82,
 0x68,  0xFF,  0x05,  0x05,  0x49,  0x74,  0x27,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x82,  0x68,  0xFF,  0x05,  0x05,  0x49,  0x74,  0x27,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x61,  0x6C,  0x6C,  0x20,  0x82,  0x68,  0xFF,  0x05,  0x04,  0x66,  0x6F,  0x72,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x83,  0x60,  0xFF,
 0x05,  0x03,  0x45,  0x76,  0x65,  0x82,  0x68,  0xFF,  0x05,  0x02,  0x72,  0x79,  0x78,  0xFF,  0x05,  0x06,  0x74,  0x68,  0x69,  0x6E,  0x67,  0x20,  0x82,  0x68,  0xFF,  0x05,  0x02,  0x49,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x64,  0x6F,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x82,  0x68,  0xFF,  0x05,  0x02,  0x49,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x74,  0x65,  0x6C,  0x6C,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,
 0x78,  0xFF,  0x05,  0x04,  0x61,  0x6C,  0x6C,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x68,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x74,  0x69,  0x6D,  0x65,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x48,  0x65,  0x61,  0x78,  0xFF,  0x05,  0x04,  0x76,  0x65,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x69,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x61,  0x20,  0x78,  0xFF,  0x05,  0x06,  0x70,  0x6C,
 0x61,  0x63,  0x65,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x03,  0x6F,  0x6E,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x06,  0x65,  0x61,  0x72,  0x74,  0x68,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x77,  0x69,  0x74,  0x68,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x84,  0x58,  0xFF,  0x05,  0x05,  0x54,  0x65,  0x6C,  0x6C,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x6D,  0x65,  0x20,  0x78,  0xFF,
 0x05,  0x04,  0x61,  0x6C,  0x6C,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x68,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x07,  0x74,  0x68,  0x69,  0x6E,  0x67,  0x73,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x77,  0x61,  0x6E,  0x74,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x03,  0x74,  0x6F,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x64,  0x6F,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x01,  0x20,
 0x00,  0xFF,  0x05,  0x01,  0x0D,  0x82,  0x68,  0xFF,  0x05,  0x08,  0x49,  0x5F,  0x68,  0x65,  0x61,  0x72,  0x64,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x74,  0x68,  0x61,  0x74,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x6C,  0x69,  0x6B,  0x65,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x74,  0x68,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x62,  0x61,  0x64,  0x20,  0x82,  0x68,  0xFF,
 0x05,  0x06,  0x67,  0x69,  0x72,  0x6C,  0x73,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x81,  0x70,  0xFF,  0x05,  0x02,  0x48,  0x6F,  0x83,  0x60,  0xFF,  0x05,  0x04,  0x6E,  0x65,  0x79,  0x20,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x69,  0x73,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x74,  0x68,  0x61,  0x74,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x74,  0x72,  0x75,  0x65,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x01,  0x20,  0x00,  0xFF,
 0x05,  0x01,  0x0D,  0x84,  0x58,  0xFF,  0x05,  0x05,  0x49,  0x74,  0x27,  0x73,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x62,  0x65,  0x74,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x65,  0x72,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x74,  0x68,  0x61,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x49,  0x20,  0x78,  0xFF,  0x05,  0x01,  0x65,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x76,  0x65,  0x72,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x01,  0x65,  0x81,  0x70,
 0xFF,  0x05,  0x04,  0x76,  0x65,  0x6E,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x6B,  0x6E,  0x65,  0x77,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x84,  0x58,  0xFF,  0x05,  0x05,  0x54,  0x68,  0x65,  0x79,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x73,  0x61,  0x79,  0x20,  0x78,  0xFF,  0x05,  0x05,  0x74,  0x68,  0x61,  0x74,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x68,  0x65,  0x20,  0x78,  0xFF,  0x05,  0x06,  0x77,  0x6F,  0x72,  0x6C,  0x64,
 0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x77,  0x61,  0x73,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x06,  0x62,  0x75,  0x69,  0x6C,  0x74,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x66,  0x6F,  0x72,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x74,  0x77,  0x6F,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x01,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x82,  0x68,  0xFF,  0x05,  0x02,  0x4F,  0x6E,  0x81,  0x70,  0xFF,  0x05,  0x03,  0x6C,  0x79,  0x20,  0x78,
 0xFF,  0x05,  0x06,  0x77,  0x6F,  0x72,  0x74,  0x68,  0x20,  0x78,  0xFF,  0x05,  0x02,  0x6C,  0x69,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x76,  0x69,  0x6E,  0x67,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x03,  0x69,  0x66,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x73,  0x6F,  0x6D,  0x65,  0x81,  0x70,  0xFF,  0x05,  0x02,  0x62,  0x6F,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x64,  0x79,  0x20,  0x85,  0x50,  0xFF,  0x05,  0x03,  0x69,  0x73,  0x20,
 0x81,  0x70,  0xFF,  0x05,  0x02,  0x6C,  0x6F,  0x81,  0x70,  0xFF,  0x05,  0x05,  0x76,  0x69,  0x6E,  0x67,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x78,  0xFF,  0x05,  0x01,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x88,  0x38,  0xFF,  0x05,  0x02,  0x42,  0x61,  0x78,  0xFF,  0x05,  0x03,  0x62,  0x79,  0x20,  0x78,  0xFF,  0x05,  0x04,  0x6E,  0x6F,  0x77,  0x20,  0x81,  0x70,  0xFF,  0x05,  0x04,  0x79,  0x6F,
 0x75,  0x20,  0x78,  0xFF,  0x05,  0x03,  0x64,  0x6F,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x00,  0xFF,  0x05,  0x01,  0x0A,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x6D,  0x6D,  0x20,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x6D,  0x6D,  0x20,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x6D,  0x6D,  0x20,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x6D,  0x6D,  0x20,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x6D,  0x6D,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x83,
 0x60,  0xFF,  0x05,  0x03,  0x6D,  0x6D,  0x20,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x6D,  0x6D,  0x20,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x6D,  0x6D,  0x20,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x6D,  0x6D,  0x20,  0x83,  0x60,  0xFF,  0x05,  0x03,  0x6D,  0x6D,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x83,  0x60,  0xFF,  0x05,  0x04,  0x6E,  0x6F,  0x77,  0x20,  0x83,  0x60,  0xFF,  0x05,  0x04,  0x79,  0x6F,  0x75,  0x20,  0x83,  0x60,  0xFF,
 0x05,  0x03,  0x64,  0x6F,  0x20,  0x00,  0xFF,  0x05,  0x01,  0x0D,  0x8C,  0x48,  0xFF,  0x2F,  0x00,  0x4D,  0x54,  0x72,  0x6B,  0x00,  0x00,  0x43,  0xFB,  0x00,  0xFF,  0x03,  0x14,  0x41,  0x63,  0x6F,  0x75,  0x73,  0x74,  0x69,  0x63,  0x20,  0x47,  0x72,  0x61,  0x6E,  0x64,  0x20,  0x50,  0x69,  0x61,  0x6E,  0x6F,  0x00,  0xE0,  0x00,  0x40,  0x00,  0xB0,  0x0B,  0x7F,  0x00,  0xB0,  0x40,  0x00,  0x00,  0xB0,  0x5D,  0x00,  0x00,  0xB0,
 0x5B,  0x00,  0x00,  0xB0,  0x79,  0x00,  0x00,  0xB0,  0x00,  0x00,  0x00,  0xB0,  0x0A,  0x40,  0x00,  0xB0,  0x07,  0x78,  0x00,  0xC0,  0x00,  0x9E,  0x00,  0x90,  0x39,  0x32,  0x00,  0x90,  0x31,  0x24,  0x06,  0x90,  0x2A,  0x1F,  0x81,  0x10,  0xB0,  0x40,  0x08,  0x04,  0xB0,  0x40,  0x66,  0x05,  0xB0,  0x40,  0x7F,  0x7B,  0x80,  0x31,  0x00,  0x07,  0x80,  0x2A,  0x00,  0x06,  0x80,  0x39,  0x00,  0x81,  0x30,  0x90,  0x2D,  0x1F,  0x09,
 0x90,  0x3D,  0x40,  0x0C,  0xB0,  0x40,  0x52,  0x01,  0x90,  0x34,  0x21,  0x03,  0xB0,  0x40,  0x18,  0x04,  0xB0,  0x40,  0x00,  0x81,  0x11,  0xB0,  0x40,  0x10,  0x04,  0xB0,  0x40,  0x68,  0x05,  0xB0,  0x40,  0x7F,  0x6F,  0x80,  0x2D,  0x00,  0x06,  0x80,  0x34,  0x00,  0x2D,  0x80,  0x3D,  0x00,  0x81,  0x16,  0xB0,  0x40,  0x7E,  0x06,  0xB0,  0x40,  0x2E,  0x02,  0x90,  0x39,  0x32,  0x01,  0x90,  0x31,  0x1F,  0x01,  0xB0,  0x40,  0x04,
 0x04,  0xB0,  0x40,  0x00,  0x04,  0x90,  0x2A,  0x1A,  0x81,  0x25,  0xB0,  0x40,  0x08,  0x04,  0xB0,  0x40,  0x66,  0x06,  0xB0,  0x40,  0x7F,  0x45,  0x80,  0x2A,  0x00,  0x2D,  0x80,  0x31,  0x00,  0x05,  0x80,  0x39,  0x00,  0x81,  0x1F,  0xB0,  0x40,  0x5E,  0x03,  0x90,  0x3D,  0x36,  0x00,  0xB0,  0x40,  0x1E,  0x00,  0x90,  0x34,  0x34,  0x07,  0xB0,  0x40,  0x00,  0x05,  0x90,  0x2D,  0x2C,  0x81,  0x19,  0xB0,  0x40,  0x04,  0x06,  0xB0,
 0x40,  0x64,  0x04,  0xB0,  0x40,  0x7F,  0x4D,  0x80,  0x34,  0x00,  0x03,  0x80,  0x2D,  0x00,  0x25,  0x80,  0x3D,  0x00,  0x81,  0x39,  0x90,  0x34,  0x34,  0x01,  0x90,  0x25,  0x20,  0x02,  0x90,  0x2C,  0x24,  0x07,  0xB0,  0x40,  0x5C,  0x05,  0xB0,  0x40,  0x1C,  0x05,  0xB0,  0x40,  0x00,  0x81,  0x3B,  0xB0,  0x40,  0x5C,  0x05,  0xB0,  0x40,  0x7F,  0x44,  0x80,  0x2C,  0x00,  0x0B,  0x80,  0x25,  0x00,  0x07,  0x80,  0x34,  0x00,  0x81,
 0x3F,  0xB0,  0x40,  0x54,  0x02,  0x90,  0x36,  0x30,  0x04,  0xB0,  0x40,  0x18,  0x01,  0x90,  0x26,  0x1A,  0x00,  0x90,  0x2D,  0x18,  0x04,  0xB0,  0x40,  0x00,  0x81,  0x19,  0xB0,  0x40,  0x5C,  0x05,  0xB0,  0x40,  0x7F,  0x53,  0x80,  0x26,  0x00,  0x01,  0x80,  0x2D,  0x00,  0x2B,  0x80,  0x36,  0x00,  0x81,  0x62,  0x90,  0x39,  0x2A,  0x07,  0x90,  0x31,  0x19,  0x07,  0x90,  0x2A,  0x16,  0x04,  0xB0,  0x40,  0x7F,  0x1B,  0xB0,  0x40,
 0x00,  0x72,  0xB0,  0x40,  0x7F,  0x82,  0x46,  0xB0,  0x40,  0x7A,  0x05,  0xB0,  0x40,  0x7F,  0x81,  0x19,  0x80,  0x31,  0x00,  0x32,  0x80,  0x2A,  0x00,  0x06,  0x80,  0x39,  0x00,  0x81,  0x47,  0x90,  0x39,  0x32,  0x06,  0x90,  0x36,  0x3C,  0x08,  0x90,  0x2A,  0x1B,  0x00,  0x90,  0x31,  0x24,  0x00,  0x90,  0x3D,  0x3E,  0x7B,  0xB0,  0x40,  0x50,  0x05,  0xB0,  0x40,  0x16,  0x05,  0xB0,  0x40,  0x00,  0x4C,  0xB0,  0x40,  0x1C,  0x05,
 0xB0,  0x40,  0x70,  0x05,  0xB0,  0x40,  0x7F,  0x2D,  0x80,  0x31,  0x00,  0x08,  0x80,  0x2A,  0x00,  0x04,  0x80,  0x39,  0x00,  0x02,  0x80,  0x36,  0x00,  0x06,  0x80,  0x3D,  0x00,  0x81,  0x2A,  0x90,  0x2D,  0x2A,  0x15,  0x90,  0x39,  0x32,  0x00,  0x90,  0x3D,  0x46,  0x05,  0x90,  0x34,  0x38,  0x11,  0x90,  0x40,  0x48,  0x05,  0xB0,  0x40,  0x56,  0x06,  0xB0,  0x40,  0x1A,  0x04,  0xB0,  0x40,  0x00,  0x60,  0xB0,  0x40,  0x1C,  0x05,
 0xB0,  0x40,  0x70,  0x05,  0xB0,  0x40,  0x7F,  0x3A,  0x80,  0x2D,  0x00,  0x08,  0x80,  0x34,  0x00,  0x0D,  0x80,  0x40,  0x00,  0x30,  0x80,  0x39,  0x00,  0x02,  0x80,  0x3D,  0x00,  0x81,  0x38,  0x90,  0x2A,  0x28,  0x0D,  0x90,  0x36,  0x34,  0x01,  0x90,  0x39,  0x40,  0x08,  0x90,  0x31,  0x20,  0x03,  0x90,  0x3D,  0x3E,  0x04,  0xB0,  0x40,  0x50,  0x08,  0xB0,  0x40,  0x16,  0x05,  0xB0,  0x40,  0x00,  0x65,  0xB0,  0x40,  0x5A,  0x06,
 0xB0,  0x40,  0x7F,  0x32,  0x80,  0x36,  0x00,  0x0B,  0x80,  0x31,  0x00,  0x08,  0x80,  0x2A,  0x00,  0x27,  0x80,  0x39,  0x00,  0x13,  0x80,  0x3D,  0x00,  0x81,  0x4B,  0x90,  0x2D,  0x2A,  0x0A,  0x90,  0x3D,  0x3E,  0x04,  0xB0,  0x40,  0x7E,  0x01,  0x90,  0x34,  0x2A,  0x04,  0xB0,  0x40,  0x2E,  0x02,  0x90,  0x40,  0x40,  0x01,  0x90,  0x39,  0x30,  0x02,  0xB0,  0x40,  0x02,  0x04,  0xB0,  0x40,  0x00,  0x62,  0xB0,  0x40,  0x56,  0x03,
 0xB0,  0x40,  0x7F,  0x3E,  0x80,  0x34,  0x00,  0x0E,  0x80,  0x39,  0x00,  0x05,  0x80,  0x40,  0x00,  0x02,  0x80,  0x2D,  0x00,  0x22,  0x80,  0x3D,  0x00,  0x81,  0x5A,  0x90,  0x2C,  0x2C,  0x02,  0x90,  0x25,  0x26,  0x04,  0x90,  0x34,  0x2C,  0x02,  0x90,  0x38,  0x38,  0x0E,  0xB0,  0x40,  0x50,  0x03,  0xB0,  0x40,  0x14,  0x06,  0xB0,  0x40,  0x00,  0x00,  0x90,  0x40,  0x3E,  0x64,  0xB0,  0x40,  0x5C,  0x05,  0xB0,  0x40,  0x7F,  0x53,
 0x80,  0x34,  0x00,  0x13,  0x80,  0x40,  0x00,  0x00,  0x80,  0x2C,  0x00,  0x05,  0x80,  0x38,  0x00,  0x09,  0x80,  0x25,  0x00,  0x81,  0x70,  0x90,  0x2D,  0x21,  0x03,  0x90,  0x26,  0x1E,  0x00,  0x90,  0x36,  0x24,  0x00,  0x90,  0x39,  0x27,  0x03,  0x90,  0x3E,  0x2C,  0x04,  0xB0,  0x40,  0x54,  0x04,  0xB0,  0x40,  0x18,  0x05,  0xB0,  0x40,  0x00,  0x81,  0x3B,  0xB0,  0x40,  0x58,  0x06,  0xB0,  0x40,  0x7F,  0x2B,  0x80,  0x26,  0x00,
 0x0B,  0x80,  0x2D,  0x00,  0x01,  0x80,  0x36,  0x00,  0x13,  0x80,  0x39,  0x00,  0x03,  0x80,  0x3E,  0x00,  0x81,  0x6D,  0x90,  0x36,  0x1D,  0x02,  0x90,  0x2A,  0x19,  0x00,  0x90,  0x31,  0x1A,  0x03,  0x90,  0x39,  0x30,  0x05,  0x90,  0x3D,  0x28,  0x02,  0xB0,  0x40,  0x52,  0x03,  0xB0,  0x40,  0x18,  0x04,  0xB0,  0x40,  0x00,  0x82,  0x5F,  0xB0,  0x40,  0x0C,  0x05,  0xB0,  0x40,  0x10,  0x05,  0xB0,  0x40,  0x5E,  0x04,  0xB0,  0x40,
 0x7F,  0x81,  0x19,  0x80,  0x39,  0x00,  0x01,  0x80,  0x36,  0x00,  0x2B,  0x80,  0x2A,  0x00,  0x0D,  0x80,  0x3D,  0x00,  0x0A,  0x80,  0x31,  0x00,  0x82,  0x58,  0x90,  0x2A,  0x2E,  0x00,  0x90,  0x31,  0x2A,  0x12,  0x90,  0x39,  0x2A,  0x19,  0xB0,  0x40,  0x6E,  0x05,  0xB0,  0x40,  0x26,  0x04,  0xB0,  0x40,  0x01,  0x05,  0xB0,  0x40,  0x00,  0x43,  0xB0,  0x40,  0x0A,  0x05,  0xB0,  0x40,  0x68,  0x00,  0x80,  0x39,  0x00,  0x04,  0xB0,
 0x40,  0x7F,  0x17,  0x90,  0x39,  0x36,  0x51,  0x80,  0x39,  0x00,  0x1D,  0x90,  0x39,  0x36,  0x2B,  0x80,  0x39,  0x00,  0x13,  0x80,  0x2A,  0x00,  0x01,  0x80,  0x31,  0x00,  0x3A,  0x90,  0x39,  0x32,  0x33,  0x80,  0x39,  0x00,  0x4C,  0x90,  0x2D,  0x2C,  0x03,  0x90,  0x34,  0x26,  0x02,  0x90,  0x3D,  0x34,  0x04,  0xB0,  0x40,  0x52,  0x05,  0xB0,  0x40,  0x1A,  0x05,  0xB0,  0x40,  0x00,  0x68,  0xB0,  0x40,  0x1C,  0x05,  0xB0,  0x40,
 0x70,  0x06,  0xB0,  0x40,  0x7F,  0x28,  0x80,  0x3D,  0x00,  0x44,  0x90,  0x3D,  0x34,  0x18,  0x80,  0x2D,  0x00,  0x12,  0x80,  0x34,  0x00,  0x2B,  0x80,  0x3D,  0x00,  0x81,  0x14,  0xB0,  0x40,  0x52,  0x02,  0x90,  0x2A,  0x2A,  0x03,  0x90,  0x39,  0x32,  0x02,  0xB0,  0x40,  0x1A,  0x01,  0x90,  0x31,  0x2C,  0x01,  0xB0,  0x40,  0x02,  0x05,  0xB0,  0x40,  0x00,  0x5B,  0xB0,  0x40,  0x60,  0x05,  0xB0,  0x40,  0x7F,  0x05,  0x80,  0x39,
 0x00,  0x0D,  0x90,  0x39,  0x30,  0x38,  0x80,  0x39,  0x00,  0x3C,  0x90,  0x39,  0x2E,  0x37,  0x80,  0x39,  0x00,  0x3D,  0x80,  0x2A,  0x00,  0x0F,  0x80,  0x31,  0x00,  0x02,  0x90,  0x39,  0x30,  0x26,  0x80,  0x39,  0x00,  0x59,  0xB0,  0x40,  0x54,  0x05,  0x90,  0x34,  0x20,  0x00,  0xB0,  0x40,  0x1A,  0x02,  0x90,  0x3D,  0x36,  0x02,  0xB0,  0x40,  0x00,  0x06,  0x90,  0x2D,  0x1E,  0x5F,  0xB0,  0x40,  0x38,  0x05,  0xB0,  0x40,  0x7E,
 0x05,  0xB0,  0x40,  0x7F,  0x2D,  0x80,  0x3D,  0x00,  0x4E,  0x90,  0x3D,  0x3A,  0x01,  0x80,  0x34,  0x00,  0x07,  0x80,  0x2D,  0x00,  0x42,  0x80,  0x3D,  0x00,  0x65,  0xB0,  0x40,  0x32,  0x05,  0xB0,  0x40,  0x0A,  0x02,  0x90,  0x25,  0x2C,  0x02,  0xB0,  0x40,  0x00,  0x00,  0x90,  0x2C,  0x1F,  0x0A,  0x90,  0x34,  0x38,  0x47,  0x80,  0x34,  0x00,  0x07,  0xB0,  0x40,  0x62,  0x05,  0xB0,  0x40,  0x7F,  0x21,  0x90,  0x34,  0x29,  0x3F,
 0x80,  0x34,  0x00,  0x40,  0x90,  0x34,  0x35,  0x33,  0x80,  0x34,  0x00,  0x45,  0x80,  0x25,  0x00,  0x0B,  0x80,  0x2C,  0x00,  0x04,  0x90,  0x34,  0x34,  0x16,  0x80,  0x34,  0x00,  0x7F,  0x90,  0x36,  0x2C,  0x07,  0xB0,  0x40,  0x54,  0x03,  0x90,  0x2D,  0x17,  0x00,  0x90,  0x26,  0x1C,  0x03,  0xB0,  0x40,  0x1A,  0x04,  0xB0,  0x40,  0x00,  0x82,  0x4B,  0xB0,  0x40,  0x20,  0x04,  0xB0,  0x40,  0x72,  0x05,  0xB0,  0x40,  0x7F,  0x84,
 0x48,  0x80,  0x36,  0x00,  0x04,  0x80,  0x26,  0x00,  0x01,  0x80,  0x2D,  0x00,  0x83,  0x2D,  0x90,  0x31,  0x19,  0x08,  0x90,  0x39,  0x36,  0x02,  0x90,  0x2A,  0x20,  0x0C,  0xB0,  0x40,  0x56,  0x05,  0xB0,  0x40,  0x1A,  0x05,  0xB0,  0x40,  0x00,  0x43,  0xB0,  0x40,  0x5C,  0x05,  0xB0,  0x40,  0x7F,  0x02,  0x80,  0x39,  0x00,  0x0C,  0x90,  0x39,  0x2A,  0x58,  0x80,  0x39,  0x00,  0x17,  0x90,  0x39,  0x30,  0x38,  0x80,  0x39,  0x00,
 0x1A,  0x80,  0x2A,  0x00,  0x06,  0x80,  0x31,  0x00,  0x27,  0x90,  0x39,  0x30,  0x46,  0x80,  0x39,  0x00,  0x2F,  0x90,  0x34,  0x24,  0x06,  0x90,  0x2D,  0x2C,  0x03,  0xB0,  0x40,  0x7A,  0x06,  0x90,  0x3D,  0x3A,  0x00,  0xB0,  0x40,  0x2C,  0x04,  0xB0,  0x40,  0x04,  0x05,  0xB0,  0x40,  0x00,  0x3F,  0xB0,  0x40,  0x5C,  0x04,  0xB0,  0x40,  0x7F,  0x0A,  0xB0,  0x40,  0x4E,  0x05,  0xB0,  0x40,  0x1E,  0x05,  0xB0,  0x40,  0x16,  0x04,
 0xB0,  0x40,  0x6E,  0x05,  0xB0,  0x40,  0x7F,  0x33,  0x80,  0x3D,  0x00,  0x4B,  0x90,  0x3D,  0x3C,  0x13,  0x80,  0x2D,  0x00,  0x21,  0x80,  0x34,  0x00,  0x0E,  0x80,  0x3D,  0x00,  0x81,  0x32,  0x90,  0x2A,  0x1D,  0x01,  0x90,  0x39,  0x31,  0x04,  0xB0,  0x40,  0x54,  0x04,  0x90,  0x31,  0x1C,  0x01,  0xB0,  0x40,  0x18,  0x04,  0xB0,  0x40,  0x00,  0x3F,  0xB0,  0x40,  0x42,  0x05,  0xB0,  0x40,  0x7F,  0x09,  0xB0,  0x40,  0x48,  0x05,
 0xB0,  0x40,  0x1A,  0x05,  0xB0,  0x40,  0x14,  0x05,  0xB0,  0x40,  0x6C,  0x03,  0x80,  0x39,  0x00,  0x01,  0xB0,  0x40,  0x7F,  0x09,  0x90,  0x39,  0x31,  0x5A,  0x80,  0x39,  0x00,  0x18,  0x90,  0x39,  0x35,  0x43,  0x80,  0x31,  0x00,  0x17,  0x80,  0x39,  0x00,  0x18,  0x80,  0x2A,  0x00,  0x02,  0x90,  0x39,  0x3B,  0x49,  0x80,  0x39,  0x00,  0x27,  0x90,  0x3D,  0x42,  0x04,  0x90,  0x2D,  0x2E,  0x02,  0x90,  0x34,  0x22,  0x03,  0xB0,
 0x40,  0x58,  0x06,  0xB0,  0x40,  0x1C,  0x04,  0xB0,  0x40,  0x0A,  0x05,  0xB0,  0x40,  0x00,  0x35,  0xB0,  0x40,  0x06,  0x04,  0xB0,  0x40,  0x66,  0x05,  0xB0,  0x40,  0x7F,  0x05,  0xB0,  0x40,  0x46,  0x05,  0xB0,  0x40,  0x1A,  0x05,  0xB0,  0x40,  0x16,  0x04,  0xB0,  0x40,  0x6E,  0x05,  0xB0,  0x40,  0x7F,  0x37,  0x80,  0x3D,  0x00,  0x2D,  0x80,  0x34,  0x00,  0x0D,  0x80,  0x2D,  0x00,  0x0E,  0x90,  0x3D,  0x4A,  0x42,  0x80,  0x3D,
 0x00,  0x81,  0x2F,  0x90,  0x25,  0x21,  0x02,  0x90,  0x34,  0x37,  0x02,  0x90,  0x2C,  0x24,  0x11,  0xB0,  0x40,  0x54,  0x05,  0xB0,  0x40,  0x1A,  0x05,  0xB0,  0x40,  0x00,  0x3A,  0xB0,  0x40,  0x06,  0x00,  0x80,  0x34,  0x00,  0x04,  0xB0,  0x40,  0x00,  0x14,  0xB0,  0x40,  0x5A,  0x05,  0xB0,  0x40,  0x7F,  0x0B,  0x90,  0x34,  0x29,  0x41,  0x80,  0x34,  0x00,  0x48,  0x90,  0x34,  0x34,  0x43,  0x80,  0x2C,  0x00,  0x00,  0x80,  0x25,
 0x00,  0x81,  0x35,  0x80,  0x34,  0x00,  0x0C,  0xB0,  0x40,  0x58,  0x01,  0x90,  0x26,  0x1C,  0x03,  0x90,  0x2D,  0x16,  0x01,  0xB0,  0x40,  0x1A,  0x05,  0xB0,  0x40,  0x00,  0x09,  0x90,  0x36,  0x24,  0x6E,  0xB0,  0x40,  0x14,  0x05,  0xB0,  0x40,  0x6C,  0x05,  0xB0,  0x40,  0x7F,  0x81,  0x03,  0x80,  0x26,  0x00,  0x06,  0x80,  0x2D,  0x00,  0x69,  0x80,  0x36,  0x00,  0x81,  0x33,  0x90,  0x39,  0x1C,  0x02,  0x90,  0x31,  0x17,  0x01,
 0x90,  0x2A,  0x16,  0x03,  0xB0,  0x40,  0x5A,  0x05,  0xB0,  0x40,  0x1E,  0x05,  0xB0,  0x40,  0x00,  0x7C,  0xB0,  0x40,  0x10,  0x06,  0xB0,  0x40,  0x6A,  0x04,  0xB0,  0x40,  0x7F,  0x1A,  0xB0,  0x40,  0x68,  0x03,  0xB0,  0x40,  0x32,  0x05,  0xB0,  0x40,  0x7C,  0x04,  0xB0,  0x40,  0x7F,  0x84,  0x42,  0x80,  0x31,  0x00,  0x08,  0x80,  0x2A,  0x00,  0x02,  0x80,  0x39,  0x00,  0x54,  0x90,  0x31,  0x20,  0x03,  0x90,  0x3D,  0x30,  0x01,
 0x90,  0x2A,  0x1E,  0x06,  0x90,  0x39,  0x30,  0x0C,  0xB0,  0x40,  0x3E,  0x05,  0xB0,  0x40,  0x0E,  0x04,  0xB0,  0x40,  0x00,  0x3A,  0x80,  0x3D,  0x00,  0x0A,  0xB0,  0x40,  0x5E,  0x05,  0xB0,  0x40,  0x7F,  0x04,  0xB0,  0x40,  0x46,  0x05,  0xB0,  0x40,  0x1A,  0x05,  0xB0,  0x40,  0x04,  0x05,  0xB0,  0x40,  0x64,  0x05,  0xB0,  0x40,  0x7F,  0x05,  0x90,  0x3D,  0x2B,  0x45,  0x80,  0x3D,  0x00,  0x2D,  0x90,  0x3D,  0x38,  0x45,  0x80,
 0x3D,  0x00,  0x17,  0x80,  0x31,  0x00,  0x01,  0x80,  0x2A,  0x00,  0x21,  0x90,  0x3D,  0x36,  0x2F,  0x80,  0x39,  0x00,  0x0A,  0x80,  0x3D,  0x00,  0x36,  0x90,  0x34,  0x34,  0x10,  0x90,  0x40,  0x3E,  0x00,  0x90,  0x2D,  0x34,  0x02,  0xB0,  0x40,  0x6C,  0x05,  0xB0,  0x40,  0x26,  0x04,  0x90,  0x3D,  0x32,  0x00,  0xB0,  0x40,  0x02,  0x05,  0xB0,  0x40,  0x00,  0x5C,  0xB0,  0x40,  0x0E,  0x04,  0xB0,  0x40,  0x6A,  0x06,  0xB0,  0x40,
 0x7F,  0x12,  0xB0,  0x40,  0x66,  0x05,  0xB0,  0x40,  0x7F,  0x56,  0x90,  0x42,  0x40,  0x03,  0x80,  0x40,  0x00,  0x02,  0x80,  0x34,  0x00,  0x02,  0x80,  0x2D,  0x00,  0x28,  0x80,  0x3D,  0x00,  0x21,  0x80,  0x42,  0x00,  0x81,  0x11,  0x90,  0x31,  0x24,  0x05,  0x90,  0x3D,  0x3E,  0x03,  0x90,  0x2A,  0x20,  0x02,  0x90,  0x39,  0x36,  0x06,  0xB0,  0x40,  0x52,  0x04,  0xB0,  0x40,  0x18,  0x05,  0xB0,  0x40,  0x00,  0x40,  0x80,  0x3D,
 0x00,  0x09,  0xB0,  0x40,  0x22,  0x06,  0xB0,  0x40,  0x74,  0x03,  0xB0,  0x40,  0x7F,  0x17,  0x90,  0x3D,  0x3A,  0x42,  0x80,  0x3D,  0x00,  0x34,  0x90,  0x3D,  0x36,  0x3F,  0x80,  0x3D,  0x00,  0x05,  0x80,  0x2A,  0x00,  0x0E,  0x80,  0x31,  0x00,  0x2B,  0x90,  0x3D,  0x38,  0x27,  0x80,  0x39,  0x00,  0x14,  0x80,  0x3D,  0x00,  0x3B,  0x90,  0x2D,  0x21,  0x00,  0xB0,  0x40,  0x52,  0x05,  0x90,  0x34,  0x26,  0x00,  0xB0,  0x40,  0x18,
 0x05,  0xB0,  0x40,  0x00,  0x01,  0x90,  0x3D,  0x34,  0x00,  0x90,  0x40,  0x3E,  0x50,  0xB0,  0x40,  0x34,  0x05,  0xB0,  0x40,  0x7C,  0x05,  0xB0,  0x40,  0x7F,  0x0E,  0xB0,  0x40,  0x58,  0x05,  0xB0,  0x40,  0x7F,  0x60,  0x90,  0x42,  0x40,  0x18,  0x80,  0x2D,  0x00,  0x05,  0x80,  0x34,  0x00,  0x0B,  0x80,  0x40,  0x00,  0x05,  0x80,  0x42,  0x00,  0x0D,  0x80,  0x3D,  0x00,  0x81,  0x49,  0x90,  0x34,  0x23,  0x02,  0x90,  0x25,  0x23,
 0x03,  0xB0,  0x40,  0x52,  0x03,  0x90,  0x2C,  0x22,  0x02,  0xB0,  0x40,  0x1A,  0x05,  0xB0,  0x40,  0x00,  0x0B,  0xB0,  0x40,  0x7F,  0x42,  0xB0,  0x40,  0x4A,  0x82,  0x02,  0x80,  0x34,  0x00,  0x18,  0x80,  0x25,  0x00,  0x02,  0x80,  0x2C,  0x00,  0x82,  0x38,  0x90,  0x26,  0x1F,  0x03,  0xB0,  0x40,  0x54,  0x00,  0x90,  0x2D,  0x17,  0x02,  0x90,  0x36,  0x25,  0x05,  0xB0,  0x40,  0x1A,  0x02,  0xB0,  0x40,  0x00,  0x03,  0x90,  0x3D,
 0x38,  0x81,  0x6D,  0x80,  0x3D,  0x00,  0x04,  0x90,  0x39,  0x32,  0x81,  0x02,  0xB0,  0x40,  0x2A,  0x05,  0xB0,  0x40,  0x26,  0x05,  0xB0,  0x40,  0x41,  0x05,  0xB0,  0x40,  0x7F,  0x83,  0x3F,  0x80,  0x26,  0x00,  0x04,  0x80,  0x2D,  0x00,  0x0C,  0x80,  0x36,  0x00,  0x25,  0x80,  0x39,  0x00,  0x82,  0x2A,  0xB0,  0x40,  0x76,  0x05,  0xB0,  0x40,  0x2A,  0x06,  0xB0,  0x40,  0x04,  0x04,  0xB0,  0x40,  0x00,  0x12,  0x90,  0x2A,  0x17,
 0x02,  0x90,  0x31,  0x20,  0x00,  0x90,  0x3D,  0x22,  0x06,  0x90,  0x39,  0x2A,  0x4B,  0x80,  0x3D,  0x00,  0x12,  0xB0,  0x40,  0x3C,  0x05,  0xB0,  0x40,  0x7F,  0x1A,  0x90,  0x3D,  0x24,  0x35,  0x80,  0x3D,  0x00,  0x52,  0x90,  0x3D,  0x2C,  0x34,  0x80,  0x3D,  0x00,  0x10,  0x80,  0x2A,  0x00,  0x01,  0x80,  0x31,  0x00,  0x32,  0x90,  0x3D,  0x2A,  0x39,  0x80,  0x3D,  0x00,  0x03,  0x80,  0x39,  0x00,  0x4C,  0x90,  0x40,  0x38,  0x03,
 0x90,  0x34,  0x1F,  0x02,  0x90,  0x2D,  0x28,  0x01,  0x90,  0x3D,  0x3C,  0x05,  0xB0,  0x40,  0x52,  0x05,  0xB0,  0x40,  0x18,  0x05,  0xB0,  0x40,  0x00,  0x53,  0xB0,  0x40,  0x52,  0x04,  0xB0,  0x40,  0x7F,  0x17,  0xB0,  0x40,  0x5C,  0x05,  0xB0,  0x40,  0x7F,  0x50,  0x80,  0x40,  0x00,  0x05,  0x80,  0x2D,  0x00,  0x09,  0x90,  0x42,  0x3C,  0x06,  0x80,  0x34,  0x00,  0x24,  0x80,  0x3D,  0x00,  0x24,  0x80,  0x42,  0x00,  0x81,  0x1F,
 0x90,  0x2A,  0x20,  0x04,  0x90,  0x3D,  0x38,  0x00,  0x90,  0x31,  0x1E,  0x03,  0x90,  0x39,  0x2E,  0x0E,  0xB0,  0x40,  0x54,  0x04,  0xB0,  0x40,  0x1A,  0x04,  0xB0,  0x40,  0x00,  0x30,  0x80,  0x3D,  0x00,  0x10,  0xB0,  0x40,  0x10,  0x03,  0xB0,  0x40,  0x6A,  0x05,  0xB0,  0x40,  0x7F,  0x05,  0xB0,  0x40,  0x4E,  0x05,  0xB0,  0x40,  0x1E,  0x06,  0xB0,  0x40,  0x0C,  0x04,  0xB0,  0x40,  0x68,  0x01,  0x90,  0x3D,  0x38,  0x04,  0xB0,
 0x40,  0x7F,  0x3C,  0x80,  0x3D,  0x00,  0x2C,  0x90,  0x3D,  0x32,  0x10,  0x80,  0x2A,  0x00,  0x15,  0x80,  0x31,  0x00,  0x1A,  0x80,  0x3D,  0x00,  0x35,  0x90,  0x3D,  0x36,  0x14,  0x80,  0x39,  0x00,  0x3C,  0x80,  0x3D,  0x00,  0x33,  0xB0,  0x40,  0x52,  0x05,  0xB0,  0x40,  0x18,  0x03,  0x90,  0x2D,  0x26,  0x00,  0x90,  0x3D,  0x2C,  0x01,  0xB0,  0x40,  0x00,  0x03,  0x90,  0x40,  0x40,  0x10,  0x90,  0x34,  0x10,  0x2C,  0xB0,  0x40,
 0x20,  0x04,  0xB0,  0x40,  0x72,  0x05,  0xB0,  0x40,  0x7F,  0x05,  0xB0,  0x40,  0x52,  0x04,  0xB0,  0x40,  0x22,  0x06,  0xB0,  0x40,  0x72,  0x04,  0xB0,  0x40,  0x7F,  0x65,  0x80,  0x34,  0x00,  0x1C,  0x80,  0x2D,  0x00,  0x04,  0x80,  0x40,  0x00,  0x02,  0x90,  0x42,  0x38,  0x1A,  0x80,  0x42,  0x00,  0x03,  0x80,  0x3D,  0x00,  0x81,  0x47,  0x90,  0x34,  0x24,  0x03,  0x90,  0x25,  0x1B,  0x03,  0x90,  0x2C,  0x20,  0x02,  0xB0,  0x40,
 0x52,  0x04,  0xB0,  0x40,  0x18,  0x05,  0xB0,  0x40,  0x00,  0x10,  0x90,  0x3B,  0x32,  0x25,  0xB0,  0x40,  0x10,  0x05,  0xB0,  0x40,  0x6A,  0x04,  0xB0,  0x40,  0x7F,  0x04,  0xB0,  0x40,  0x50,  0x06,  0xB0,  0x40,  0x20,  0x04,  0xB0,  0x40,  0x1A,  0x05,  0xB0,  0x40,  0x70,  0x00,  0x80,  0x3B,  0x00,  0x04,  0xB0,  0x40,  0x7F,  0x2B,  0x90,  0x3B,  0x38,  0x43,  0x80,  0x3B,  0x00,  0x2E,  0x90,  0x3B,  0x3E,  0x41,  0x80,  0x3B,  0x00,
 0x1C,  0x80,  0x25,  0x00,  0x0C,  0x80,  0x2C,  0x00,  0x00,  0x80,  0x34,  0x00,  0x14,  0x90,  0x3B,  0x2C,  0x3E,  0x80,  0x3B,  0x00,  0x42,  0xB0,  0x40,  0x76,  0x05,  0xB0,  0x40,  0x2A,  0x01,  0x90,  0x36,  0x32,  0x02,  0x90,  0x26,  0x1E,  0x02,  0x90,  0x2D,  0x17,  0x02,  0x90,  0x3D,  0x2E,  0x00,  0xB0,  0x40,  0x04,  0x03,  0xB0,  0x40,  0x00,  0x73,  0xB0,  0x40,  0x5E,  0x04,  0xB0,  0x40,  0x7F,  0x19,  0xB0,  0x40,  0x64,  0x04,
 0xB0,  0x40,  0x7F,  0x53,  0x90,  0x39,  0x3A,  0x17,  0x80,  0x3D,  0x00,  0x05,  0x80,  0x2D,  0x00,  0x05,  0x80,  0x26,  0x00,  0x3F,  0x80,  0x39,  0x00,  0x18,  0x80,  0x36,  0x00,  0x81,  0x00,  0xB0,  0x40,  0x5C,  0x01,  0x90,  0x2A,  0x1C,  0x01,  0x90,  0x39,  0x2C,  0x01,  0x90,  0x31,  0x20,  0x03,  0xB0,  0x40,  0x1E,  0x04,  0xB0,  0x40,  0x00,  0x6D,  0x90,  0x36,  0x1C,  0x0B,  0xB0,  0x40,  0x1C,  0x05,  0xB0,  0x40,  0x70,  0x05,
 0xB0,  0x40,  0x7F,  0x21,  0xB0,  0x40,  0x62,  0x05,  0xB0,  0x40,  0x2C,  0x05,  0xB0,  0x40,  0x0E,  0x05,  0xB0,  0x40,  0x12,  0x04,  0xB0,  0x40,  0x6C,  0x05,  0xB0,  0x40,  0x7F,  0x82,  0x0B,  0x80,  0x36,  0x00,  0x41,  0x80,  0x39,  0x00,  0x17,  0x80,  0x31,  0x00,  0x1B,  0x80,  0x2A,  0x00,  0x81,  0x10,  0x90,  0x3D,  0x2D,  0x81,  0x03,  0x90,  0x31,  0x14,  0x24,  0x80,  0x3D,  0x00,  0x50,  0x90,  0x34,  0x29,  0x01,  0x90,  0x25,
 0x1F,  0x03,  0x90,  0x2C,  0x1D,  0x10,  0xB0,  0x40,  0x44,  0x04,  0xB0,  0x40,  0x12,  0x05,  0xB0,  0x40,  0x00,  0x16,  0x80,  0x31,  0x00,  0x1F,  0xB0,  0x40,  0x0A,  0x05,  0xB0,  0x40,  0x68,  0x05,  0xB0,  0x40,  0x40,  0x05,  0xB0,  0x40,  0x16,  0x05,  0xB0,  0x40,  0x01,  0x05,  0xB0,  0x40,  0x62,  0x04,  0xB0,  0x40,  0x7F,  0x03,  0x80,  0x34,  0x00,  0x0C,  0x90,  0x34,  0x2A,  0x66,  0x80,  0x34,  0x00,  0x13,  0x90,  0x34,  0x31,
 0x41,  0x80,  0x25,  0x00,  0x06,  0x80,  0x2C,  0x00,  0x81,  0x1A,  0x80,  0x34,  0x00,  0x3D,  0x90,  0x26,  0x1B,  0x01,  0xB0,  0x40,  0x54,  0x04,  0xB0,  0x40,  0x18,  0x02,  0x90,  0x2D,  0x14,  0x03,  0x90,  0x36,  0x2A,  0x01,  0xB0,  0x40,  0x00,  0x81,  0x5C,  0xB0,  0x40,  0x12,  0x05,  0xB0,  0x40,  0x64,  0x05,  0xB0,  0x40,  0x7F,  0x82,  0x25,  0xB0,  0x40,  0x58,  0x04,  0xB0,  0x40,  0x26,  0x05,  0xB0,  0x40,  0x70,  0x06,  0xB0,
 0x40,  0x7F,  0x82,  0x53,  0x80,  0x26,  0x00,  0x0D,  0x80,  0x2D,  0x00,  0x34,  0x80,  0x36,  0x00,  0x81,  0x7C,  0x90,  0x3B,  0x34,  0x81,  0x10,  0x90,  0x3D,  0x2E,  0x00,  0x90,  0x34,  0x20,  0x01,  0x90,  0x39,  0x1F,  0x07,  0xB0,  0x40,  0x4E,  0x04,  0x90,  0x28,  0x21,  0x01,  0xB0,  0x40,  0x18,  0x05,  0xB0,  0x40,  0x00,  0x0A,  0x80,  0x3B,  0x00,  0x52,  0xB0,  0x40,  0x26,  0x04,  0xB0,  0x40,  0x76,  0x06,  0xB0,  0x40,  0x7F,
 0x81,  0x5C,  0x80,  0x3D,  0x00,  0x00,  0x80,  0x34,  0x00,  0x0B,  0x90,  0x3B,  0x32,  0x2D,  0x80,  0x39,  0x00,  0x65,  0x90,  0x39,  0x2C,  0x03,  0x90,  0x3D,  0x39,  0x0E,  0x90,  0x34,  0x19,  0x05,  0x80,  0x3B,  0x00,  0x82,  0x41,  0x80,  0x3D,  0x00,  0x07,  0x80,  0x34,  0x00,  0x02,  0x80,  0x39,  0x00,  0x13,  0x90,  0x3B,  0x36,  0x81,  0x0C,  0x90,  0x39,  0x2E,  0x03,  0x90,  0x3D,  0x37,  0x09,  0x90,  0x34,  0x1B,  0x16,  0x80,
 0x3B,  0x00,  0x82,  0x29,  0x90,  0x40,  0x3E,  0x20,  0x80,  0x34,  0x00,  0x35,  0x80,  0x39,  0x00,  0x11,  0x80,  0x3D,  0x00,  0x1F,  0x90,  0x34,  0x21,  0x0C,  0x90,  0x39,  0x32,  0x06,  0x90,  0x3D,  0x36,  0x12,  0x80,  0x40,  0x00,  0x74,  0x90,  0x3B,  0x24,  0x81,  0x35,  0x80,  0x3D,  0x00,  0x0A,  0x80,  0x34,  0x00,  0x13,  0x80,  0x39,  0x00,  0x06,  0x80,  0x3B,  0x00,  0x1C,  0x80,  0x28,  0x00,  0x1C,  0x90,  0x3D,  0x36,  0x02,
 0x90,  0x39,  0x36,  0x03,  0x90,  0x28,  0x24,  0x07,  0x90,  0x34,  0x18,  0x05,  0xB0,  0x40,  0x5E,  0x05,  0xB0,  0x40,  0x1E,  0x06,  0xB0,  0x40,  0x00,  0x50,  0xB0,  0x40,  0x26,  0x05,  0xB0,  0x40,  0x76,  0x05,  0xB0,  0x40,  0x7F,  0x81,  0x72,  0x80,  0x34,  0x00,  0x00,  0x80,  0x39,  0x00,  0x0C,  0x90,  0x3B,  0x36,  0x15,  0x80,  0x3D,  0x00,  0x7D,  0x90,  0x39,  0x36,  0x04,  0x90,  0x34,  0x26,  0x01,  0x90,  0x3D,  0x3E,  0x20,
 0x80,  0x3B,  0x00,  0x81,  0x74,  0x90,  0x3B,  0x36,  0x35,  0x80,  0x39,  0x00,  0x1C,  0x80,  0x34,  0x00,  0x38,  0x80,  0x3D,  0x00,  0x03,  0x90,  0x34,  0x20,  0x00,  0x90,  0x39,  0x26,  0x15,  0x90,  0x3D,  0x38,  0x18,  0x80,  0x3B,  0x00,  0x81,  0x31,  0x80,  0x34,  0x00,  0x38,  0x80,  0x3D,  0x00,  0x05,  0x80,  0x39,  0x00,  0x3F,  0x80,  0x28,  0x00,  0x1C,  0x90,  0x34,  0x3A,  0x70,  0x90,  0x28,  0x2E,  0x06,  0x90,  0x39,  0x21,
 0x02,  0x90,  0x3B,  0x3E,  0x01,  0x80,  0x34,  0x00,  0x1E,  0x90,  0x3D,  0x3C,  0x00,  0x80,  0x39,  0x00,  0x14,  0x80,  0x3B,  0x00,  0x47,  0x90,  0x3B,  0x3C,  0x35,  0x80,  0x3D,  0x00,  0x37,  0x80,  0x28,  0x00,  0x02,  0x80,  0x3B,  0x00,  0x05,  0x90,  0x39,  0x44,  0x30,  0x80,  0x39,  0x00,  0x46,  0x90,  0x38,  0x38,  0x54,  0x80,  0x38,  0x00,  0x30,  0xB0,  0x40,  0x78,  0x00,  0x90,  0x36,  0x30,  0x01,  0x90,  0x32,  0x3C,  0x03,
 0xB0,  0x40,  0x2C,  0x01,  0x90,  0x26,  0x22,  0x03,  0xB0,  0x40,  0x06,  0x05,  0xB0,  0x40,  0x00,  0x51,  0xB0,  0x40,  0x2A,  0x05,  0xB0,  0x40,  0x78,  0x05,  0xB0,  0x40,  0x7F,  0x5B,  0x80,  0x32,  0x00,  0x18,  0x80,  0x36,  0x00,  0x82,  0x29,  0x90,  0x36,  0x28,  0x02,  0x90,  0x39,  0x3A,  0x02,  0x90,  0x32,  0x24,  0x05,  0x90,  0x3D,  0x36,  0x51,  0x80,  0x3D,  0x00,  0x2E,  0x90,  0x3D,  0x40,  0x3C,  0x80,  0x3D,  0x00,  0x35,
 0x90,  0x3D,  0x36,  0x45,  0x80,  0x3D,  0x00,  0x2E,  0x90,  0x3D,  0x38,  0x0B,  0x80,  0x32,  0x00,  0x05,  0x80,  0x36,  0x00,  0x1A,  0x80,  0x39,  0x00,  0x14,  0x80,  0x3D,  0x00,  0x4C,  0x90,  0x32,  0x26,  0x03,  0x90,  0x36,  0x20,  0x01,  0x90,  0x3D,  0x30,  0x06,  0x90,  0x39,  0x32,  0x81,  0x7C,  0x90,  0x3B,  0x3A,  0x0B,  0x80,  0x3D,  0x00,  0x18,  0x80,  0x36,  0x00,  0x01,  0x80,  0x39,  0x00,  0x1C,  0x80,  0x32,  0x00,  0x81,
 0x26,  0x90,  0x32,  0x32,  0x02,  0x90,  0x36,  0x34,  0x02,  0x90,  0x39,  0x3E,  0x00,  0x80,  0x3B,  0x00,  0x81,  0x09,  0x80,  0x36,  0x00,  0x11,  0x80,  0x32,  0x00,  0x2C,  0x80,  0x26,  0x00,  0x32,  0x90,  0x3B,  0x42,  0x6A,  0x90,  0x3D,  0x3C,  0x34,  0x90,  0x21,  0x1F,  0x00,  0x90,  0x34,  0x1E,  0x09,  0x90,  0x2D,  0x14,  0x01,  0xB0,  0x40,  0x58,  0x05,  0xB0,  0x40,  0x1C,  0x04,  0xB0,  0x40,  0x00,  0x78,  0xB0,  0x40,  0x3A,
 0x05,  0xB0,  0x40,  0x7F,  0x19,  0xB0,  0x40,  0x60,  0x05,  0xB0,  0x40,  0x7F,  0x52,  0x80,  0x21,  0x00,  0x23,  0x80,  0x34,  0x00,  0x04,  0x80,  0x2D,  0x00,  0x81,  0x00,  0x80,  0x39,  0x00,  0x0E,  0x80,  0x3B,  0x00,  0x0E,  0x80,  0x3D,  0x00,  0x46,  0x90,  0x2D,  0x20,  0x02,  0x90,  0x39,  0x2C,  0x03,  0x90,  0x3D,  0x40,  0x02,  0x90,  0x3B,  0x28,  0x0C,  0x90,  0x34,  0x1C,  0x42,  0x80,  0x3D,  0x00,  0x26,  0x90,  0x3D,  0x3F,
 0x4C,  0x80,  0x3D,  0x00,  0x28,  0x90,  0x3D,  0x38,  0x46,  0x80,  0x39,  0x00,  0x10,  0x80,  0x3D,  0x00,  0x0B,  0x80,  0x2D,  0x00,  0x1C,  0x90,  0x3D,  0x34,  0x11,  0x80,  0x3B,  0x00,  0x13,  0x80,  0x34,  0x00,  0x0F,  0x80,  0x3D,  0x00,  0x50,  0x90,  0x39,  0x2C,  0x00,  0x90,  0x3D,  0x3C,  0x02,  0x90,  0x3B,  0x30,  0x02,  0x90,  0x34,  0x1E,  0x03,  0x90,  0x2D,  0x1A,  0x6B,  0x80,  0x3B,  0x00,  0x46,  0x80,  0x39,  0x00,  0x34,
 0x90,  0x3B,  0x3C,  0x26,  0x80,  0x34,  0x00,  0x06,  0x80,  0x2D,  0x00,  0x35,  0x80,  0x3D,  0x00,  0x81,  0x17,  0x90,  0x2B,  0x2A,  0x05,  0x90,  0x32,  0x1F,  0x02,  0x80,  0x3B,  0x00,  0x06,  0x90,  0x39,  0x39,  0x09,  0xB0,  0x40,  0x54,  0x05,  0xB0,  0x40,  0x1A,  0x05,  0xB0,  0x40,  0x00,  0x60,  0xB0,  0x40,  0x01,  0x05,  0xB0,  0x40,  0x62,  0x04,  0xB0,  0x40,  0x7F,  0x14,  0xB0,  0x40,  0x58,  0x05,  0xB0,  0x40,  0x7F,  0x4B,
 0x90,  0x3B,  0x3E,  0x03,  0x80,  0x2B,  0x00,  0x05,  0x80,  0x39,  0x00,  0x22,  0x80,  0x32,  0x00,  0x57,  0x90,  0x39,  0x3D,  0x02,  0x80,  0x3B,  0x00,  0x34,  0x90,  0x36,  0x2E,  0x07,  0x90,  0x31,  0x2A,  0x09,  0x90,  0x2A,  0x20,  0x14,  0xB0,  0x40,  0x6C,  0x05,  0xB0,  0x40,  0x26,  0x05,  0xB0,  0x40,  0x02,  0x05,  0xB0,  0x40,  0x00,  0x6A,  0xB0,  0x40,  0x58,  0x04,  0xB0,  0x40,  0x7F,  0x13,  0xB0,  0x40,  0x60,  0x05,  0xB0,
 0x40,  0x7F,  0x4C,  0x80,  0x36,  0x00,  0x1B,  0x80,  0x2A,  0x00,  0x0C,  0x80,  0x31,  0x00,  0x81,  0x4C,  0x90,  0x36,  0x26,  0x01,  0x90,  0x3B,  0x3E,  0x01,  0x90,  0x31,  0x22,  0x04,  0x90,  0x39,  0x3A,  0x06,  0x90,  0x2A,  0x17,  0x21,  0x80,  0x39,  0x00,  0x55,  0x90,  0x3D,  0x3A,  0x2D,  0x80,  0x3B,  0x00,  0x17,  0x80,  0x3D,  0x00,  0x36,  0x90,  0x3D,  0x30,  0x24,  0x80,  0x39,  0x00,  0x19,  0x80,  0x3D,  0x00,  0x0C,  0x80,
 0x36,  0x00,  0x11,  0x80,  0x31,  0x00,  0x0C,  0x80,  0x2A,  0x00,  0x06,  0x90,  0x3D,  0x36,  0x3E,  0x80,  0x3D,  0x00,  0x52,  0x90,  0x31,  0x17,  0x00,  0x90,  0x39,  0x2A,  0x00,  0x90,  0x2A,  0x1D,  0x03,  0x90,  0x3D,  0x38,  0x01,  0x90,  0x36,  0x2A,  0x81,  0x60,  0x80,  0x3D,  0x00,  0x0A,  0x90,  0x3B,  0x40,  0x26,  0x80,  0x36,  0x00,  0x06,  0x80,  0x39,  0x00,  0x02,  0x80,  0x2A,  0x00,  0x10,  0x80,  0x31,  0x00,  0x5D,  0x80,
 0x3B,  0x00,  0x5A,  0x90,  0x31,  0x1F,  0x09,  0x90,  0x3D,  0x38,  0x03,  0x90,  0x2A,  0x20,  0x03,  0x90,  0x36,  0x2C,  0x00,  0x90,  0x39,  0x3A,  0x81,  0x2E,  0x90,  0x40,  0x3E,  0x07,  0x80,  0x31,  0x00,  0x1D,  0x80,  0x36,  0x00,  0x00,  0x80,  0x2A,  0x00,  0x0E,  0x80,  0x40,  0x00,  0x0A,  0x80,  0x39,  0x00,  0x00,  0x80,  0x3D,  0x00,  0x81,  0x28,  0x90,  0x39,  0x2A,  0x02,  0xB0,  0x40,  0x52,  0x01,  0x90,  0x3E,  0x36,  0x03,
 0xB0,  0x40,  0x1A,  0x06,  0x90,  0x32,  0x24,  0x00,  0xB0,  0x40,  0x02,  0x03,  0x90,  0x42,  0x3C,  0x01,  0xB0,  0x40,  0x00,  0x00,  0x90,  0x26,  0x1A,  0x74,  0xB0,  0x40,  0x5A,  0x04,  0xB0,  0x40,  0x7F,  0x1D,  0xB0,  0x40,  0x5C,  0x04,  0xB0,  0x40,  0x7F,  0x81,  0x01,  0x80,  0x42,  0x00,  0x06,  0x80,  0x3E,  0x00,  0x02,  0x80,  0x39,  0x00,  0x1C,  0x80,  0x32,  0x00,  0x81,  0x44,  0x90,  0x32,  0x24,  0x03,  0x90,  0x3E,  0x34,
 0x01,  0x90,  0x42,  0x34,  0x04,  0x90,  0x39,  0x3A,  0x82,  0x2B,  0x80,  0x3E,  0x00,  0x01,  0x80,  0x39,  0x00,  0x00,  0x80,  0x42,  0x00,  0x1B,  0x80,  0x32,  0x00,  0x81,  0x2D,  0x90,  0x42,  0x32,  0x00,  0x90,  0x3E,  0x2E,  0x03,  0x90,  0x39,  0x32,  0x03,  0x90,  0x32,  0x1D,  0x81,  0x79,  0x90,  0x45,  0x30,  0x48,  0x80,  0x3E,  0x00,  0x06,  0x80,  0x32,  0x00,  0x81,  0x36,  0x90,  0x32,  0x28,  0x0A,  0x90,  0x3D,  0x3C,  0x09,
 0x80,  0x45,  0x00,  0x15,  0x80,  0x42,  0x00,  0x81,  0x3E,  0x80,  0x3D,  0x00,  0x1F,  0x80,  0x39,  0x00,  0x25,  0x80,  0x32,  0x00,  0x04,  0x80,  0x26,  0x00,  0x4D,  0x90,  0x26,  0x2C,  0x02,  0x90,  0x36,  0x2D,  0x03,  0x90,  0x3D,  0x2D,  0x02,  0x90,  0x39,  0x2F,  0x0B,  0x90,  0x32,  0x2A,  0x09,  0xB0,  0x40,  0x3C,  0x05,  0xB0,  0x40,  0x0E,  0x05,  0xB0,  0x40,  0x00,  0x60,  0xB0,  0x40,  0x60,  0x05,  0xB0,  0x40,  0x7F,  0x5D,
 0x80,  0x3D,  0x00,  0x5D,  0x80,  0x32,  0x00,  0x30,  0x80,  0x39,  0x00,  0x69,  0x90,  0x3D,  0x3F,  0x04,  0x90,  0x39,  0x3C,  0x05,  0x90,  0x32,  0x2A,  0x50,  0x80,  0x3D,  0x00,  0x23,  0x80,  0x36,  0x00,  0x06,  0x90,  0x3D,  0x3F,  0x45,  0x80,  0x3D,  0x00,  0x2E,  0x90,  0x3D,  0x39,  0x47,  0x80,  0x3D,  0x00,  0x2B,  0x90,  0x3D,  0x44,  0x09,  0x80,  0x32,  0x00,  0x17,  0x80,  0x39,  0x00,  0x18,  0x80,  0x3D,  0x00,  0x3E,  0x90,
 0x3D,  0x3E,  0x04,  0x90,  0x39,  0x3E,  0x02,  0x90,  0x32,  0x32,  0x81,  0x70,  0x90,  0x3B,  0x46,  0x3D,  0x80,  0x3D,  0x00,  0x0D,  0x80,  0x39,  0x00,  0x01,  0x80,  0x32,  0x00,  0x81,  0x26,  0x90,  0x32,  0x2E,  0x0B,  0x90,  0x39,  0x46,  0x1C,  0x80,  0x3B,  0x00,  0x81,  0x59,  0x90,  0x3B,  0x40,  0x10,  0x80,  0x26,  0x00,  0x05,  0x80,  0x32,  0x00,  0x62,  0x90,  0x3D,  0x44,  0x79,  0x90,  0x34,  0x3E,  0x14,  0x90,  0x2D,  0x20,
 0x0F,  0xB0,  0x40,  0x52,  0x05,  0xB0,  0x40,  0x18,  0x05,  0xB0,  0x40,  0x00,  0x56,  0xB0,  0x40,  0x5E,  0x05,  0xB0,  0x40,  0x7F,  0x13,  0xB0,  0x40,  0x56,  0x04,  0xB0,  0x40,  0x7F,  0x5F,  0x80,  0x34,  0x00,  0x11,  0x80,  0x2D,  0x00,  0x14,  0x80,  0x39,  0x00,  0x07,  0x80,  0x3B,  0x00,  0x13,  0x80,  0x3D,  0x00,  0x81,  0x28,  0x90,  0x34,  0x3E,  0x00,  0x90,  0x3D,  0x46,  0x03,  0x90,  0x2D,  0x2C,  0x03,  0x90,  0x39,  0x40,
 0x03,  0x90,  0x3B,  0x34,  0x44,  0x80,  0x3D,  0x00,  0x30,  0x90,  0x3D,  0x32,  0x51,  0x80,  0x3D,  0x00,  0x2C,  0x90,  0x3D,  0x3E,  0x07,  0x80,  0x3B,  0x00,  0x46,  0x80,  0x3D,  0x00,  0x23,  0x80,  0x2D,  0x00,  0x07,  0x90,  0x3D,  0x3E,  0x1A,  0x80,  0x39,  0x00,  0x10,  0x80,  0x34,  0x00,  0x17,  0x80,  0x3D,  0x00,  0x4B,  0x90,  0x2D,  0x1F,  0x00,  0x90,  0x34,  0x2A,  0x09,  0x90,  0x3D,  0x3E,  0x01,  0x90,  0x39,  0x3A,  0x02,
 0x90,  0x3B,  0x2C,  0x81,  0x31,  0x80,  0x3B,  0x00,  0x15,  0x80,  0x39,  0x00,  0x18,  0x90,  0x3B,  0x44,  0x1B,  0x80,  0x3D,  0x00,  0x0C,  0x80,  0x2D,  0x00,  0x07,  0x80,  0x34,  0x00,  0x81,  0x3D,  0x80,  0x3B,  0x00,  0x04,  0x90,  0x2B,  0x30,  0x01,  0x90,  0x32,  0x2A,  0x0A,  0x90,  0x39,  0x40,  0x0B,  0xB0,  0x40,  0x52,  0x06,  0xB0,  0x40,  0x1A,  0x03,  0xB0,  0x40,  0x00,  0x60,  0xB0,  0x40,  0x01,  0x05,  0xB0,  0x40,  0x62,
 0x05,  0xB0,  0x40,  0x7F,  0x13,  0xB0,  0x40,  0x58,  0x05,  0xB0,  0x40,  0x7F,  0x4D,  0x80,  0x2B,  0x00,  0x14,  0x90,  0x3B,  0x30,  0x02,  0x80,  0x39,  0x00,  0x2C,  0x80,  0x32,  0x00,  0x1A,  0x80,  0x3B,  0x00,  0x2C,  0x90,  0x39,  0x42,  0x28,  0xB0,  0x40,  0x56,  0x05,  0xB0,  0x40,  0x1A,  0x01,  0x90,  0x36,  0x3E,  0x04,  0xB0,  0x40,  0x00,  0x08,  0x90,  0x2A,  0x2C,  0x04,  0x90,  0x31,  0x2C,  0x6C,  0xB0,  0x40,  0x1E,  0x05,
 0xB0,  0x40,  0x72,  0x05,  0xB0,  0x40,  0x7F,  0x13,  0xB0,  0x40,  0x5A,  0x05,  0xB0,  0x40,  0x7F,  0x53,  0x80,  0x36,  0x00,  0x38,  0x80,  0x2A,  0x00,  0x09,  0x80,  0x31,  0x00,  0x0A,  0x80,  0x39,  0x00,  0x81,  0x28,  0x90,  0x2A,  0x26,  0x04,  0x90,  0x36,  0x36,  0x05,  0x90,  0x3B,  0x42,  0x00,  0x90,  0x31,  0x30,  0x03,  0x90,  0x39,  0x3E,  0x81,  0x00,  0x90,  0x3D,  0x38,  0x16,  0x80,  0x3B,  0x00,  0x2D,  0x80,  0x3D,  0x00,
 0x2D,  0x90,  0x3D,  0x3E,  0x40,  0x80,  0x3D,  0x00,  0x24,  0x80,  0x31,  0x00,  0x12,  0x90,  0x3D,  0x36,  0x06,  0x80,  0x2A,  0x00,  0x0A,  0x80,  0x36,  0x00,  0x1B,  0x80,  0x39,  0x00,  0x19,  0x80,  0x3D,  0x00,  0x46,  0x90,  0x36,  0x30,  0x00,  0x90,  0x2A,  0x19,  0x02,  0x90,  0x3D,  0x3D,  0x04,  0x90,  0x31,  0x32,  0x05,  0x90,  0x39,  0x3C,  0x81,  0x77,  0x90,  0x3B,  0x39,  0x03,  0x80,  0x3D,  0x00,  0x1F,  0x80,  0x39,  0x00,
 0x11,  0x80,  0x31,  0x00,  0x10,  0x80,  0x36,  0x00,  0x54,  0x80,  0x2A,  0x00,  0x0E,  0x80,  0x3B,  0x00,  0x39,  0x90,  0x2A,  0x0E,  0x0A,  0x90,  0x31,  0x24,  0x01,  0x90,  0x36,  0x2E,  0x01,  0x90,  0x3D,  0x40,  0x08,  0x90,  0x39,  0x42,  0x06,  0xB0,  0x40,  0x54,  0x06,  0xB0,  0x40,  0x18,  0x03,  0xB0,  0x40,  0x00,  0x3F,  0xB0,  0x40,  0x5E,  0x06,  0xB0,  0x40,  0x7F,  0x1C,  0x80,  0x36,  0x00,  0x14,  0x80,  0x31,  0x00,  0x10,
 0x80,  0x2A,  0x00,  0x50,  0x90,  0x34,  0x2E,  0x02,  0x90,  0x40,  0x4C,  0x1B,  0x80,  0x34,  0x00,  0x00,  0x80,  0x3D,  0x00,  0x02,  0x80,  0x39,  0x00,  0x27,  0x80,  0x40,  0x00,  0x81,  0x33,  0xB0,  0x40,  0x5E,  0x01,  0x90,  0x39,  0x3C,  0x03,  0x90,  0x3E,  0x34,  0x00,  0x90,  0x32,  0x32,  0x03,  0xB0,  0x40,  0x1E,  0x00,  0x90,  0x42,  0x42,  0x03,  0xB0,  0x40,  0x00,  0x03,  0x90,  0x26,  0x1B,  0x81,  0x13,  0xB0,  0x40,  0x54,
 0x03,  0xB0,  0x40,  0x7F,  0x1C,  0xB0,  0x40,  0x5A,  0x06,  0xB0,  0x40,  0x7F,  0x6D,  0x80,  0x3E,  0x00,  0x09,  0x80,  0x42,  0x00,  0x0F,  0x80,  0x39,  0x00,  0x03,  0x80,  0x32,  0x00,  0x81,  0x38,  0x90,  0x39,  0x40,  0x03,  0x90,  0x32,  0x21,  0x03,  0x90,  0x3E,  0x34,  0x00,  0x90,  0x42,  0x48,  0x81,  0x12,  0x80,  0x39,  0x00,  0x34,  0x80,  0x42,  0x00,  0x50,  0x80,  0x32,  0x00,  0x17,  0x80,  0x3E,  0x00,  0x81,  0x3E,  0x90,
 0x39,  0x1E,  0x0F,  0x90,  0x3E,  0x36,  0x02,  0x90,  0x32,  0x22,  0x0C,  0x90,  0x42,  0x38,  0x81,  0x32,  0x80,  0x3E,  0x00,  0x55,  0x80,  0x32,  0x00,  0x81,  0x0E,  0x80,  0x42,  0x00,  0x04,  0x90,  0x45,  0x41,  0x51,  0x80,  0x39,  0x00,  0x3E,  0x80,  0x45,  0x00,  0x81,  0x30,  0x80,  0x26,  0x00,  0x12,  0x90,  0x32,  0x28,  0x81,  0x74,  0x80,  0x32,  0x00,  0x17,  0x90,  0x3D,  0x39,  0x77,  0x80,  0x3D,  0x00,  0x01,  0x90,  0x3B,
 0x30,  0x81,  0x28,  0xB0,  0x40,  0x5C,  0x05,  0xB0,  0x40,  0x1E,  0x03,  0x90,  0x29,  0x1E,  0x03,  0xB0,  0x40,  0x02,  0x03,  0xB0,  0x40,  0x00,  0x03,  0x90,  0x32,  0x19,  0x06,  0x90,  0x35,  0x1C,  0x85,  0x4C,  0xB0,  0x40,  0x02,  0x05,  0xB0,  0x40,  0x12,  0x05,  0xB0,  0x40,  0x14,  0x06,  0xB0,  0x40,  0x20,  0x03,  0xB0,  0x40,  0x2E,  0x05,  0xB0,  0x40,  0x60,  0x05,  0xB0,  0x40,  0x7F,  0x7D,  0xB0,  0x40,  0x7A,  0x06,  0xB0,
 0x40,  0x41,  0x04,  0xB0,  0x40,  0x22,  0x05,  0xB0,  0x40,  0x08,  0x04,  0xB0,  0x40,  0x00,  0x05,  0xB0,  0x40,  0x60,  0x06,  0xB0,  0x40,  0x7F,  0x61,  0x80,  0x32,  0x00,  0x02,  0x80,  0x35,  0x00,  0x06,  0x80,  0x29,  0x00,  0x81,  0x07,  0x90,  0x39,  0x2A,  0x0C,  0x80,  0x3B,  0x00,  0x19,  0x90,  0x39,  0x32,  0x1D,  0x90,  0x3B,  0x2C,  0x0D,  0x80,  0x39,  0x00,  0x07,  0x80,  0x39,  0x00,  0x07,  0x90,  0x3B,  0x28,  0x30,  0x90,
 0x3D,  0x3C,  0x00,  0x80,  0x3B,  0x00,  0x0C,  0x90,  0x39,  0x32,  0x0A,  0x80,  0x3B,  0x00,  0x7B,  0xB0,  0x40,  0x66,  0x03,  0x90,  0x31,  0x1A,  0x02,  0xB0,  0x40,  0x22,  0x04,  0xB0,  0x40,  0x01,  0x05,  0xB0,  0x40,  0x00,  0x00,  0x90,  0x36,  0x36,  0x05,  0x90,  0x2A,  0x15,  0x0A,  0x90,  0x3D,  0x32,  0x0B,  0x80,  0x3D,  0x00,  0x29,  0xB0,  0x40,  0x06,  0x05,  0xB0,  0x40,  0x66,  0x05,  0xB0,  0x40,  0x7F,  0x81,  0x18,  0x80,
 0x36,  0x00,  0x1D,  0x80,  0x2A,  0x00,  0x02,  0x80,  0x31,  0x00,  0x06,  0x80,  0x39,  0x00,  0x81,  0x2C,  0x80,  0x3D,  0x00,  0x16,  0x90,  0x3D,  0x2A,  0x04,  0x90,  0x40,  0x36,  0x0A,  0x90,  0x2D,  0x20,  0x02,  0x90,  0x39,  0x27,  0x04,  0x90,  0x34,  0x2A,  0x0F,  0xB0,  0x40,  0x64,  0x04,  0xB0,  0x40,  0x22,  0x05,  0xB0,  0x40,  0x01,  0x05,  0xB0,  0x40,  0x00,  0x6F,  0xB0,  0x40,  0x62,  0x04,  0xB0,  0x40,  0x7F,  0x1D,  0xB0,
 0x40,  0x48,  0x07,  0xB0,  0x40,  0x7F,  0x45,  0x80,  0x2D,  0x00,  0x0D,  0x80,  0x3D,  0x00,  0x14,  0x80,  0x34,  0x00,  0x2C,  0x80,  0x39,  0x00,  0x0B,  0x80,  0x40,  0x00,  0x7E,  0x90,  0x31,  0x1F,  0x02,  0x90,  0x36,  0x2A,  0x00,  0x90,  0x2A,  0x1C,  0x06,  0x90,  0x39,  0x3A,  0x02,  0x90,  0x3D,  0x32,  0x0D,  0xB0,  0x40,  0x70,  0x04,  0xB0,  0x40,  0x28,  0x05,  0xB0,  0x40,  0x02,  0x05,  0xB0,  0x40,  0x00,  0x56,  0xB0,  0x40,
 0x4E,  0x05,  0xB0,  0x40,  0x7F,  0x69,  0x80,  0x36,  0x00,  0x0C,  0x80,  0x2A,  0x00,  0x04,  0x80,  0x31,  0x00,  0x12,  0x80,  0x39,  0x00,  0x29,  0x80,  0x3D,  0x00,  0x81,  0x38,  0x90,  0x2D,  0x21,  0x03,  0x90,  0x3D,  0x30,  0x08,  0xB0,  0x40,  0x54,  0x04,  0xB0,  0x40,  0x1A,  0x02,  0x90,  0x34,  0x24,  0x01,  0x90,  0x39,  0x22,  0x02,  0xB0,  0x40,  0x00,  0x01,  0x90,  0x40,  0x36,  0x51,  0xB0,  0x40,  0x18,  0x05,  0xB0,  0x40,
 0x6E,  0x06,  0xB0,  0x40,  0x7F,  0x12,  0xB0,  0x40,  0x56,  0x04,  0xB0,  0x40,  0x7F,  0x3F,  0x80,  0x34,  0x00,  0x0B,  0x80,  0x40,  0x00,  0x0C,  0x80,  0x39,  0x00,  0x0A,  0x80,  0x2D,  0x00,  0x38,  0x80,  0x3D,  0x00,  0x81,  0x4D,  0x90,  0x34,  0x32,  0x14,  0x90,  0x25,  0x1E,  0x03,  0x90,  0x38,  0x30,  0x00,  0xB0,  0x40,  0x4E,  0x03,  0x90,  0x2C,  0x1C,  0x03,  0xB0,  0x40,  0x16,  0x03,  0xB0,  0x40,  0x00,  0x0C,  0x90,  0x40,
 0x38,  0x41,  0xB0,  0x40,  0x20,  0x05,  0xB0,  0x40,  0x72,  0x04,  0xB0,  0x40,  0x7F,  0x14,  0xB0,  0x40,  0x5E,  0x06,  0xB0,  0x40,  0x7F,  0x67,  0x80,  0x34,  0x00,  0x14,  0x80,  0x40,  0x00,  0x04,  0x80,  0x25,  0x00,  0x05,  0x80,  0x2C,  0x00,  0x02,  0x80,  0x38,  0x00,  0x82,  0x04,  0x90,  0x26,  0x1E,  0x07,  0x90,  0x36,  0x32,  0x02,  0x90,  0x2D,  0x17,  0x02,  0x90,  0x39,  0x36,  0x01,  0x90,  0x3E,  0x2C,  0x0A,  0xB0,  0x40,
 0x54,  0x05,  0xB0,  0x40,  0x18,  0x05,  0xB0,  0x40,  0x00,  0x7D,  0xB0,  0x40,  0x16,  0x04,  0xB0,  0x40,  0x6E,  0x05,  0xB0,  0x40,  0x7F,  0x81,  0x11,  0x80,  0x39,  0x00,  0x81,  0x60,  0x90,  0x38,  0x20,  0x10,  0xB0,  0x40,  0x52,  0x05,  0xB0,  0x40,  0x18,  0x05,  0xB0,  0x40,  0x00,  0x81,  0x14,  0xB0,  0x40,  0x0C,  0x06,  0xB0,  0x40,  0x68,  0x04,  0xB0,  0x40,  0x7F,  0x63,  0x80,  0x38,  0x00,  0x82,  0x01,  0x90,  0x39,  0x16,
 0x46,  0xD0,  0x03,  0x05,  0xD0,  0x07,  0x09,  0xD0,  0x06,  0x05,  0xD0,  0x08,  0x05,  0xD0,  0x0A,  0x05,  0xD0,  0x0B,  0x05,  0xD0,  0x0A,  0x04,  0xD0,  0x08,  0x05,  0xD0,  0x0D,  0x09,  0xD0,  0x0A,  0x05,  0xD0,  0x0C,  0x05,  0xD0,  0x0D,  0x09,  0xD0,  0x10,  0x0A,  0xD0,  0x11,  0x05,  0xD0,  0x13,  0x02,  0x80,  0x36,  0x00,  0x02,  0xD0,  0x14,  0x05,  0xD0,  0x16,  0x05,  0xD0,  0x18,  0x0A,  0xD0,  0x1B,  0x04,  0xD0,  0x1E,  0x05,
 0xD0,  0x1D,  0x09,  0xD0,  0x1F,  0x03,  0x80,  0x26,  0x00,  0x07,  0xD0,  0x21,  0x01,  0x80,  0x2D,  0x00,  0x03,  0xD0,  0x22,  0x04,  0x80,  0x3E,  0x00,  0x07,  0xD0,  0x23,  0x12,  0xD0,  0x1B,  0x04,  0xD0,  0x0A,  0x05,  0xD0,  0x00,  0x0B,  0xB0,  0x40,  0x7F,  0x09,  0xB0,  0x40,  0x7A,  0x06,  0xD0,  0x01,  0x04,  0xD0,  0x02,  0x04,  0xD0,  0x01,  0x05,  0xD0,  0x02,  0x05,  0xD0,  0x05,  0x06,  0x90,  0x39,  0x30,  0x05,  0x90,  0x3D,
 0x34,  0x07,  0x90,  0x31,  0x1E,  0x01,  0x90,  0x2A,  0x1A,  0x09,  0x90,  0x45,  0x38,  0x30,  0xB0,  0x40,  0x00,  0x1A,  0x80,  0x45,  0x00,  0x1C,  0xB0,  0x40,  0x7F,  0x11,  0x90,  0x45,  0x38,  0x2D,  0x80,  0x39,  0x00,  0x0E,  0x80,  0x45,  0x00,  0x12,  0x80,  0x2A,  0x00,  0x16,  0x80,  0x31,  0x00,  0x16,  0x90,  0x45,  0x42,  0x2C,  0x80,  0x45,  0x00,  0x2A,  0x80,  0x39,  0x00,  0x1B,  0x90,  0x45,  0x44,  0x0E,  0x80,  0x3D,  0x00,
 0x05,  0x80,  0x45,  0x00,  0x5D,  0xB0,  0x40,  0x2A,  0x03,  0x90,  0x34,  0x28,  0x02,  0xB0,  0x40,  0x00,  0x02,  0x90,  0x2D,  0x22,  0x03,  0x90,  0x49,  0x3A,  0x04,  0x90,  0x40,  0x3E,  0x01,  0x90,  0x3D,  0x32,  0x5A,  0xB0,  0x40,  0x5C,  0x06,  0xB0,  0x40,  0x7F,  0x22,  0x80,  0x49,  0x00,  0x58,  0x90,  0x49,  0x4E,  0x1C,  0x80,  0x3D,  0x00,  0x01,  0x80,  0x40,  0x00,  0x02,  0x80,  0x34,  0x00,  0x0B,  0x80,  0x2D,  0x00,  0x02,
 0x80,  0x49,  0x00,  0x81,  0x47,  0x90,  0x39,  0x34,  0x05,  0x90,  0x3D,  0x40,  0x04,  0xB0,  0x40,  0x50,  0x00,  0x90,  0x31,  0x24,  0x02,  0x90,  0x2A,  0x18,  0x02,  0xB0,  0x40,  0x14,  0x05,  0xB0,  0x40,  0x00,  0x03,  0x90,  0x45,  0x44,  0x46,  0x80,  0x45,  0x00,  0x0E,  0xB0,  0x40,  0x5C,  0x04,  0xB0,  0x40,  0x7F,  0x1B,  0x90,  0x45,  0x48,  0x3D,  0x80,  0x45,  0x00,  0x3A,  0x90,  0x45,  0x48,  0x1A,  0x80,  0x2A,  0x00,  0x21,
 0x80,  0x31,  0x00,  0x01,  0x80,  0x45,  0x00,  0x32,  0x80,  0x39,  0x00,  0x04,  0x90,  0x45,  0x42,  0x1A,  0x80,  0x3D,  0x00,  0x10,  0x80,  0x45,  0x00,  0x4E,  0x90,  0x2D,  0x30,  0x02,  0x90,  0x34,  0x21,  0x07,  0x90,  0x3D,  0x3C,  0x00,  0x90,  0x49,  0x4C,  0x01,  0x90,  0x40,  0x44,  0x0A,  0xB0,  0x40,  0x50,  0x05,  0xB0,  0x40,  0x12,  0x05,  0xB0,  0x40,  0x00,  0x4B,  0xB0,  0x40,  0x52,  0x05,  0xB0,  0x40,  0x7F,  0x31,  0x80,
 0x49,  0x00,  0x4A,  0x90,  0x49,  0x50,  0x04,  0x80,  0x34,  0x00,  0x10,  0x80,  0x2D,  0x00,  0x06,  0x80,  0x3D,  0x00,  0x02,  0x80,  0x40,  0x00,  0x0F,  0x80,  0x49,  0x00,  0x81,  0x3C,  0x90,  0x25,  0x22,  0x05,  0x90,  0x2C,  0x24,  0x03,  0x90,  0x38,  0x2E,  0x01,  0x90,  0x34,  0x1D,  0x02,  0xB0,  0x40,  0x50,  0x05,  0xB0,  0x40,  0x12,  0x05,  0xB0,  0x40,  0x00,  0x04,  0x90,  0x40,  0x4E,  0x46,  0x80,  0x40,  0x00,  0x07,  0xB0,
 0x40,  0x3C,  0x05,  0xB0,  0x40,  0x7F,  0x25,  0x90,  0x40,  0x3E,  0x3D,  0x80,  0x40,  0x00,  0x3D,  0x90,  0x40,  0x42,  0x35,  0x80,  0x40,  0x00,  0x2D,  0x80,  0x25,  0x00,  0x02,  0x80,  0x2C,  0x00,  0x01,  0x90,  0x40,  0x40,  0x04,  0x80,  0x34,  0x00,  0x1D,  0x80,  0x38,  0x00,  0x07,  0x80,  0x40,  0x00,  0x56,  0x90,  0x39,  0x36,  0x03,  0x90,  0x2D,  0x2E,  0x05,  0xB0,  0x40,  0x50,  0x01,  0x90,  0x26,  0x2A,  0x04,  0xB0,  0x40,
 0x12,  0x02,  0x90,  0x36,  0x25,  0x03,  0xB0,  0x40,  0x00,  0x06,  0x90,  0x42,  0x36,  0x71,  0xB0,  0x40,  0x56,  0x05,  0xB0,  0x40,  0x7F,  0x38,  0x80,  0x36,  0x00,  0x00,  0x80,  0x39,  0x00,  0x27,  0x80,  0x42,  0x00,  0x25,  0x80,  0x26,  0x00,  0x0E,  0x80,  0x2D,  0x00,  0x81,  0x57,  0x90,  0x5A,  0x34,  0x4F,  0x80,  0x5A,  0x00,  0x81,  0x17,  0x90,  0x55,  0x32,  0x33,  0x80,  0x55,  0x00,  0x81,  0x4D,  0x90,  0x51,  0x30,  0x81,
 0x03,  0x80,  0x51,  0x00,  0x63,  0x90,  0x4E,  0x32,  0x0F,  0x80,  0x4E,  0x00,  0x81,  0x52,  0xB0,  0x40,  0x54,  0x04,  0xB0,  0x40,  0x1C,  0x03,  0x90,  0x39,  0x30,  0x04,  0xB0,  0x40,  0x00,  0x02,  0x90,  0x3D,  0x3A,  0x0B,  0x90,  0x2A,  0x1D,  0x02,  0x90,  0x31,  0x21,  0x06,  0x90,  0x45,  0x34,  0x3B,  0xB0,  0x40,  0x54,  0x04,  0x80,  0x45,  0x00,  0x01,  0xB0,  0x40,  0x7F,  0x2F,  0x90,  0x45,  0x3C,  0x3C,  0x80,  0x45,  0x00,
 0x33,  0x80,  0x2A,  0x00,  0x06,  0x90,  0x45,  0x45,  0x11,  0x80,  0x31,  0x00,  0x22,  0x80,  0x45,  0x00,  0x22,  0x80,  0x39,  0x00,  0x0E,  0x90,  0x45,  0x44,  0x1A,  0x80,  0x3D,  0x00,  0x0F,  0x80,  0x45,  0x00,  0x50,  0xB0,  0x40,  0x52,  0x02,  0x90,  0x3D,  0x36,  0x03,  0xB0,  0x40,  0x1A,  0x04,  0xB0,  0x40,  0x00,  0x03,  0x90,  0x2D,  0x1C,  0x00,  0x90,  0x40,  0x40,  0x01,  0x90,  0x49,  0x3E,  0x02,  0x90,  0x34,  0x1D,  0x47,
 0xB0,  0x40,  0x2C,  0x05,  0xB0,  0x40,  0x78,  0x05,  0xB0,  0x40,  0x7F,  0x17,  0x80,  0x40,  0x00,  0x09,  0x80,  0x3D,  0x00,  0x13,  0x80,  0x49,  0x00,  0x17,  0x80,  0x34,  0x00,  0x04,  0x80,  0x2D,  0x00,  0x3C,  0x90,  0x4C,  0x42,  0x16,  0x80,  0x4C,  0x00,  0x81,  0x54,  0x90,  0x39,  0x38,  0x02,  0x90,  0x2A,  0x28,  0x01,  0x90,  0x31,  0x2C,  0x02,  0x90,  0x3D,  0x3A,  0x06,  0xB0,  0x40,  0x52,  0x05,  0xB0,  0x40,  0x1A,  0x04,
 0xB0,  0x40,  0x00,  0x05,  0x90,  0x45,  0x42,  0x3D,  0x80,  0x45,  0x00,  0x06,  0xB0,  0x40,  0x5E,  0x05,  0xB0,  0x40,  0x7F,  0x24,  0x90,  0x45,  0x42,  0x2F,  0x80,  0x45,  0x00,  0x3F,  0x90,  0x45,  0x4A,  0x16,  0x80,  0x2A,  0x00,  0x0E,  0x80,  0x31,  0x00,  0x05,  0x80,  0x45,  0x00,  0x40,  0x80,  0x39,  0x00,  0x09,  0x90,  0x45,  0x4C,  0x18,  0x80,  0x3D,  0x00,  0x16,  0x80,  0x45,  0x00,  0x48,  0x90,  0x2D,  0x1C,  0x01,  0xB0,
 0x40,  0x52,  0x05,  0xB0,  0x40,  0x1A,  0x01,  0x90,  0x34,  0x20,  0x03,  0x90,  0x49,  0x3C,  0x03,  0xB0,  0x40,  0x00,  0x01,  0x90,  0x3D,  0x28,  0x0E,  0x90,  0x40,  0x2A,  0x2E,  0xB0,  0x40,  0x54,  0x04,  0xB0,  0x40,  0x7F,  0x27,  0x80,  0x40,  0x00,  0x05,  0x80,  0x3D,  0x00,  0x12,  0x80,  0x49,  0x00,  0x16,  0x80,  0x34,  0x00,  0x16,  0x80,  0x2D,  0x00,  0x3E,  0x90,  0x4C,  0x3C,  0x11,  0x80,  0x4C,  0x00,  0x81,  0x5D,  0x90,
 0x38,  0x34,  0x02,  0x90,  0x2C,  0x2E,  0x01,  0x90,  0x25,  0x2C,  0x03,  0x90,  0x40,  0x30,  0x06,  0xB0,  0x40,  0x52,  0x05,  0xB0,  0x40,  0x1A,  0x05,  0xB0,  0x40,  0x00,  0x3E,  0x80,  0x40,  0x00,  0x0A,  0xB0,  0x40,  0x4C,  0x04,  0xB0,  0x40,  0x7F,  0x22,  0x90,  0x40,  0x3A,  0x41,  0x80,  0x40,  0x00,  0x3A,  0x90,  0x40,  0x46,  0x0E,  0x80,  0x38,  0x00,  0x0B,  0x80,  0x25,  0x00,  0x01,  0x80,  0x2C,  0x00,  0x10,  0x80,  0x40,
 0x00,  0x81,  0x3E,  0xB0,  0x40,  0x52,  0x03,  0x90,  0x2D,  0x20,  0x02,  0x90,  0x26,  0x1C,  0x01,  0xB0,  0x40,  0x1A,  0x01,  0x90,  0x3D,  0x2A,  0x03,  0x90,  0x49,  0x32,  0x01,  0xB0,  0x40,  0x00,  0x51,  0xB0,  0x40,  0x3C,  0x05,  0xB0,  0x40,  0x7F,  0x2B,  0x90,  0x45,  0x36,  0x1B,  0x80,  0x49,  0x00,  0x16,  0x80,  0x3D,  0x00,  0x08,  0x80,  0x26,  0x00,  0x07,  0x80,  0x2D,  0x00,  0x11,  0x80,  0x45,  0x00,  0x82,  0x10,  0x90,
 0x31,  0x24,  0x02,  0x90,  0x5A,  0x38,  0x00,  0xB0,  0x40,  0x5E,  0x05,  0xB0,  0x40,  0x1E,  0x04,  0x90,  0x2A,  0x1A,  0x01,  0xB0,  0x40,  0x04,  0x04,  0xB0,  0x40,  0x00,  0x65,  0xB0,  0x40,  0x2A,  0x05,  0xB0,  0x40,  0x78,  0x05,  0xB0,  0x40,  0x7F,  0x65,  0x80,  0x5A,  0x00,  0x08,  0x90,  0x55,  0x44,  0x3D,  0x80,  0x55,  0x00,  0x81,  0x24,  0x90,  0x51,  0x38,  0x5D,  0x80,  0x51,  0x00,  0x4C,  0x80,  0x2A,  0x00,  0x17,  0x80,
 0x31,  0x00,  0x30,  0x90,  0x4E,  0x44,  0x10,  0x80,  0x4E,  0x00,  0x81,  0x67,  0xB0,  0x40,  0x52,  0x05,  0xB0,  0x40,  0x1A,  0x01,  0x90,  0x31,  0x20,  0x04,  0xB0,  0x40,  0x00,  0x00,  0x90,  0x3D,  0x2A,  0x01,  0x90,  0x2A,  0x1D,  0x03,  0x90,  0x45,  0x36,  0x02,  0x90,  0x49,  0x3A,  0x48,  0xB0,  0x40,  0x2C,  0x03,  0xB0,  0x40,  0x78,  0x05,  0xB0,  0x40,  0x7F,  0x10,  0x80,  0x49,  0x00,  0x19,  0x90,  0x49,  0x2C,  0x3D,  0x80,
 0x49,  0x00,  0x3F,  0x80,  0x2A,  0x00,  0x02,  0x90,  0x49,  0x34,  0x0E,  0x80,  0x31,  0x00,  0x14,  0x80,  0x49,  0x00,  0x44,  0x90,  0x49,  0x42,  0x11,  0x80,  0x3D,  0x00,  0x18,  0x80,  0x45,  0x00,  0x3E,  0x80,  0x49,  0x00,  0x16,  0xB0,  0x40,  0x52,  0x06,  0xB0,  0x40,  0x1A,  0x03,  0x90,  0x34,  0x21,  0x02,  0xB0,  0x40,  0x00,  0x01,  0x90,  0x2D,  0x20,  0x06,  0x90,  0x49,  0x3B,  0x01,  0x90,  0x4C,  0x45,  0x10,  0x90,  0x40,
 0x20,  0x25,  0xB0,  0x40,  0x28,  0x06,  0xB0,  0x40,  0x76,  0x04,  0xB0,  0x40,  0x7F,  0x2C,  0x80,  0x40,  0x00,  0x14,  0x80,  0x49,  0x00,  0x10,  0x80,  0x4C,  0x00,  0x20,  0x80,  0x34,  0x00,  0x0A,  0x80,  0x2D,  0x00,  0x36,  0x90,  0x4E,  0x42,  0x10,  0x80,  0x4E,  0x00,  0x81,  0x61,  0xB0,  0x40,  0x52,  0x04,  0x90,  0x3D,  0x30,  0x00,  0x90,  0x2A,  0x1F,  0x01,  0xB0,  0x40,  0x1A,  0x03,  0x90,  0x31,  0x22,  0x03,  0xB0,  0x40,
 0x00,  0x00,  0x90,  0x45,  0x38,  0x06,  0x90,  0x49,  0x3E,  0x32,  0xB0,  0x40,  0x5C,  0x05,  0xB0,  0x40,  0x7F,  0x10,  0x80,  0x49,  0x00,  0x2C,  0x90,  0x49,  0x40,  0x40,  0x80,  0x49,  0x00,  0x31,  0x80,  0x2A,  0x00,  0x06,  0x90,  0x49,  0x3E,  0x05,  0x80,  0x31,  0x00,  0x2D,  0x80,  0x49,  0x00,  0x3F,  0x80,  0x3D,  0x00,  0x0B,  0x90,  0x49,  0x3E,  0x19,  0x80,  0x45,  0x00,  0x0A,  0x80,  0x49,  0x00,  0x46,  0xB0,  0x40,  0x52,
 0x05,  0xB0,  0x40,  0x24,  0x06,  0x90,  0x2D,  0x1E,  0x00,  0xB0,  0x40,  0x00,  0x03,  0x90,  0x49,  0x3C,  0x05,  0x90,  0x4C,  0x4A,  0x09,  0x90,  0x34,  0x16,  0x0D,  0x90,  0x40,  0x10,  0x15,  0xB0,  0x40,  0x1E,  0x05,  0xB0,  0x40,  0x72,  0x05,  0xB0,  0x40,  0x7F,  0x31,  0x80,  0x40,  0x00,  0x18,  0x80,  0x49,  0x00,  0x16,  0x80,  0x4C,  0x00,  0x20,  0x80,  0x34,  0x00,  0x0D,  0x80,  0x2D,  0x00,  0x1B,  0x90,  0x4E,  0x58,  0x0E,
 0x80,  0x4E,  0x00,  0x81,  0x51,  0xB0,  0x40,  0x5C,  0x05,  0xB0,  0x40,  0x1E,  0x00,  0x90,  0x25,  0x1A,  0x04,  0xB0,  0x40,  0x01,  0x04,  0xB0,  0x40,  0x00,  0x04,  0x90,  0x2C,  0x24,  0x01,  0x90,  0x3B,  0x22,  0x04,  0x90,  0x40,  0x38,  0x13,  0x90,  0x47,  0x3A,  0x1E,  0xB0,  0x40,  0x50,  0x06,  0xB0,  0x40,  0x7F,  0x0F,  0x80,  0x47,  0x00,  0x39,  0x90,  0x47,  0x32,  0x32,  0x80,  0x47,  0x00,  0x4D,  0x90,  0x47,  0x40,  0x1E,
 0x80,  0x3B,  0x00,  0x07,  0x80,  0x25,  0x00,  0x09,  0x80,  0x47,  0x00,  0x09,  0x80,  0x40,  0x00,  0x0E,  0x80,  0x2C,  0x00,  0x27,  0x90,  0x47,  0x34,  0x11,  0x80,  0x47,  0x00,  0x63,  0x90,  0x3D,  0x3E,  0x02,  0x90,  0x2D,  0x21,  0x01,  0xB0,  0x40,  0x52,  0x03,  0x90,  0x26,  0x2A,  0x03,  0xB0,  0x40,  0x1A,  0x04,  0xB0,  0x40,  0x00,  0x0A,  0x90,  0x49,  0x36,  0x0B,  0x90,  0x45,  0x1A,  0x1F,  0x80,  0x45,  0x00,  0x58,  0xB0,
 0x40,  0x22,  0x05,  0xB0,  0x40,  0x74,  0x04,  0xB0,  0x40,  0x7F,  0x59,  0x90,  0x45,  0x2C,  0x17,  0x80,  0x26,  0x00,  0x01,  0x80,  0x3D,  0x00,  0x0C,  0x80,  0x45,  0x00,  0x00,  0x80,  0x2D,  0x00,  0x07,  0x80,  0x49,  0x00,  0x81,  0x50,  0x90,  0x5A,  0x3C,  0x39,  0x80,  0x5A,  0x00,  0x81,  0x28,  0x90,  0x55,  0x36,  0x2E,  0x80,  0x55,  0x00,  0x81,  0x36,  0x90,  0x51,  0x34,  0x3B,  0x80,  0x51,  0x00,  0x81,  0x32,  0x90,  0x4E,
 0x36,  0x0F,  0x80,  0x4E,  0x00,  0x81,  0x62,  0xB0,  0x40,  0x58,  0x05,  0xB0,  0x40,  0x1C,  0x05,  0xB0,  0x40,  0x00,  0x02,  0x90,  0x2A,  0x21,  0x00,  0x90,  0x3D,  0x26,  0x05,  0x90,  0x31,  0x22,  0x01,  0x90,  0x45,  0x32,  0x0A,  0x90,  0x49,  0x36,  0x35,  0xB0,  0x40,  0x4C,  0x05,  0xB0,  0x40,  0x7F,  0x00,  0x80,  0x49,  0x00,  0x3B,  0x90,  0x49,  0x2C,  0x3A,  0x80,  0x49,  0x00,  0x3D,  0x90,  0x49,  0x3C,  0x31,  0x80,  0x2A,
 0x00,  0x05,  0x80,  0x31,  0x00,  0x02,  0x80,  0x49,  0x00,  0x3D,  0x90,  0x49,  0x40,  0x11,  0x80,  0x3D,  0x00,  0x10,  0x80,  0x45,  0x00,  0x14,  0x80,  0x49,  0x00,  0x2E,  0xB0,  0x40,  0x56,  0x05,  0xB0,  0x40,  0x1A,  0x02,  0x90,  0x2D,  0x28,  0x02,  0xB0,  0x40,  0x00,  0x01,  0x90,  0x34,  0x20,  0x10,  0x90,  0x49,  0x3E,  0x03,  0x90,  0x4C,  0x50,  0x06,  0x90,  0x40,  0x32,  0x21,  0xB0,  0x40,  0x18,  0x04,  0xB0,  0x40,  0x6E,
 0x05,  0xB0,  0x40,  0x7F,  0x48,  0x80,  0x40,  0x00,  0x1B,  0x80,  0x49,  0x00,  0x08,  0x80,  0x4C,  0x00,  0x10,  0x80,  0x34,  0x00,  0x0F,  0x80,  0x2D,  0x00,  0x26,  0x90,  0x4E,  0x4C,  0x10,  0x80,  0x4E,  0x00,  0x81,  0x51,  0x90,  0x31,  0x28,  0x02,  0x90,  0x2A,  0x1D,  0x02,  0x90,  0x3D,  0x32,  0x03,  0x90,  0x45,  0x3E,  0x00,  0xB0,  0x40,  0x56,  0x03,  0x90,  0x49,  0x3C,  0x02,  0xB0,  0x40,  0x1A,  0x06,  0xB0,  0x40,  0x00,
 0x35,  0xB0,  0x40,  0x2C,  0x04,  0xB0,  0x40,  0x78,  0x05,  0xB0,  0x40,  0x7F,  0x04,  0x80,  0x49,  0x00,  0x01,  0xB0,  0x40,  0x5E,  0x05,  0xB0,  0x40,  0x7F,  0x26,  0x90,  0x49,  0x3E,  0x41,  0x80,  0x49,  0x00,  0x38,  0x80,  0x2A,  0x00,  0x03,  0x90,  0x49,  0x46,  0x06,  0x80,  0x31,  0x00,  0x29,  0x80,  0x49,  0x00,  0x37,  0x90,  0x49,  0x32,  0x10,  0x80,  0x3D,  0x00,  0x14,  0x80,  0x45,  0x00,  0x0D,  0x80,  0x49,  0x00,  0x24,
 0xB0,  0x40,  0x56,  0x05,  0xB0,  0x40,  0x1A,  0x06,  0xB0,  0x40,  0x00,  0x0F,  0x90,  0x34,  0x24,  0x06,  0x90,  0x2D,  0x20,  0x10,  0x90,  0x4C,  0x50,  0x03,  0x90,  0x49,  0x3A,  0x03,  0xB0,  0x40,  0x1E,  0x05,  0xB0,  0x40,  0x72,  0x04,  0xB0,  0x40,  0x41,  0x04,  0xB0,  0x40,  0x32,  0x02,  0x90,  0x40,  0x12,  0x03,  0xB0,  0x40,  0x7C,  0x05,  0xB0,  0x40,  0x7F,  0x5A,  0x80,  0x40,  0x00,  0x1F,  0x80,  0x49,  0x00,  0x02,  0x80,
 0x4C,  0x00,  0x0F,  0x80,  0x34,  0x00,  0x14,  0x80,  0x2D,  0x00,  0x1C,  0x90,  0x4E,  0x58,  0x11,  0x80,  0x4E,  0x00,  0x81,  0x4D,  0xB0,  0x40,  0x52,  0x06,  0xB0,  0x40,  0x18,  0x02,  0x90,  0x25,  0x1F,  0x02,  0xB0,  0x40,  0x00,  0x04,  0x90,  0x2C,  0x1F,  0x01,  0x90,  0x3B,  0x2E,  0x04,  0x90,  0x40,  0x3A,  0x0D,  0x90,  0x47,  0x44,  0x23,  0xB0,  0x40,  0x2C,  0x06,  0xB0,  0x40,  0x78,  0x05,  0xB0,  0x40,  0x7F,  0x0C,  0x80,
 0x47,  0x00,  0x36,  0x90,  0x47,  0x3E,  0x3B,  0x80,  0x47,  0x00,  0x3C,  0x90,  0x47,  0x38,  0x2D,  0x80,  0x47,  0x00,  0x03,  0x80,  0x3B,  0x00,  0x15,  0x80,  0x25,  0x00,  0x07,  0x80,  0x2C,  0x00,  0x20,  0x90,  0x47,  0x32,  0x19,  0x80,  0x47,  0x00,  0x02,  0x80,  0x40,  0x00,  0x59,  0x90,  0x3D,  0x3C,  0x03,  0x90,  0x2D,  0x2A,  0x00,  0xB0,  0x40,  0x66,  0x04,  0x90,  0x45,  0x22,  0x01,  0xB0,  0x40,  0x24,  0x03,  0x90,  0x26,
 0x2A,  0x02,  0xB0,  0x40,  0x02,  0x05,  0xB0,  0x40,  0x00,  0x05,  0x90,  0x49,  0x38,  0x31,  0x80,  0x45,  0x00,  0x29,  0xB0,  0x40,  0x50,  0x06,  0xB0,  0x40,  0x7F,  0x7E,  0x90,  0x45,  0x2E,  0x03,  0x80,  0x26,  0x00,  0x06,  0x80,  0x49,  0x00,  0x01,  0x80,  0x3D,  0x00,  0x05,  0x80,  0x45,  0x00,  0x06,  0x80,  0x2D,  0x00,  0x81,  0x49,  0x90,  0x39,  0x26,  0x01,  0x90,  0x2A,  0x1E,  0x06,  0x90,  0x3D,  0x2C,  0x03,  0x90,  0x31,
 0x21,  0x02,  0xB0,  0x40,  0x52,  0x00,  0x90,  0x45,  0x3E,  0x04,  0xB0,  0x40,  0x1A,  0x05,  0xB0,  0x40,  0x00,  0x72,  0x90,  0x42,  0x2E,  0x1C,  0x80,  0x45,  0x00,  0x07,  0xB0,  0x40,  0x36,  0x05,  0xB0,  0x40,  0x7E,  0x05,  0xB0,  0x40,  0x7F,  0x81,  0x24,  0x80,  0x39,  0x00,  0x10,  0x80,  0x3D,  0x00,  0x12,  0x80,  0x42,  0x00,  0x54,  0x80,  0x2A,  0x00,  0x07,  0x80,  0x31,  0x00,  0x81,  0x05,  0x90,  0x45,  0x3E,  0x7C,  0x90,
 0x49,  0x4A,  0x10,  0x80,  0x45,  0x00,  0x15,  0x80,  0x49,  0x00,  0x50,  0x90,  0x36,  0x24,  0x13,  0x80,  0x36,  0x00,  0x73,  0xB0,  0x40,  0x50,  0x02,  0x90,  0x38,  0x3C,  0x01,  0x90,  0x2C,  0x24,  0x01,  0xB0,  0x40,  0x18,  0x01,  0x90,  0x40,  0x40,  0x03,  0x90,  0x25,  0x21,  0x01,  0xB0,  0x40,  0x00,  0x4C,  0x80,  0x40,  0x00,  0x15,  0xB0,  0x40,  0x24,  0x04,  0xB0,  0x40,  0x74,  0x05,  0xB0,  0x40,  0x7F,  0x19,  0x90,  0x40,
 0x38,  0x27,  0x80,  0x2C,  0x00,  0x0A,  0x80,  0x25,  0x00,  0x06,  0x80,  0x40,  0x00,  0x46,  0x90,  0x40,  0x38,  0x1A,  0x80,  0x38,  0x00,  0x21,  0x80,  0x40,  0x00,  0x81,  0x1C,  0xB0,  0x40,  0x5A,  0x04,  0xB0,  0x40,  0x1C,  0x04,  0x90,  0x39,  0x3C,  0x01,  0xB0,  0x40,  0x00,  0x06,  0x90,  0x26,  0x22,  0x04,  0x90,  0x2D,  0x1D,  0x11,  0x90,  0x42,  0x42,  0x4A,  0xB0,  0x40,  0x1E,  0x05,  0xB0,  0x40,  0x72,  0x05,  0xB0,  0x40,
 0x7F,  0x4B,  0x80,  0x39,  0x00,  0x1E,  0x80,  0x42,  0x00,  0x03,  0x80,  0x26,  0x00,  0x1D,  0x80,  0x2D,  0x00,  0x81,  0x7F,  0x90,  0x5A,  0x42,  0x50,  0x80,  0x5A,  0x00,  0x81,  0x21,  0x90,  0x55,  0x3E,  0x37,  0x80,  0x55,  0x00,  0x81,  0x39,  0x90,  0x51,  0x32,  0x13,  0x80,  0x51,  0x00,  0x81,  0x7C,  0x90,  0x4E,  0x32,  0x16,  0x80,  0x4E,  0x00,  0x5B,  0x90,  0x47,  0x31,  0x3A,  0x90,  0x45,  0x3A,  0x05,  0x90,  0x49,  0x48,
 0x00,  0x90,  0x40,  0x46,  0x0C,  0x90,  0x28,  0x36,  0x08,  0x90,  0x34,  0x2A,  0x10,  0xB0,  0x40,  0x00,  0x1B,  0xB0,  0x40,  0x0E,  0x05,  0xB0,  0x40,  0x56,  0x05,  0xB0,  0x40,  0x7F,  0x72,  0x80,  0x47,  0x00,  0x20,  0x80,  0x28,  0x00,  0x25,  0x80,  0x34,  0x00,  0x7B,  0x90,  0x47,  0x46,  0x00,  0x80,  0x40,  0x00,  0x0A,  0x80,  0x49,  0x00,  0x0F,  0x80,  0x45,  0x00,  0x5F,  0x90,  0x45,  0x42,  0x0D,  0x90,  0x49,  0x48,  0x02,
 0x90,  0x40,  0x38,  0x05,  0x90,  0x3B,  0x2A,  0x0A,  0x90,  0x34,  0x21,  0x03,  0x90,  0x39,  0x21,  0x0A,  0x80,  0x47,  0x00,  0x82,  0x26,  0x80,  0x34,  0x00,  0x02,  0x80,  0x39,  0x00,  0x0E,  0x80,  0x3B,  0x00,  0x0B,  0x80,  0x40,  0x00,  0x01,  0x90,  0x47,  0x46,  0x0E,  0x80,  0x49,  0x00,  0x13,  0x80,  0x45,  0x00,  0x6B,  0x90,  0x45,  0x34,  0x08,  0x90,  0x40,  0x2C,  0x00,  0x90,  0x3B,  0x20,  0x05,  0x90,  0x34,  0x19,  0x04,
 0x90,  0x49,  0x4C,  0x07,  0x90,  0x39,  0x20,  0x82,  0x25,  0x80,  0x39,  0x00,  0x14,  0x80,  0x34,  0x00,  0x16,  0x80,  0x47,  0x00,  0x01,  0x90,  0x4C,  0x52,  0x03,  0x80,  0x3B,  0x00,  0x12,  0x80,  0x49,  0x00,  0x0C,  0x80,  0x40,  0x00,  0x0A,  0x80,  0x45,  0x00,  0x58,  0x80,  0x4C,  0x00,  0x06,  0x90,  0x34,  0x2A,  0x06,  0x90,  0x45,  0x3A,  0x00,  0x90,  0x40,  0x30,  0x00,  0x90,  0x47,  0x32,  0x03,  0x90,  0x3B,  0x24,  0x02,
 0x90,  0x39,  0x30,  0x18,  0x90,  0x49,  0x40,  0x31,  0x80,  0x47,  0x00,  0x15,  0x80,  0x45,  0x00,  0x16,  0x80,  0x39,  0x00,  0x0B,  0x80,  0x3B,  0x00,  0x0C,  0x80,  0x34,  0x00,  0x18,  0x90,  0x47,  0x34,  0x30,  0x80,  0x49,  0x00,  0x10,  0x80,  0x40,  0x00,  0x81,  0x2A,  0xB0,  0x40,  0x52,  0x02,  0x90,  0x28,  0x38,  0x04,  0xB0,  0x40,  0x1A,  0x00,  0x90,  0x45,  0x32,  0x00,  0x90,  0x34,  0x38,  0x05,  0xB0,  0x40,  0x00,  0x03,
 0x90,  0x40,  0x1C,  0x04,  0x90,  0x49,  0x44,  0x66,  0xB0,  0x40,  0x3C,  0x06,  0xB0,  0x40,  0x7F,  0x15,  0x80,  0x47,  0x00,  0x67,  0x80,  0x28,  0x00,  0x24,  0x80,  0x34,  0x00,  0x6D,  0x90,  0x47,  0x46,  0x0D,  0x80,  0x40,  0x00,  0x0E,  0x80,  0x45,  0x00,  0x05,  0x80,  0x49,  0x00,  0x5F,  0x90,  0x45,  0x36,  0x03,  0x90,  0x3B,  0x26,  0x05,  0x90,  0x34,  0x26,  0x05,  0x90,  0x40,  0x30,  0x04,  0x90,  0x39,  0x2C,  0x00,  0x90,
 0x49,  0x48,  0x2B,  0x80,  0x47,  0x00,  0x81,  0x73,  0x80,  0x40,  0x00,  0x48,  0x80,  0x39,  0x00,  0x01,  0x90,  0x47,  0x44,  0x02,  0x80,  0x34,  0x00,  0x0F,  0x80,  0x3B,  0x00,  0x09,  0x80,  0x49,  0x00,  0x1C,  0x80,  0x45,  0x00,  0x05,  0x90,  0x34,  0x32,  0x06,  0x90,  0x45,  0x36,  0x00,  0x90,  0x3B,  0x30,  0x06,  0x90,  0x40,  0x34,  0x02,  0x90,  0x39,  0x2A,  0x03,  0x90,  0x49,  0x4E,  0x6D,  0x80,  0x47,  0x00,  0x6B,  0x80,
 0x40,  0x00,  0x69,  0x80,  0x45,  0x00,  0x03,  0x80,  0x39,  0x00,  0x07,  0x80,  0x49,  0x00,  0x04,  0x80,  0x3B,  0x00,  0x06,  0x80,  0x34,  0x00,  0x04,  0x90,  0x40,  0x4C,  0x78,  0x90,  0x45,  0x4A,  0x08,  0x90,  0x34,  0x36,  0x04,  0x90,  0x3B,  0x2E,  0x01,  0x90,  0x39,  0x36,  0x03,  0x90,  0x47,  0x52,  0x04,  0x80,  0x40,  0x00,  0x2B,  0x90,  0x49,  0x50,  0x04,  0x80,  0x45,  0x00,  0x18,  0x80,  0x47,  0x00,  0x2D,  0x80,  0x39,
 0x00,  0x14,  0x90,  0x47,  0x50,  0x0C,  0x80,  0x3B,  0x00,  0x10,  0x80,  0x34,  0x00,  0x20,  0x80,  0x49,  0x00,  0x37,  0x90,  0x45,  0x4E,  0x0C,  0x80,  0x47,  0x00,  0x25,  0xD0,  0x06,  0x13,  0xD0,  0x04,  0x05,  0xD0,  0x03,  0x0B,  0x80,  0x45,  0x00,  0x08,  0xD0,  0x02,  0x05,  0xD0,  0x00,  0x0F,  0x90,  0x44,  0x39,  0x55,  0x80,  0x44,  0x00,  0x27,  0xB0,  0x40,  0x6C,  0x06,  0xB0,  0x40,  0x26,  0x00,  0x90,  0x26,  0x44,  0x04,
 0x90,  0x32,  0x4A,  0x01,  0xB0,  0x40,  0x0A,  0x02,  0x90,  0x3E,  0x4E,  0x00,  0x90,  0x42,  0x44,  0x03,  0xB0,  0x40,  0x00,  0x0D,  0x90,  0x39,  0x38,  0x59,  0xD0,  0x01,  0x06,  0xD0,  0x10,  0x02,  0xB0,  0x40,  0x5C,  0x02,  0xD0,  0x20,  0x03,  0xB0,  0x40,  0x7F,  0x01,  0xD0,  0x21,  0x18,  0xD0,  0x22,  0x09,  0xD0,  0x21,  0x05,  0xD0,  0x20,  0x0E,  0xD0,  0x1E,  0x05,  0xD0,  0x1B,  0x05,  0xD0,  0x19,  0x05,  0xD0,  0x17,  0x05,
 0xD0,  0x13,  0x05,  0xD0,  0x0E,  0x04,  0xD0,  0x07,  0x05,  0xD0,  0x00,  0x00,  0x80,  0x26,  0x00,  0x0E,  0x80,  0x39,  0x00,  0x0C,  0x80,  0x32,  0x00,  0x01,  0x80,  0x3E,  0x00,  0x11,  0x80,  0x42,  0x00,  0x81,  0x73,  0x90,  0x32,  0x3C,  0x03,  0x90,  0x39,  0x32,  0x09,  0x90,  0x3E,  0x2A,  0x03,  0x90,  0x42,  0x42,  0x00,  0x90,  0x45,  0x46,  0x00,  0x90,  0x49,  0x46,  0x58,  0x80,  0x49,  0x00,  0x2F,  0x90,  0x49,  0x40,  0x4C,
 0x80,  0x49,  0x00,  0x27,  0x90,  0x49,  0x4E,  0x4B,  0x80,  0x49,  0x00,  0x0B,  0x80,  0x39,  0x00,  0x03,  0x80,  0x32,  0x00,  0x16,  0x90,  0x49,  0x50,  0x02,  0x80,  0x42,  0x00,  0x07,  0x80,  0x3E,  0x00,  0x13,  0x80,  0x45,  0x00,  0x17,  0x90,  0x32,  0x3A,  0x02,  0x90,  0x49,  0x4E,  0x01,  0x90,  0x42,  0x42,  0x01,  0x90,  0x45,  0x44,  0x00,  0x80,  0x49,  0x00,  0x04,  0x90,  0x3E,  0x24,  0x00,  0x90,  0x39,  0x38,  0x81,  0x6C,
 0x90,  0x47,  0x4A,  0x21,  0x80,  0x49,  0x00,  0x20,  0x80,  0x45,  0x00,  0x09,  0x80,  0x42,  0x00,  0x06,  0x80,  0x39,  0x00,  0x21,  0x80,  0x3E,  0x00,  0x55,  0x80,  0x47,  0x00,  0x30,  0x90,  0x45,  0x56,  0x07,  0x90,  0x3E,  0x2C,  0x03,  0x90,  0x39,  0x3C,  0x02,  0x90,  0x40,  0x4E,  0x81,  0x06,  0xD0,  0x01,  0x05,  0xD0,  0x02,  0x09,  0xD0,  0x03,  0x06,  0xD0,  0x04,  0x25,  0xD0,  0x03,  0x0B,  0xD0,  0x02,  0x02,  0x80,  0x40,
 0x00,  0x04,  0x80,  0x3E,  0x00,  0x02,  0xD0,  0x01,  0x01,  0x80,  0x39,  0x00,  0x08,  0xD0,  0x00,  0x0C,  0x90,  0x47,  0x56,  0x0A,  0x80,  0x32,  0x00,  0x16,  0xD0,  0x0A,  0x09,  0x80,  0x45,  0x00,  0x0E,  0xD0,  0x09,  0x0F,  0xD0,  0x07,  0x03,  0xD0,  0x06,  0x06,  0xD0,  0x05,  0x04,  0xD0,  0x03,  0x05,  0xD0,  0x02,  0x05,  0xD0,  0x01,  0x04,  0xD0,  0x00,  0x22,  0x90,  0x49,  0x40,  0x02,  0x80,  0x47,  0x00,  0x68,  0x90,  0x40,
 0x40,  0x00,  0x90,  0x3D,  0x40,  0x05,  0x90,  0x2D,  0x4A,  0x06,  0x90,  0x39,  0x3A,  0x17,  0xB0,  0x40,  0x00,  0x26,  0xB0,  0x40,  0x7F,  0x23,  0xD0,  0x03,  0x05,  0xD0,  0x09,  0x04,  0xD0,  0x0E,  0x02,  0xD0,  0x0D,  0x03,  0xD0,  0x14,  0x02,  0xD0,  0x0E,  0x03,  0xD0,  0x19,  0x05,  0xD0,  0x1E,  0x01,  0xD0,  0x0D,  0x05,  0xD0,  0x22,  0x00,  0xD0,  0x0E,  0x05,  0xD0,  0x27,  0x03,  0xD0,  0x4B,  0x02,  0xD0,  0x0D,  0x03,  0xD0,
 0x6D,  0x06,  0xD0,  0x0C,  0x0A,  0xD0,  0x0A,  0x09,  0xD0,  0x09,  0x03,  0xD0,  0x6F,  0x02,  0xD0,  0x06,  0x03,  0xD0,  0x70,  0x03,  0xD0,  0x02,  0x04,  0xD0,  0x00,  0x1F,  0xD0,  0x71,  0x09,  0xD0,  0x72,  0x0A,  0xD0,  0x73,  0x2A,  0xD0,  0x74,  0x0A,  0xD0,  0x73,  0x18,  0xD0,  0x71,  0x04,  0xD0,  0x70,  0x05,  0xD0,  0x71,  0x0A,  0xD0,  0x70,  0x13,  0xD0,  0x6F,  0x04,  0xD0,  0x6E,  0x0A,  0xD0,  0x6D,  0x05,  0xD0,  0x6B,  0x09,
 0xD0,  0x6A,  0x04,  0xD0,  0x69,  0x0A,  0xD0,  0x68,  0x05,  0xD0,  0x67,  0x05,  0xD0,  0x66,  0x09,  0xD0,  0x65,  0x04,  0xD0,  0x64,  0x1E,  0xD0,  0x65,  0x05,  0xD0,  0x66,  0x03,  0xD0,  0x67,  0x0A,  0xD0,  0x68,  0x04,  0xD0,  0x6A,  0x0F,  0xD0,  0x69,  0x05,  0xD0,  0x64,  0x04,  0xD0,  0x56,  0x05,  0xD0,  0x3A,  0x05,  0xD0,  0x17,  0x04,  0x80,  0x3D,  0x00,  0x01,  0x80,  0x39,  0x00,  0x00,  0x80,  0x40,  0x00,  0x01,  0xD0,  0x0F,
 0x03,  0xD0,  0x00,  0x02,  0x80,  0x49,  0x00,  0x01,  0x90,  0x2D,  0x3C,  0x02,  0x90,  0x3D,  0x3A,  0x00,  0x80,  0x2D,  0x00,  0x03,  0x90,  0x40,  0x42,  0x01,  0x90,  0x39,  0x2E,  0x00,  0x90,  0x49,  0x46,  0x4C,  0x80,  0x49,  0x00,  0x31,  0x90,  0x49,  0x42,  0x41,  0x80,  0x49,  0x00,  0x2B,  0x90,  0x49,  0x42,  0x3A,  0x80,  0x49,  0x00,  0x24,  0x80,  0x3D,  0x00,  0x0D,  0x80,  0x2D,  0x00,  0x0C,  0x90,  0x49,  0x50,  0x14,  0x80,
 0x39,  0x00,  0x27,  0x80,  0x40,  0x00,  0x05,  0x80,  0x49,  0x00,  0x3C,  0x90,  0x40,  0x42,  0x03,  0x90,  0x2D,  0x3C,  0x01,  0x90,  0x3D,  0x46,  0x03,  0x90,  0x49,  0x50,  0x04,  0x90,  0x39,  0x2E,  0x81,  0x0C,  0x80,  0x3D,  0x00,  0x04,  0x80,  0x49,  0x00,  0x1A,  0x80,  0x40,  0x00,  0x10,  0x80,  0x39,  0x00,  0x11,  0x80,  0x2D,  0x00,  0x15,  0x90,  0x47,  0x50,  0x51,  0x80,  0x47,  0x00,  0x81,  0x22,  0x90,  0x3E,  0x50,  0x00,
 0xB0,  0x40,  0x5E,  0x03,  0x90,  0x3B,  0x50,  0x01,  0x90,  0x45,  0x48,  0x00,  0x90,  0x37,  0x42,  0x00,  0x90,  0x2B,  0x48,  0x05,  0xB0,  0x40,  0x20,  0x02,  0xB0,  0x40,  0x00,  0x5A,  0xB0,  0x40,  0x2E,  0x06,  0xB0,  0x40,  0x7A,  0x04,  0xB0,  0x40,  0x7F,  0x53,  0x80,  0x3B,  0x00,  0x0D,  0x80,  0x45,  0x00,  0x0E,  0x80,  0x37,  0x00,  0x15,  0x80,  0x2B,  0x00,  0x07,  0x80,  0x3E,  0x00,  0x06,  0x90,  0x47,  0x34,  0x06,  0xD0,
 0x02,  0x04,  0xD0,  0x04,  0x04,  0xD0,  0x07,  0x05,  0xD0,  0x1C,  0x02,  0x80,  0x47,  0x00,  0x03,  0xD0,  0x30,  0x0E,  0xD0,  0x2F,  0x12,  0xD0,  0x2E,  0x0A,  0xD0,  0x2D,  0x0A,  0xD0,  0x2A,  0x08,  0xD0,  0x29,  0x06,  0xD0,  0x28,  0x08,  0xD0,  0x26,  0x05,  0xD0,  0x24,  0x07,  0x90,  0x45,  0x56,  0x04,  0xD0,  0x22,  0x04,  0xD0,  0x20,  0x04,  0xD0,  0x1F,  0x05,  0xD0,  0x1E,  0x05,  0xD0,  0x1B,  0x04,  0xD0,  0x19,  0x05,  0xD0,
 0x17,  0x06,  0xD0,  0x14,  0x04,  0xD0,  0x12,  0x04,  0xD0,  0x0F,  0x05,  0xD0,  0x0C,  0x05,  0xD0,  0x0A,  0x05,  0xD0,  0x06,  0x05,  0xD0,  0x03,  0x05,  0xD0,  0x00,  0x2F,  0x90,  0x3D,  0x46,  0x08,  0x90,  0x2A,  0x42,  0x01,  0x90,  0x39,  0x36,  0x02,  0x90,  0x36,  0x34,  0x0B,  0xB0,  0x40,  0x52,  0x06,  0xB0,  0x40,  0x1A,  0x04,  0xB0,  0x40,  0x00,  0x38,  0xD0,  0x01,  0x04,  0xD0,  0x03,  0x05,  0xD0,  0x06,  0x06,  0xD0,  0x08,
 0x04,  0xD0,  0x1A,  0x04,  0xD0,  0x2B,  0x0E,  0xD0,  0x2C,  0x0E,  0xB0,  0x40,  0x4C,  0x03,  0xB0,  0x40,  0x7F,  0x0C,  0xD0,  0x2B,  0x18,  0xD0,  0x2A,  0x05,  0xD0,  0x28,  0x09,  0xD0,  0x27,  0x05,  0xD0,  0x25,  0x05,  0xD0,  0x24,  0x05,  0xD0,  0x22,  0x04,  0xD0,  0x20,  0x04,  0xD0,  0x1C,  0x05,  0xD0,  0x16,  0x05,  0xD0,  0x0D,  0x05,  0xD0,  0x04,  0x03,  0x80,  0x39,  0x00,  0x07,  0xD0,  0x00,  0x05,  0x80,  0x3D,  0x00,  0x08,
 0x80,  0x2A,  0x00,  0x12,  0x80,  0x45,  0x00,  0x04,  0x80,  0x36,  0x00,  0x33,  0xD0,  0x0D,  0x05,  0xD0,  0x1D,  0x05,  0xD0,  0x1B,  0x04,  0xD0,  0x1A,  0x05,  0xD0,  0x19,  0x05,  0xD0,  0x18,  0x05,  0xD0,  0x17,  0x04,  0xD0,  0x15,  0x05,  0xD0,  0x12,  0x05,  0xD0,  0x10,  0x06,  0xD0,  0x0E,  0x04,  0xD0,  0x0C,  0x04,  0xD0,  0x11,  0x05,  0xD0,  0x0D,  0x05,  0xD0,  0x0B,  0x04,  0xD0,  0x08,  0x06,  0xD0,  0x04,  0x04,  0xD0,  0x00,
 0x20,  0x90,  0x36,  0x42,  0x00,  0x90,  0x3D,  0x40,  0x03,  0x90,  0x47,  0x4C,  0x04,  0x90,  0x42,  0x3E,  0x03,  0x90,  0x45,  0x42,  0x81,  0x08,  0x90,  0x49,  0x4E,  0x44,  0x80,  0x49,  0x00,  0x30,  0x90,  0x49,  0x48,  0x48,  0x80,  0x49,  0x00,  0x0E,  0x80,  0x36,  0x00,  0x10,  0x80,  0x3D,  0x00,  0x0F,  0x90,  0x49,  0x44,  0x02,  0x80,  0x42,  0x00,  0x1B,  0x80,  0x45,  0x00,  0x06,  0x80,  0x47,  0x00,  0x1E,  0x80,  0x49,  0x00,
 0x3A,  0x90,  0x49,  0x46,  0x02,  0x90,  0x3D,  0x2E,  0x02,  0x90,  0x36,  0x2E,  0x01,  0x90,  0x45,  0x4A,  0x00,  0x90,  0x42,  0x46,  0x81,  0x6D,  0x80,  0x49,  0x00,  0x04,  0x90,  0x47,  0x4C,  0x27,  0x80,  0x42,  0x00,  0x09,  0x80,  0x3D,  0x00,  0x02,  0x80,  0x36,  0x00,  0x24,  0x80,  0x45,  0x00,  0x44,  0x80,  0x47,  0x00,  0x4B,  0x90,  0x49,  0x56,  0x02,  0x90,  0x45,  0x50,  0x03,  0x90,  0x42,  0x50,  0x02,  0x90,  0x3D,  0x3A,
 0x05,  0x90,  0x36,  0x40,  0x67,  0x80,  0x3D,  0x00,  0x30,  0x80,  0x36,  0x00,  0x4B,  0x90,  0x4C,  0x5C,  0x19,  0x80,  0x42,  0x00,  0x02,  0x80,  0x49,  0x00,  0x27,  0x80,  0x45,  0x00,  0x0A,  0x80,  0x4C,  0x00,  0x19,  0xD0,  0x01,  0x05,  0xD0,  0x03,  0x04,  0xD0,  0x05,  0x07,  0xD0,  0x07,  0x03,  0xD0,  0x18,  0x04,  0xD0,  0x26,  0x0A,  0xD0,  0x25,  0x05,  0xD0,  0x24,  0x04,  0xD0,  0x22,  0x05,  0xD0,  0x21,  0x0A,  0xD0,  0x1F,
 0x05,  0xD0,  0x1C,  0x04,  0xD0,  0x1B,  0x05,  0xD0,  0x19,  0x05,  0xD0,  0x16,  0x04,  0xD0,  0x12,  0x05,  0xD0,  0x11,  0x06,  0xD0,  0x0F,  0x04,  0xD0,  0x0B,  0x05,  0xD0,  0x06,  0x04,  0xD0,  0x02,  0x05,  0xD0,  0x00,  0x27,  0x90,  0x32,  0x44,  0x02,  0x90,  0x42,  0x36,  0x03,  0xB0,  0x40,  0x52,  0x00,  0x90,  0x3E,  0x34,  0x03,  0x90,  0x39,  0x40,  0x04,  0xB0,  0x40,  0x26,  0x00,  0x90,  0x45,  0x52,  0x03,  0xB0,  0x40,  0x00,
 0x09,  0x90,  0x4E,  0x5C,  0x52,  0xD0,  0x01,  0x05,  0xD0,  0x03,  0x05,  0xD0,  0x05,  0x04,  0xD0,  0x16,  0x05,  0xB0,  0x40,  0x04,  0x01,  0xD0,  0x27,  0x04,  0xB0,  0x40,  0x64,  0x05,  0xB0,  0x40,  0x7F,  0x09,  0xD0,  0x25,  0x0E,  0xD0,  0x24,  0x0A,  0xD0,  0x23,  0x05,  0xD0,  0x22,  0x04,  0xD0,  0x20,  0x05,  0xD0,  0x1E,  0x05,  0xD0,  0x1D,  0x05,  0xD0,  0x1C,  0x04,  0xD0,  0x1B,  0x05,  0xD0,  0x19,  0x05,  0xD0,  0x16,  0x04,
 0xD0,  0x15,  0x06,  0xD0,  0x11,  0x04,  0xD0,  0x0E,  0x05,  0xD0,  0x0D,  0x05,  0xD0,  0x09,  0x04,  0xD0,  0x04,  0x06,  0xD0,  0x00,  0x28,  0x80,  0x45,  0x00,  0x02,  0x80,  0x42,  0x00,  0x16,  0x80,  0x39,  0x00,  0x06,  0x80,  0x4E,  0x00,  0x1E,  0x80,  0x3E,  0x00,  0x69,  0x90,  0x45,  0x40,  0x05,  0x90,  0x4A,  0x5A,  0x02,  0x90,  0x3E,  0x30,  0x04,  0x90,  0x39,  0x42,  0x0A,  0x90,  0x4E,  0x52,  0x81,  0x33,  0x80,  0x45,  0x00,
 0x57,  0x80,  0x39,  0x00,  0x06,  0x90,  0x45,  0x1C,  0x03,  0x80,  0x4E,  0x00,  0x09,  0x80,  0x4A,  0x00,  0x10,  0x80,  0x45,  0x00,  0x15,  0x80,  0x3E,  0x00,  0x81,  0x35,  0x90,  0x3E,  0x2E,  0x05,  0x90,  0x4A,  0x58,  0x00,  0x90,  0x39,  0x32,  0x01,  0x90,  0x4E,  0x56,  0x03,  0x90,  0x45,  0x4D,  0x81,  0x72,  0x90,  0x51,  0x5C,  0x1D,  0x80,  0x39,  0x00,  0x5A,  0x80,  0x4A,  0x00,  0x08,  0x80,  0x4E,  0x00,  0x04,  0xD0,  0x01,
 0x00,  0x80,  0x3E,  0x00,  0x06,  0x80,  0x51,  0x00,  0x03,  0xD0,  0x00,  0x70,  0x90,  0x3E,  0x34,  0x07,  0x90,  0x39,  0x38,  0x07,  0x90,  0x45,  0x3F,  0x0A,  0x90,  0x49,  0x4F,  0x31,  0x80,  0x45,  0x00,  0x70,  0x80,  0x49,  0x00,  0x00,  0x80,  0x3E,  0x00,  0x04,  0x80,  0x45,  0x00,  0x14,  0x80,  0x39,  0x00,  0x68,  0x80,  0x32,  0x00,  0x67,  0x90,  0x3E,  0x3E,  0x02,  0x90,  0x32,  0x4C,  0x00,  0x90,  0x42,  0x46,  0x07,  0x90,
 0x26,  0x50,  0x00,  0x90,  0x39,  0x44,  0x0E,  0xB0,  0x40,  0x50,  0x05,  0xB0,  0x40,  0x1C,  0x06,  0xB0,  0x40,  0x00,  0x48,  0xD0,  0x03,  0x05,  0xD0,  0x06,  0x06,  0xD0,  0x09,  0x04,  0xD0,  0x0B,  0x05,  0xD0,  0x21,  0x03,  0xB0,  0x40,  0x2A,  0x02,  0xD0,  0x37,  0x02,  0xB0,  0x40,  0x78,  0x05,  0xB0,  0x40,  0x7F,  0x03,  0xD0,  0x36,  0x08,  0xD0,  0x35,  0x04,  0xD0,  0x34,  0x06,  0xD0,  0x32,  0x06,  0xD0,  0x30,  0x03,  0xD0,
 0x2C,  0x05,  0xD0,  0x28,  0x04,  0xD0,  0x24,  0x05,  0xD0,  0x1E,  0x05,  0xD0,  0x17,  0x02,  0x80,  0x39,  0x00,  0x03,  0xD0,  0x0F,  0x04,  0x80,  0x3E,  0x00,  0x01,  0xD0,  0x05,  0x01,  0x80,  0x26,  0x00,  0x03,  0xD0,  0x00,  0x0C,  0x80,  0x32,  0x00,  0x06,  0x80,  0x42,  0x00,  0x82,  0x24,  0x90,  0x32,  0x2E,  0x03,  0x90,  0x39,  0x3C,  0x01,  0x90,  0x42,  0x4A,  0x00,  0x90,  0x3E,  0x38,  0x03,  0x90,  0x45,  0x4A,  0x01,  0x90,
 0x49,  0x4C,  0x55,  0x80,  0x49,  0x00,  0x26,  0x90,  0x49,  0x4C,  0x53,  0x80,  0x49,  0x00,  0x2A,  0x90,  0x49,  0x4A,  0x46,  0x80,  0x39,  0x00,  0x06,  0x80,  0x49,  0x00,  0x28,  0x90,  0x49,  0x46,  0x1B,  0x80,  0x42,  0x00,  0x03,  0x80,  0x3E,  0x00,  0x14,  0x80,  0x45,  0x00,  0x1A,  0x80,  0x49,  0x00,  0x37,  0x90,  0x39,  0x3C,  0x03,  0x90,  0x3E,  0x2E,  0x00,  0x90,  0x49,  0x4A,  0x00,  0x90,  0x45,  0x4E,  0x04,  0x90,  0x42,
 0x4C,  0x81,  0x75,  0x80,  0x42,  0x00,  0x03,  0x90,  0x47,  0x4C,  0x1B,  0x80,  0x39,  0x00,  0x0F,  0x80,  0x49,  0x00,  0x1A,  0x80,  0x45,  0x00,  0x0D,  0x80,  0x3E,  0x00,  0x81,  0x25,  0x90,  0x3E,  0x3A,  0x03,  0x90,  0x45,  0x56,  0x00,  0x90,  0x39,  0x3C,  0x04,  0x90,  0x40,  0x58,  0x1A,  0x80,  0x47,  0x00,  0x4A,  0xD0,  0x03,  0x14,  0xD0,  0x02,  0x13,  0xD0,  0x01,  0x09,  0xD0,  0x02,  0x0A,  0xD0,  0x01,  0x0A,  0xD0,  0x00,
 0x02,  0x80,  0x3E,  0x00,  0x03,  0xD0,  0x06,  0x09,  0x80,  0x39,  0x00,  0x04,  0xD0,  0x0C,  0x05,  0xD0,  0x0A,  0x05,  0xD0,  0x08,  0x05,  0xD0,  0x06,  0x05,  0xD0,  0x04,  0x04,  0xD0,  0x03,  0x03,  0x80,  0x32,  0x00,  0x02,  0xD0,  0x0B,  0x04,  0xD0,  0x0A,  0x05,  0xD0,  0x09,  0x05,  0xD0,  0x0A,  0x01,  0x80,  0x40,  0x00,  0x03,  0xD0,  0x14,  0x09,  0x90,  0x47,  0x5F,  0x01,  0xD0,  0x0C,  0x09,  0xD0,  0x0D,  0x23,  0xD0,  0x0C,
 0x03,  0xD0,  0x0B,  0x0A,  0xD0,  0x09,  0x0B,  0xD0,  0x08,  0x03,  0xD0,  0x07,  0x05,  0xD0,  0x06,  0x04,  0xD0,  0x04,  0x0A,  0xD0,  0x03,  0x05,  0xD0,  0x00,  0x1D,  0x90,  0x49,  0x40,  0x1C,  0x90,  0x2D,  0x40,  0x03,  0xB0,  0x40,  0x56,  0x01,  0x90,  0x3D,  0x50,  0x04,  0xB0,  0x40,  0x1C,  0x04,  0x90,  0x39,  0x2E,  0x01,  0x90,  0x34,  0x3C,  0x00,  0xB0,  0x40,  0x00,  0x56,  0xB0,  0x40,  0x18,  0x05,  0xB0,  0x40,  0x6E,  0x05,
 0xB0,  0x40,  0x7F,  0x03,  0xD0,  0x08,  0x0E,  0xD0,  0x07,  0x06,  0xD0,  0x05,  0x08,  0xD0,  0x04,  0x06,  0xD0,  0x03,  0x08,  0xD0,  0x01,  0x05,  0xD0,  0x00,  0x51,  0x80,  0x3D,  0x00,  0x09,  0x80,  0x34,  0x00,  0x0A,  0x80,  0x2D,  0x00,  0x14,  0x80,  0x39,  0x00,  0x12,  0x80,  0x45,  0x00,  0x12,  0x80,  0x47,  0x00,  0x0E,  0x80,  0x49,  0x00,  0x81,  0x34,  0x90,  0x2D,  0x3C,  0x02,  0x90,  0x40,  0x54,  0x03,  0x90,  0x3D,  0x4A,
 0x00,  0x90,  0x39,  0x2E,  0x02,  0x90,  0x49,  0x52,  0x03,  0x90,  0x34,  0x36,  0x48,  0x80,  0x49,  0x00,  0x27,  0x90,  0x49,  0x4E,  0x48,  0x80,  0x49,  0x00,  0x31,  0x90,  0x49,  0x40,  0x41,  0x80,  0x49,  0x00,  0x34,  0x90,  0x49,  0x40,  0x06,  0x80,  0x34,  0x00,  0x13,  0x80,  0x39,  0x00,  0x1C,  0x80,  0x3D,  0x00,  0x19,  0x80,  0x49,  0x00,  0x09,  0x80,  0x40,  0x00,  0x2E,  0x90,  0x40,  0x4C,  0x03,  0x90,  0x3D,  0x50,  0x02,
 0x90,  0x39,  0x3A,  0x04,  0x90,  0x49,  0x5A,  0x03,  0x90,  0x34,  0x30,  0x76,  0x80,  0x40,  0x00,  0x04,  0x80,  0x3D,  0x00,  0x07,  0x80,  0x49,  0x00,  0x3D,  0x80,  0x34,  0x00,  0x05,  0x80,  0x39,  0x00,  0x11,  0x80,  0x2D,  0x00,  0x15,  0x90,  0x47,  0x52,  0x58,  0x80,  0x47,  0x00,  0x81,  0x15,  0x90,  0x3E,  0x58,  0x03,  0x90,  0x45,  0x52,  0x02,  0x90,  0x3B,  0x4E,  0x00,  0x90,  0x2B,  0x46,  0x00,  0x90,  0x37,  0x3E,  0x03,
 0x90,  0x32,  0x4A,  0x05,  0xB0,  0x40,  0x52,  0x05,  0xB0,  0x40,  0x1A,  0x05,  0xB0,  0x40,  0x00,  0x50,  0xD0,  0x05,  0x04,  0xD0,  0x13,  0x02,  0xB0,  0x40,  0x54,  0x04,  0xD0,  0x12,  0x02,  0xB0,  0x40,  0x7F,  0x16,  0xD0,  0x11,  0x04,  0xD0,  0x10,  0x04,  0xD0,  0x0D,  0x05,  0xD0,  0x0C,  0x05,  0xD0,  0x09,  0x06,  0xD0,  0x05,  0x03,  0xD0,  0x00,  0x27,  0x80,  0x37,  0x00,  0x0A,  0x80,  0x3B,  0x00,  0x17,  0x80,  0x32,  0x00,
 0x02,  0x80,  0x2B,  0x00,  0x07,  0x90,  0x47,  0x52,  0x07,  0x80,  0x45,  0x00,  0x03,  0x80,  0x3E,  0x00,  0x17,  0x80,  0x47,  0x00,  0x57,  0x90,  0x45,  0x5A,  0x73,  0x90,  0x3D,  0x46,  0x03,  0x90,  0x42,  0x42,  0x04,  0x90,  0x2A,  0x38,  0x03,  0x90,  0x36,  0x3C,  0x00,  0x90,  0x39,  0x42,  0x03,  0x90,  0x31,  0x3C,  0x05,  0xB0,  0x40,  0x58,  0x05,  0xB0,  0x40,  0x1C,  0x04,  0xB0,  0x40,  0x00,  0x2D,  0xD0,  0x03,  0x05,  0xD0,
 0x06,  0x05,  0xD0,  0x09,  0x05,  0xD0,  0x0B,  0x05,  0xD0,  0x0E,  0x05,  0xD0,  0x10,  0x04,  0xD0,  0x13,  0x05,  0xD0,  0x15,  0x04,  0xD0,  0x26,  0x05,  0xB0,  0x40,  0x50,  0x00,  0xD0,  0x36,  0x05,  0xB0,  0x40,  0x7F,  0x13,  0xD0,  0x35,  0x05,  0xD0,  0x34,  0x1D,  0xD0,  0x33,  0x05,  0xD0,  0x32,  0x07,  0xD0,  0x2F,  0x05,  0xD0,  0x2D,  0x04,  0xD0,  0x2C,  0x05,  0xD0,  0x28,  0x05,  0xD0,  0x25,  0x05,  0xD0,  0x23,  0x05,  0xD0,
 0x20,  0x04,  0xD0,  0x1C,  0x05,  0xD0,  0x17,  0x04,  0xD0,  0x11,  0x06,  0xD0,  0x0B,  0x04,  0xD0,  0x03,  0x05,  0xD0,  0x00,  0x0C,  0x80,  0x39,  0x00,  0x0C,  0x80,  0x2A,  0x00,  0x08,  0x80,  0x3D,  0x00,  0x0A,  0x80,  0x31,  0x00,  0x02,  0x80,  0x42,  0x00,  0x0D,  0x80,  0x45,  0x00,  0x03,  0x80,  0x36,  0x00,  0x81,  0x10,  0x90,  0x3D,  0x44,  0x03,  0x90,  0x36,  0x44,  0x02,  0x90,  0x47,  0x56,  0x00,  0x90,  0x42,  0x4C,  0x07,
 0x90,  0x45,  0x42,  0x79,  0x90,  0x49,  0x52,  0x4C,  0x80,  0x49,  0x00,  0x27,  0x90,  0x49,  0x4E,  0x52,  0x80,  0x49,  0x00,  0x09,  0x80,  0x42,  0x00,  0x19,  0x80,  0x36,  0x00,  0x06,  0x90,  0x49,  0x40,  0x0D,  0x80,  0x3D,  0x00,  0x09,  0x80,  0x45,  0x00,  0x10,  0x80,  0x47,  0x00,  0x10,  0x80,  0x49,  0x00,  0x41,  0x90,  0x3D,  0x36,  0x04,  0x90,  0x49,  0x4E,  0x02,  0x90,  0x36,  0x2C,  0x03,  0x90,  0x42,  0x46,  0x02,  0x90,
 0x45,  0x42,  0x81,  0x72,  0x90,  0x47,  0x4C,  0x08,  0x80,  0x49,  0x00,  0x35,  0x80,  0x42,  0x00,  0x05,  0x80,  0x3D,  0x00,  0x19,  0x80,  0x36,  0x00,  0x03,  0x80,  0x45,  0x00,  0x4F,  0x80,  0x47,  0x00,  0x3F,  0x90,  0x49,  0x52,  0x01,  0x90,  0x3D,  0x3C,  0x03,  0x90,  0x45,  0x54,  0x00,  0x90,  0x42,  0x52,  0x00,  0x90,  0x36,  0x38,  0x76,  0x80,  0x3D,  0x00,  0x35,  0x80,  0x36,  0x00,  0x44,  0x90,  0x4C,  0x58,  0x15,  0x80,
 0x42,  0x00,  0x15,  0x80,  0x49,  0x00,  0x0C,  0x80,  0x45,  0x00,  0x19,  0x80,  0x4C,  0x00,  0x81,  0x15,  0x90,  0x32,  0x40,  0x02,  0x90,  0x42,  0x3E,  0x0D,  0x90,  0x39,  0x3E,  0x05,  0x90,  0x3E,  0x34,  0x06,  0xB0,  0x40,  0x52,  0x01,  0x90,  0x45,  0x4E,  0x04,  0xB0,  0x40,  0x18,  0x04,  0xB0,  0x40,  0x00,  0x07,  0x90,  0x4E,  0x5A,  0x54,  0xB0,  0x40,  0x4E,  0x00,  0xD0,  0x02,  0x03,  0xD0,  0x04,  0x02,  0xB0,  0x40,  0x7F,
 0x03,  0xD0,  0x16,  0x05,  0xD0,  0x27,  0x0E,  0xD0,  0x26,  0x0A,  0xD0,  0x25,  0x13,  0xD0,  0x24,  0x05,  0xD0,  0x23,  0x09,  0xD0,  0x21,  0x0A,  0xD0,  0x20,  0x09,  0xD0,  0x1E,  0x07,  0x80,  0x42,  0x00,  0x03,  0xD0,  0x25,  0x04,  0xD0,  0x24,  0x05,  0xD0,  0x23,  0x05,  0xD0,  0x20,  0x04,  0xD0,  0x1D,  0x05,  0xD0,  0x1A,  0x05,  0xD0,  0x17,  0x05,  0xD0,  0x13,  0x04,  0xD0,  0x0D,  0x05,  0xD0,  0x05,  0x04,  0xD0,  0x00,  0x05,
 0x90,  0x42,  0x28,  0x1E,  0x80,  0x45,  0x00,  0x10,  0x80,  0x42,  0x00,  0x14,  0x80,  0x4E,  0x00,  0x03,  0x80,  0x39,  0x00,  0x2E,  0x80,  0x3E,  0x00,  0x47,  0x90,  0x45,  0x3A,  0x04,  0x90,  0x4A,  0x54,  0x04,  0x90,  0x3E,  0x32,  0x04,  0x90,  0x39,  0x36,  0x06,  0x90,  0x4E,  0x52,  0x63,  0xD0,  0x01,  0x0A,  0xD0,  0x00,  0x59,  0x80,  0x45,  0x00,  0x48,  0x80,  0x39,  0x00,  0x04,  0x90,  0x45,  0x15,  0x01,  0x80,  0x4E,  0x00,
 0x05,  0x80,  0x4A,  0x00,  0x15,  0x80,  0x45,  0x00,  0x02,  0x80,  0x3E,  0x00,  0x81,  0x35,  0x90,  0x45,  0x38,  0x0F,  0x90,  0x3E,  0x1F,  0x01,  0x90,  0x4A,  0x50,  0x03,  0x90,  0x39,  0x22,  0x07,  0x90,  0x4E,  0x58,  0x81,  0x64,  0x90,  0x51,  0x5E,  0x0E,  0x80,  0x39,  0x00,  0x28,  0x80,  0x3E,  0x00,  0x81,  0x00,  0x80,  0x51,  0x00,  0x0A,  0x80,  0x4E,  0x00,  0x01,  0x80,  0x4A,  0x00,  0x3C,  0x90,  0x3E,  0x21,  0x0C,  0x90,
 0x39,  0x1F,  0x21,  0x80,  0x45,  0x00,  0x50,  0x90,  0x49,  0x50,  0x3D,  0x80,  0x3E,  0x00,  0x05,  0x80,  0x39,  0x00,  0x2D,  0x80,  0x49,  0x00,  0x04,  0x90,  0x47,  0x4E,  0x53,  0x80,  0x32,  0x00,  0x6E,  0x90,  0x3E,  0x28,  0x06,  0x90,  0x41,  0x30,  0x02,  0xB0,  0x40,  0x50,  0x02,  0x90,  0x29,  0x20,  0x00,  0x90,  0x35,  0x2A,  0x02,  0xB0,  0x40,  0x20,  0x05,  0xB0,  0x40,  0x00,  0x29,  0xB0,  0x40,  0x00,  0x84,  0x08,  0xB0,
 0x40,  0x0C,  0x05,  0xB0,  0x40,  0x2C,  0x05,  0xB0,  0x40,  0x78,  0x04,  0xB0,  0x40,  0x7F,  0x82,  0x42,  0x80,  0x3E,  0x00,  0x0C,  0x80,  0x41,  0x00,  0x11,  0x80,  0x29,  0x00,  0x00,  0x80,  0x35,  0x00,  0x5A,  0x80,  0x47,  0x00,  0x81,  0x46,  0x90,  0x47,  0x45,  0x52,  0x80,  0x47,  0x00,  0x08,  0x90,  0x45,  0x39,  0x66,  0x90,  0x47,  0x3B,  0x03,  0x80,  0x45,  0x00,  0x14,  0xB0,  0x40,  0x7F,  0x12,  0x90,  0x49,  0x41,  0x03,
 0x80,  0x47,  0x00,  0x81,  0x4F,  0x90,  0x45,  0x3F,  0x12,  0x80,  0x49,  0x00,  0x22,  0x80,  0x45,  0x00,  0x3D,  0x90,  0x45,  0x33,  0x81,  0x03,  0x90,  0x39,  0x30,  0x05,  0x90,  0x2A,  0x36,  0x03,  0x90,  0x31,  0x34,  0x04,  0x90,  0x3D,  0x46,  0x00,  0x90,  0x2D,  0x2E,  0x00,  0x90,  0x34,  0x30,  0x02,  0x90,  0x39,  0x42,  0x12,  0x80,  0x45,  0x00,  0x41,  0xB0,  0x40,  0x54,  0x05,  0xB0,  0x40,  0x7F,  0x81,  0x0B,  0x80,  0x2A,
 0x00,  0x04,  0x80,  0x31,  0x00,  0x09,  0x90,  0x40,  0x4E,  0x04,  0x80,  0x3D,  0x00,  0x03,  0x80,  0x34,  0x00,  0x0D,  0x80,  0x2D,  0x00,  0x1A,  0x80,  0x39,  0x00,  0x32,  0x80,  0x39,  0x00,  0x0F,  0x90,  0x3D,  0x42,  0x07,  0x80,  0x40,  0x00,  0x7B,  0x90,  0x31,  0x36,  0x06,  0x90,  0x2A,  0x28,  0x00,  0x90,  0x39,  0x42,  0x03,  0x90,  0x36,  0x3E,  0x01,  0xB0,  0x40,  0x52,  0x05,  0xB0,  0x40,  0x18,  0x06,  0xB0,  0x40,  0x00,
 0x4A,  0x80,  0x3D,  0x00,  0x10,  0xB0,  0x40,  0x12,  0x04,  0xB0,  0x40,  0x6C,  0x05,  0xB0,  0x40,  0x7F,  0x78,  0x90,  0x3D,  0x4A,  0x23,  0x80,  0x2A,  0x00,  0x01,  0x80,  0x36,  0x00,  0x09,  0x80,  0x31,  0x00,  0x06,  0x80,  0x39,  0x00,  0x33,  0x80,  0x3D,  0x00,  0x81,  0x01,  0xB0,  0x40,  0x54,  0x06,  0xB0,  0x40,  0x1A,  0x04,  0xB0,  0x40,  0x00,  0x01,  0x90,  0x2D,  0x30,  0x00,  0x90,  0x34,  0x3C,  0x03,  0x90,  0x3D,  0x48,
 0x05,  0x90,  0x39,  0x46,  0x49,  0xB0,  0x40,  0x28,  0x04,  0xB0,  0x40,  0x76,  0x06,  0xB0,  0x40,  0x7F,  0x81,  0x13,  0x80,  0x34,  0x00,  0x05,  0x90,  0x40,  0x58,  0x16,  0x80,  0x3D,  0x00,  0x07,  0x80,  0x2D,  0x00,  0x07,  0x80,  0x39,  0x00,  0x47,  0x90,  0x3D,  0x4C,  0x06,  0x80,  0x40,  0x00,  0x74,  0xB0,  0x40,  0x52,  0x03,  0x90,  0x25,  0x2C,  0x02,  0xB0,  0x40,  0x1A,  0x06,  0xB0,  0x40,  0x00,  0x00,  0x90,  0x38,  0x38,
 0x00,  0x90,  0x2C,  0x30,  0x0B,  0x90,  0x34,  0x2C,  0x5E,  0xB0,  0x40,  0x5E,  0x04,  0xB0,  0x40,  0x7F,  0x01,  0xD0,  0x05,  0x0A,  0xD0,  0x07,  0x04,  0xD0,  0x06,  0x18,  0xD0,  0x07,  0x05,  0xD0,  0x06,  0x05,  0xD0,  0x05,  0x04,  0xD0,  0x06,  0x0A,  0xD0,  0x04,  0x18,  0xD0,  0x03,  0x04,  0xD0,  0x01,  0x05,  0xD0,  0x00,  0x2A,  0x80,  0x34,  0x00,  0x06,  0x80,  0x38,  0x00,  0x03,  0x80,  0x2C,  0x00,  0x0F,  0x80,  0x25,  0x00,
 0x09,  0x80,  0x3D,  0x00,  0x81,  0x43,  0xB0,  0x40,  0x50,  0x03,  0x90,  0x2D,  0x36,  0x01,  0xB0,  0x40,  0x18,  0x05,  0xB0,  0x40,  0x00,  0x00,  0x90,  0x26,  0x2E,  0x02,  0x90,  0x39,  0x42,  0x03,  0x90,  0x3D,  0x3E,  0x01,  0x90,  0x36,  0x42,  0x81,  0x27,  0xB0,  0x40,  0x1C,  0x05,  0xB0,  0x40,  0x70,  0x05,  0xB0,  0x40,  0x7F,  0x40,  0x90,  0x3B,  0x48,  0x1E,  0x80,  0x3D,  0x00,  0x10,  0x80,  0x26,  0x00,  0x08,  0x80,  0x36,
 0x00,  0x02,  0x80,  0x2D,  0x00,  0x07,  0x80,  0x39,  0x00,  0x0A,  0x80,  0x3B,  0x00,  0x2C,  0x90,  0x3B,  0x44,  0x81,  0x07,  0xB0,  0x40,  0x52,  0x05,  0xB0,  0x40,  0x18,  0x01,  0x90,  0x31,  0x22,  0x03,  0x90,  0x39,  0x42,  0x01,  0xB0,  0x40,  0x00,  0x00,  0x90,  0x2A,  0x21,  0x02,  0x90,  0x36,  0x40,  0x0F,  0x80,  0x3B,  0x00,  0x83,  0x24,  0xB0,  0x40,  0x1A,  0x68,  0x80,  0x31,  0x00,  0x06,  0x80,  0x2A,  0x00,  0x02,  0x80,
 0x36,  0x00,  0x02,  0x80,  0x39,  0x00,  0x86,  0x50,  0x90,  0x2A,  0x32,  0x00,  0x90,  0x36,  0x2F,  0x00,  0x90,  0x39,  0x3D,  0x00,  0x90,  0x3D,  0x2A,  0x7B,  0xB0,  0x40,  0x7F,  0x82,  0x33,  0x80,  0x2A,  0x00,  0x14,  0xB0,  0x40,  0x00,  0x00,  0x80,  0x39,  0x00,  0x0D,  0x80,  0x3D,  0x00,  0x06,  0x80,  0x36,  0x00,  0x0B,  0x90,  0x2D,  0x32,  0x00,  0x90,  0x34,  0x2A,  0x00,  0x90,  0x39,  0x37,  0x00,  0x90,  0x3D,  0x35,  0x74,
 0xB0,  0x40,  0x7F,  0x82,  0x49,  0x80,  0x2D,  0x00,  0x07,  0x80,  0x34,  0x00,  0x06,  0x80,  0x3D,  0x00,  0x01,  0x80,  0x39,  0x00,  0x09,  0xB0,  0x40,  0x00,  0x0C,  0x90,  0x2A,  0x37,  0x00,  0x90,  0x36,  0x35,  0x00,  0x90,  0x39,  0x36,  0x00,  0x90,  0x3D,  0x30,  0x63,  0xB0,  0x40,  0x7F,  0x82,  0x63,  0x80,  0x2A,  0x00,  0x02,  0xB0,  0x40,  0x00,  0x08,  0x80,  0x3D,  0x00,  0x02,  0x80,  0x39,  0x00,  0x03,  0x80,  0x36,  0x00,
 0x0B,  0x90,  0x2D,  0x30,  0x00,  0x90,  0x34,  0x30,  0x00,  0x90,  0x39,  0x3B,  0x00,  0x90,  0x3D,  0x2F,  0x6C,  0xB0,  0x40,  0x7F,  0x82,  0x47,  0x80,  0x2D,  0x00,  0x06,  0x80,  0x34,  0x00,  0x00,  0x80,  0x39,  0x00,  0x00,  0x80,  0x3D,  0x00,  0x11,  0xB0,  0x40,  0x00,  0x16,  0x90,  0x25,  0x31,  0x00,  0x90,  0x31,  0x3C,  0x00,  0x90,  0x34,  0x28,  0x00,  0x90,  0x38,  0x32,  0x60,  0xB0,  0x40,  0x7F,  0x82,  0x33,  0x80,  0x25,
 0x00,  0x35,  0x80,  0x31,  0x00,  0x04,  0x80,  0x34,  0x00,  0x04,  0x80,  0x38,  0x00,  0x08,  0xB0,  0x40,  0x00,  0x08,  0x90,  0x26,  0x2E,  0x00,  0x90,  0x32,  0x2D,  0x00,  0x90,  0x36,  0x2D,  0x00,  0x90,  0x39,  0x2D,  0x83,  0x2A,  0x80,  0x26,  0x00,  0x0C,  0x80,  0x39,  0x00,  0x03,  0x80,  0x32,  0x00,  0x04,  0x80,  0x36,  0x00,  0x23,  0x90,  0x2A,  0x32,  0x00,  0x90,  0x36,  0x2F,  0x00,  0x90,  0x39,  0x3D,  0x00,  0x90,  0x3D,
 0x2A,  0x86,  0x7F,  0x80,  0x2A,  0x00,  0x14,  0x80,  0x39,  0x00,  0x0D,  0x80,  0x3D,  0x00,  0x06,  0x80,  0x36,  0x00,  0x1A,  0x90,  0x28,  0x50,  0x00,  0x90,  0x45,  0x40,  0x00,  0x90,  0x40,  0x2A,  0x01,  0x90,  0x34,  0x4A,  0x02,  0x90,  0x49,  0x4C,  0x0C,  0xB0,  0x40,  0x50,  0x04,  0xB0,  0x40,  0x28,  0x06,  0xB0,  0x40,  0x01,  0x04,  0xB0,  0x40,  0x00,  0x2F,  0xD0,  0x02,  0x05,  0xD0,  0x05,  0x04,  0xD0,  0x07,  0x05,  0xD0,
 0x09,  0x06,  0xD0,  0x0B,  0x04,  0xD0,  0x1B,  0x04,  0xD0,  0x2D,  0x05,  0xD0,  0x2E,  0x05,  0xD0,  0x2C,  0x06,  0xD0,  0x29,  0x04,  0xD0,  0x2A,  0x01,  0xB0,  0x40,  0x18,  0x03,  0xD0,  0x2C,  0x02,  0xB0,  0x40,  0x6E,  0x03,  0xD0,  0x29,  0x02,  0xB0,  0x40,  0x7F,  0x0C,  0xD0,  0x27,  0x05,  0xD0,  0x25,  0x05,  0xD0,  0x22,  0x09,  0xD0,  0x24,  0x05,  0xD0,  0x22,  0x05,  0xD0,  0x1E,  0x09,  0xD0,  0x1F,  0x0F,  0xD0,  0x1A,  0x04,
 0xD0,  0x17,  0x05,  0xD0,  0x16,  0x04,  0xD0,  0x13,  0x05,  0xD0,  0x11,  0x05,  0xD0,  0x0F,  0x05,  0xD0,  0x0D,  0x04,  0x80,  0x28,  0x00,  0x01,  0xD0,  0x13,  0x05,  0xD0,  0x11,  0x04,  0xD0,  0x0F,  0x05,  0xD0,  0x0E,  0x09,  0xD0,  0x0C,  0x01,  0x80,  0x34,  0x00,  0x04,  0xD0,  0x13,  0x09,  0xD0,  0x11,  0x06,  0xD0,  0x0F,  0x03,  0xD0,  0x0E,  0x05,  0xD0,  0x0C,  0x05,  0xD0,  0x0B,  0x05,  0xD0,  0x0A,  0x09,  0xD0,  0x06,  0x05,
 0xD0,  0x03,  0x05,  0xD0,  0x01,  0x05,  0xD0,  0x00,  0x2E,  0x80,  0x40,  0x00,  0x08,  0x90,  0x47,  0x52,  0x16,  0x80,  0x49,  0x00,  0x02,  0x80,  0x45,  0x00,  0x5A,  0x90,  0x45,  0x46,  0x0A,  0x90,  0x3B,  0x32,  0x02,  0x90,  0x40,  0x3C,  0x00,  0x90,  0x34,  0x28,  0x03,  0x90,  0x39,  0x3A,  0x05,  0x90,  0x49,  0x50,  0x28,  0x80,  0x47,  0x00,  0x82,  0x1A,  0x80,  0x39,  0x00,  0x0F,  0x80,  0x3B,  0x00,  0x04,  0x80,  0x34,  0x00,
 0x1B,  0x80,  0x49,  0x00,  0x04,  0x90,  0x47,  0x46,  0x19,  0x80,  0x40,  0x00,  0x0C,  0x80,  0x45,  0x00,  0x4C,  0x90,  0x45,  0x46,  0x03,  0x90,  0x40,  0x44,  0x0C,  0x90,  0x3B,  0x2C,  0x02,  0x90,  0x34,  0x30,  0x00,  0x90,  0x39,  0x34,  0x11,  0x90,  0x49,  0x50,  0x82,  0x36,  0x80,  0x3B,  0x00,  0x05,  0x80,  0x39,  0x00,  0x01,  0x80,  0x34,  0x00,  0x1F,  0x90,  0x4C,  0x5E,  0x09,  0x80,  0x47,  0x00,  0x06,  0x80,  0x49,  0x00,
 0x16,  0x80,  0x45,  0x00,  0x09,  0x80,  0x40,  0x00,  0x44,  0x90,  0x47,  0x4A,  0x03,  0x80,  0x4C,  0x00,  0x05,  0x90,  0x40,  0x42,  0x03,  0x90,  0x3B,  0x38,  0x00,  0x90,  0x45,  0x3C,  0x04,  0x90,  0x39,  0x3C,  0x01,  0x90,  0x34,  0x26,  0x16,  0x90,  0x49,  0x4C,  0x1B,  0x80,  0x47,  0x00,  0x27,  0x80,  0x45,  0x00,  0x1E,  0x80,  0x39,  0x00,  0x07,  0x80,  0x3B,  0x00,  0x07,  0x90,  0x47,  0x32,  0x08,  0x80,  0x34,  0x00,  0x3B,
 0x80,  0x49,  0x00,  0x64,  0x80,  0x40,  0x00,  0x81,  0x0F,  0x80,  0x47,  0x00,  0x11,  0x90,  0x47,  0x18,  0x07,  0x90,  0x28,  0x4A,  0x04,  0x90,  0x45,  0x44,  0x02,  0x90,  0x40,  0x42,  0x04,  0x90,  0x34,  0x48,  0x02,  0xB0,  0x40,  0x50,  0x01,  0x90,  0x49,  0x50,  0x02,  0xB0,  0x40,  0x1A,  0x06,  0xB0,  0x40,  0x00,  0x4B,  0x80,  0x47,  0x00,  0x0C,  0xD0,  0x01,  0x00,  0xB0,  0x40,  0x16,  0x04,  0xB0,  0x40,  0x6E,  0x01,  0xD0,
 0x10,  0x03,  0xB0,  0x40,  0x7F,  0x06,  0xD0,  0x0E,  0x0A,  0xD0,  0x0F,  0x08,  0xD0,  0x0D,  0x05,  0xD0,  0x0C,  0x0B,  0xD0,  0x0A,  0x08,  0xD0,  0x09,  0x05,  0xD0,  0x07,  0x0B,  0xD0,  0x05,  0x08,  0xD0,  0x02,  0x06,  0xD0,  0x01,  0x08,  0xD0,  0x00,  0x32,  0x80,  0x28,  0x00,  0x2E,  0x80,  0x34,  0x00,  0x50,  0x80,  0x40,  0x00,  0x03,  0x80,  0x45,  0x00,  0x03,  0x90,  0x47,  0x54,  0x1F,  0x80,  0x49,  0x00,  0x05,  0x90,  0x45,
 0x4A,  0x0A,  0x90,  0x34,  0x32,  0x02,  0x90,  0x40,  0x42,  0x03,  0x90,  0x3B,  0x2C,  0x01,  0x90,  0x39,  0x38,  0x02,  0x90,  0x49,  0x4E,  0x81,  0x0F,  0x80,  0x47,  0x00,  0x81,  0x49,  0x80,  0x39,  0x00,  0x01,  0x80,  0x34,  0x00,  0x15,  0x80,  0x3B,  0x00,  0x04,  0x90,  0x47,  0x3E,  0x10,  0x80,  0x40,  0x00,  0x06,  0x80,  0x49,  0x00,  0x16,  0x80,  0x45,  0x00,  0x5D,  0x90,  0x34,  0x2A,  0x01,  0x90,  0x3B,  0x30,  0x03,  0x90,
 0x45,  0x40,  0x02,  0x90,  0x40,  0x3E,  0x01,  0x90,  0x39,  0x32,  0x01,  0x90,  0x49,  0x50,  0x45,  0x80,  0x47,  0x00,  0x81,  0x3E,  0x80,  0x40,  0x00,  0x2A,  0x80,  0x45,  0x00,  0x36,  0x80,  0x49,  0x00,  0x08,  0x90,  0x40,  0x46,  0x81,  0x00,  0x90,  0x45,  0x2A,  0x02,  0x90,  0x47,  0x50,  0x19,  0x90,  0x49,  0x5C,  0x0F,  0x80,  0x40,  0x00,  0x02,  0x80,  0x45,  0x00,  0x06,  0x80,  0x47,  0x00,  0x53,  0x90,  0x47,  0x50,  0x2F,
 0x80,  0x49,  0x00,  0x25,  0x80,  0x39,  0x00,  0x09,  0x80,  0x3B,  0x00,  0x1B,  0x80,  0x34,  0x00,  0x07,  0x90,  0x45,  0x4C,  0x07,  0x80,  0x47,  0x00,  0x35,  0x80,  0x45,  0x00,  0x3E,  0x90,  0x44,  0x4E,  0x1A,  0x80,  0x44,  0x00,  0x08,  0x90,  0x26,  0x4A,  0x01,  0x90,  0x42,  0x54,  0x00,  0x90,  0x32,  0x4E,  0x04,  0x90,  0x3E,  0x58,  0x02,  0xB0,  0x40,  0x5E,  0x05,  0xB0,  0x40,  0x1E,  0x00,  0x90,  0x39,  0x4A,  0x04,  0xB0,
 0x40,  0x00,  0x59,  0xD0,  0x0F,  0x04,  0xD0,  0x21,  0x0B,  0xD0,  0x1E,  0x02,  0xB0,  0x40,  0x18,  0x03,  0xD0,  0x1F,  0x02,  0xB0,  0x40,  0x6E,  0x02,  0xD0,  0x22,  0x03,  0xB0,  0x40,  0x7F,  0x01,  0xD0,  0x1F,  0x05,  0xD0,  0x1C,  0x06,  0xD0,  0x1B,  0x04,  0xD0,  0x1C,  0x04,  0xD0,  0x1D,  0x05,  0xD0,  0x1B,  0x05,  0xD0,  0x19,  0x0E,  0xD0,  0x15,  0x05,  0xD0,  0x14,  0x05,  0xD0,  0x15,  0x04,  0xD0,  0x14,  0x05,  0xD0,  0x12,
 0x05,  0xD0,  0x0F,  0x04,  0xD0,  0x0C,  0x04,  0xD0,  0x04,  0x05,  0xD0,  0x00,  0x00,  0x80,  0x26,  0x00,  0x0F,  0x80,  0x39,  0x00,  0x0E,  0x80,  0x3E,  0x00,  0x01,  0x80,  0x32,  0x00,  0x15,  0x80,  0x42,  0x00,  0x81,  0x6C,  0x90,  0x32,  0x56,  0x04,  0x90,  0x39,  0x4C,  0x03,  0x90,  0x45,  0x4E,  0x00,  0x90,  0x49,  0x52,  0x00,  0x90,  0x42,  0x4C,  0x04,  0x90,  0x3E,  0x48,  0x55,  0x80,  0x49,  0x00,  0x2C,  0x90,  0x49,  0x4A,
 0x4D,  0x80,  0x49,  0x00,  0x27,  0x90,  0x49,  0x52,  0x2A,  0x80,  0x42,  0x00,  0x27,  0x80,  0x49,  0x00,  0x00,  0x80,  0x39,  0x00,  0x01,  0x80,  0x32,  0x00,  0x17,  0x80,  0x3E,  0x00,  0x0D,  0x90,  0x49,  0x50,  0x0F,  0x80,  0x45,  0x00,  0x33,  0x80,  0x49,  0x00,  0x1A,  0x90,  0x49,  0x52,  0x02,  0x90,  0x32,  0x50,  0x04,  0x90,  0x42,  0x50,  0x00,  0x90,  0x45,  0x50,  0x00,  0x90,  0x3E,  0x3E,  0x00,  0x90,  0x39,  0x4E,  0x81,
 0x7D,  0x90,  0x47,  0x54,  0x0B,  0x80,  0x42,  0x00,  0x1A,  0x80,  0x45,  0x00,  0x19,  0x80,  0x49,  0x00,  0x21,  0x80,  0x32,  0x00,  0x01,  0x80,  0x39,  0x00,  0x26,  0x80,  0x3E,  0x00,  0x45,  0x80,  0x47,  0x00,  0x32,  0x90,  0x32,  0x56,  0x00,  0x90,  0x45,  0x61,  0x04,  0x90,  0x39,  0x50,  0x00,  0x90,  0x3E,  0x46,  0x03,  0x90,  0x40,  0x58,  0x60,  0xD0,  0x03,  0x05,  0xD0,  0x07,  0x05,  0xD0,  0x04,  0x01,  0x80,  0x3E,  0x00,
 0x04,  0xD0,  0x08,  0x04,  0x80,  0x39,  0x00,  0x05,  0xD0,  0x13,  0x09,  0xD0,  0x11,  0x0A,  0xD0,  0x10,  0x05,  0xD0,  0x0E,  0x03,  0x80,  0x32,  0x00,  0x02,  0xD0,  0x18,  0x0E,  0xD0,  0x17,  0x04,  0xD0,  0x18,  0x0A,  0xD0,  0x17,  0x0E,  0xD0,  0x13,  0x05,  0xD0,  0x11,  0x04,  0xD0,  0x0E,  0x05,  0xD0,  0x08,  0x05,  0xD0,  0x05,  0x05,  0xD0,  0x03,  0x04,  0x80,  0x40,  0x00,  0x01,  0xD0,  0x08,  0x05,  0xD0,  0x07,  0x13,  0x90,
 0x47,  0x5C,  0x04,  0xD0,  0x00,  0x6A,  0x80,  0x45,  0x00,  0x06,  0x90,  0x49,  0x4C,  0x0F,  0x80,  0x47,  0x00,  0x16,  0x90,  0x2D,  0x4C,  0x09,  0x90,  0x39,  0x44,  0x01,  0x90,  0x34,  0x4C,  0x01,  0x90,  0x3D,  0x38,  0x08,  0x90,  0x40,  0x46,  0x07,  0xB0,  0x40,  0x3A,  0x04,  0xB0,  0x40,  0x0E,  0x05,  0xB0,  0x40,  0x00,  0x1D,  0xD0,  0x01,  0x04,  0xD0,  0x05,  0x06,  0xD0,  0x08,  0x04,  0xD0,  0x0B,  0x05,  0xD0,  0x0E,  0x04,
 0xD0,  0x11,  0x05,  0xD0,  0x14,  0x06,  0xD0,  0x17,  0x04,  0xD0,  0x19,  0x05,  0xD0,  0x1C,  0x04,  0xD0,  0x1E,  0x05,  0xD0,  0x30,  0x06,  0xD0,  0x41,  0x00,  0xB0,  0x40,  0x0A,  0x04,  0xB0,  0x40,  0x68,  0x04,  0xD0,  0x40,  0x01,  0xB0,  0x40,  0x7F,  0x04,  0xD0,  0x3F,  0x05,  0xD0,  0x3E,  0x0E,  0xD0,  0x3C,  0x05,  0xD0,  0x3A,  0x05,  0xD0,  0x38,  0x06,  0xD0,  0x37,  0x04,  0xD0,  0x36,  0x04,  0xD0,  0x35,  0x04,  0xD0,  0x34,
 0x0A,  0xD0,  0x31,  0x05,  0xD0,  0x2D,  0x05,  0xD0,  0x2C,  0x05,  0xD0,  0x2A,  0x04,  0xD0,  0x28,  0x04,  0xD0,  0x26,  0x05,  0xD0,  0x23,  0x05,  0xD0,  0x21,  0x05,  0xD0,  0x1E,  0x04,  0xD0,  0x19,  0x05,  0xD0,  0x15,  0x05,  0xD0,  0x10,  0x04,  0xD0,  0x0A,  0x06,  0xD0,  0x04,  0x04,  0xD0,  0x00,  0x26,  0x80,  0x34,  0x00,  0x09,  0x80,  0x2D,  0x00,  0x02,  0x80,  0x3D,  0x00,  0x04,  0x80,  0x40,  0x00,  0x0B,  0x80,  0x39,  0x00,
 0x45,  0x80,  0x49,  0x00,  0x81,  0x02,  0x90,  0x2D,  0x4C,  0x00,  0x90,  0x40,  0x5F,  0x01,  0x90,  0x3D,  0x58,  0x01,  0x90,  0x49,  0x5C,  0x00,  0x90,  0x39,  0x4A,  0x03,  0x90,  0x34,  0x4A,  0x48,  0x80,  0x49,  0x00,  0x35,  0x90,  0x49,  0x50,  0x4A,  0x80,  0x49,  0x00,  0x32,  0x90,  0x49,  0x4E,  0x47,  0x80,  0x49,  0x00,  0x00,  0x80,  0x3D,  0x00,  0x05,  0x80,  0x34,  0x00,  0x0D,  0x80,  0x2D,  0x00,  0x1B,  0x90,  0x49,  0x3E,
 0x04,  0x80,  0x39,  0x00,  0x27,  0x80,  0x40,  0x00,  0x18,  0x80,  0x49,  0x00,  0x3A,  0x90,  0x2D,  0x50,  0x06,  0x90,  0x34,  0x4C,  0x00,  0x90,  0x39,  0x44,  0x01,  0x90,  0x40,  0x5E,  0x01,  0x90,  0x49,  0x5A,  0x04,  0x90,  0x3D,  0x54,  0x7C,  0x80,  0x3D,  0x00,  0x0E,  0x80,  0x49,  0x00,  0x00,  0x80,  0x40,  0x00,  0x08,  0x80,  0x39,  0x00,  0x08,  0x80,  0x34,  0x00,  0x0A,  0x80,  0x2D,  0x00,  0x48,  0x90,  0x47,  0x56,  0x81,
 0x1A,  0x80,  0x47,  0x00,  0x4F,  0x90,  0x2B,  0x52,  0x02,  0x90,  0x45,  0x5E,  0x00,  0x90,  0x3E,  0x5F,  0x04,  0x90,  0x3B,  0x5A,  0x00,  0x90,  0x37,  0x48,  0x00,  0x90,  0x32,  0x54,  0x00,  0x90,  0x39,  0x4A,  0x06,  0xB0,  0x40,  0x52,  0x05,  0xB0,  0x40,  0x18,  0x05,  0xB0,  0x40,  0x00,  0x52,  0xD0,  0x0A,  0x03,  0xD0,  0x16,  0x09,  0xD0,  0x14,  0x02,  0xB0,  0x40,  0x62,  0x03,  0xD0,  0x13,  0x02,  0xB0,  0x40,  0x7F,  0x11,
 0xD0,  0x10,  0x05,  0xD0,  0x0E,  0x04,  0xD0,  0x0A,  0x05,  0xD0,  0x04,  0x05,  0xD0,  0x00,  0x12,  0x80,  0x3B,  0x00,  0x1A,  0x80,  0x39,  0x00,  0x07,  0x80,  0x37,  0x00,  0x07,  0x80,  0x2B,  0x00,  0x10,  0x80,  0x32,  0x00,  0x00,  0x80,  0x3E,  0x00,  0x06,  0x80,  0x45,  0x00,  0x2C,  0x90,  0x47,  0x56,  0x22,  0x80,  0x47,  0x00,  0x5B,  0x90,  0x45,  0x5C,  0x36,  0x90,  0x2A,  0x56,  0x03,  0x90,  0x36,  0x4C,  0x00,  0x90,  0x3D,
 0x54,  0x01,  0x90,  0x31,  0x4E,  0x02,  0x90,  0x42,  0x58,  0x05,  0x90,  0x39,  0x52,  0x0E,  0xB0,  0x40,  0x50,  0x05,  0xB0,  0x40,  0x1A,  0x05,  0xB0,  0x40,  0x00,  0x07,  0xD0,  0x01,  0x05,  0xD0,  0x08,  0x05,  0xD0,  0x0F,  0x05,  0xD0,  0x15,  0x05,  0xD0,  0x1B,  0x04,  0xD0,  0x20,  0x05,  0xD0,  0x25,  0x04,  0xD0,  0x29,  0x05,  0xD0,  0x2D,  0x06,  0xD0,  0x31,  0x03,  0xD0,  0x35,  0x06,  0xD0,  0x38,  0x03,  0xD0,  0x3B,  0x05,
 0xD0,  0x52,  0x05,  0xD0,  0x67,  0x05,  0xD0,  0x66,  0x04,  0xD0,  0x65,  0x05,  0xD0,  0x64,  0x05,  0xD0,  0x62,  0x09,  0xD0,  0x60,  0x04,  0xB0,  0x40,  0x26,  0x01,  0xD0,  0x61,  0x03,  0xB0,  0x40,  0x76,  0x02,  0xD0,  0x60,  0x03,  0xB0,  0x40,  0x7F,  0x0B,  0xD0,  0x5E,  0x05,  0xD0,  0x5C,  0x04,  0xD0,  0x5A,  0x0F,  0xD0,  0x57,  0x04,  0xD0,  0x54,  0x05,  0xD0,  0x50,  0x05,  0xD0,  0x4A,  0x05,  0xD0,  0x48,  0x04,  0xD0,  0x46,
 0x05,  0xD0,  0x42,  0x06,  0xD0,  0x3D,  0x04,  0xD0,  0x39,  0x05,  0xD0,  0x37,  0x04,  0xD0,  0x33,  0x05,  0xD0,  0x2E,  0x05,  0xD0,  0x29,  0x04,  0xD0,  0x24,  0x06,  0xD0,  0x1E,  0x05,  0xD0,  0x16,  0x04,  0xD0,  0x0B,  0x04,  0xD0,  0x00,  0x09,  0x80,  0x39,  0x00,  0x08,  0x80,  0x3D,  0x00,  0x14,  0x80,  0x42,  0x00,  0x0A,  0x80,  0x31,  0x00,  0x02,  0x80,  0x36,  0x00,  0x04,  0x80,  0x2A,  0x00,  0x4E,  0x80,  0x45,  0x00,  0x6A,
 0x90,  0x2A,  0x4A,  0x06,  0x90,  0x36,  0x48,  0x01,  0x90,  0x31,  0x46,  0x03,  0x90,  0x47,  0x5C,  0x02,  0x90,  0x42,  0x5A,  0x02,  0x90,  0x45,  0x56,  0x60,  0xD0,  0x0C,  0x04,  0xD0,  0x16,  0x0E,  0xD0,  0x15,  0x02,  0x90,  0x49,  0x56,  0x04,  0xD0,  0x0D,  0x09,  0xD0,  0x0E,  0x09,  0xD0,  0x0D,  0x0A,  0xD0,  0x0B,  0x04,  0xD0,  0x0A,  0x05,  0xD0,  0x08,  0x04,  0xD0,  0x06,  0x05,  0xD0,  0x03,  0x05,  0xD0,  0x01,  0x05,  0xD0,
 0x00,  0x0B,  0x80,  0x49,  0x00,  0x35,  0x90,  0x49,  0x48,  0x0F,  0x80,  0x42,  0x00,  0x25,  0x80,  0x45,  0x00,  0x10,  0x80,  0x49,  0x00,  0x04,  0x80,  0x36,  0x00,  0x1A,  0x80,  0x2A,  0x00,  0x01,  0x80,  0x31,  0x00,  0x0C,  0x90,  0x49,  0x44,  0x1F,  0x80,  0x47,  0x00,  0x23,  0x80,  0x49,  0x00,  0x43,  0x90,  0x2A,  0x4A,  0x01,  0x90,  0x49,  0x54,  0x03,  0x90,  0x36,  0x42,  0x04,  0x90,  0x42,  0x5A,  0x00,  0x90,  0x31,  0x44,
 0x00,  0x90,  0x45,  0x56,  0x1A,  0xB0,  0x40,  0x52,  0x05,  0xB0,  0x40,  0x24,  0x06,  0xB0,  0x40,  0x02,  0x03,  0xB0,  0x40,  0x00,  0x34,  0xB0,  0x40,  0x4C,  0x04,  0xB0,  0x40,  0x7F,  0x03,  0xD0,  0x01,  0x0E,  0xD0,  0x00,  0x81,  0x00,  0x90,  0x47,  0x54,  0x11,  0x80,  0x49,  0x00,  0x0A,  0x80,  0x42,  0x00,  0x03,  0x80,  0x31,  0x00,  0x0B,  0x80,  0x36,  0x00,  0x0F,  0x80,  0x2A,  0x00,  0x15,  0x80,  0x45,  0x00,  0x1A,  0x80,
 0x47,  0x00,  0x7E,  0x90,  0x2A,  0x4C,  0x05,  0x90,  0x36,  0x44,  0x00,  0x90,  0x49,  0x5C,  0x01,  0x90,  0x45,  0x5A,  0x00,  0x90,  0x31,  0x40,  0x05,  0x90,  0x42,  0x5C,  0x00,  0xB0,  0x40,  0x56,  0x05,  0xB0,  0x40,  0x1A,  0x04,  0xB0,  0x40,  0x00,  0x3F,  0xB0,  0x40,  0x60,  0x05,  0xB0,  0x40,  0x7F,  0x06,  0xB0,  0x40,  0x7C,  0x04,  0xB0,  0x40,  0x7F,  0x0A,  0xD0,  0x0E,  0x0A,  0xD0,  0x0D,  0x0E,  0xD0,  0x0C,  0x05,  0xD0,
 0x0B,  0x05,  0xD0,  0x09,  0x04,  0xD0,  0x06,  0x05,  0xD0,  0x04,  0x05,  0xD0,  0x03,  0x05,  0xD0,  0x02,  0x05,  0xD0,  0x00,  0x49,  0x90,  0x4C,  0x5E,  0x05,  0x80,  0x36,  0x00,  0x07,  0x80,  0x31,  0x00,  0x11,  0x80,  0x2A,  0x00,  0x12,  0x80,  0x49,  0x00,  0x0E,  0x80,  0x42,  0x00,  0x08,  0x80,  0x45,  0x00,  0x1A,  0x80,  0x4C,  0x00,  0x81,  0x13,  0x90,  0x42,  0x50,  0x05,  0xB0,  0x40,  0x50,  0x01,  0x90,  0x26,  0x5C,  0x01,
 0x90,  0x32,  0x61,  0x00,  0x90,  0x45,  0x5E,  0x04,  0xB0,  0x40,  0x18,  0x03,  0x90,  0x4E,  0x6D,  0x01,  0xB0,  0x40,  0x00,  0x30,  0xD0,  0x05,  0x06,  0xD0,  0x0A,  0x04,  0xD0,  0x10,  0x04,  0xD0,  0x15,  0x05,  0xD0,  0x19,  0x05,  0xD0,  0x1D,  0x06,  0xD0,  0x21,  0x04,  0xD0,  0x25,  0x04,  0xD0,  0x28,  0x05,  0xD0,  0x2C,  0x05,  0xD0,  0x46,  0x06,  0xD0,  0x5F,  0x09,  0xD0,  0x5D,  0x08,  0xD0,  0x5E,  0x04,  0xD0,  0x5D,  0x05,
 0xD0,  0x5B,  0x0A,  0xD0,  0x58,  0x05,  0xD0,  0x56,  0x05,  0xD0,  0x53,  0x04,  0xD0,  0x51,  0x05,  0xD0,  0x4D,  0x05,  0xD0,  0x48,  0x01,  0xB0,  0x40,  0x01,  0x03,  0xD0,  0x42,  0x02,  0xB0,  0x40,  0x62,  0x03,  0xD0,  0x3C,  0x03,  0xB0,  0x40,  0x7F,  0x03,  0xD0,  0x37,  0x04,  0xD0,  0x31,  0x05,  0xD0,  0x2C,  0x04,  0xD0,  0x27,  0x05,  0xD0,  0x23,  0x05,  0xD0,  0x1D,  0x04,  0xD0,  0x18,  0x06,  0xD0,  0x13,  0x00,  0x80,  0x42,
 0x00,  0x04,  0xD0,  0x14,  0x04,  0xD0,  0x0D,  0x05,  0xD0,  0x06,  0x05,  0xD0,  0x00,  0x19,  0x90,  0x42,  0x10,  0x1F,  0x80,  0x4E,  0x00,  0x0B,  0x80,  0x45,  0x00,  0x10,  0x80,  0x42,  0x00,  0x37,  0x80,  0x32,  0x00,  0x81,  0x04,  0x90,  0x42,  0x40,  0x0C,  0x90,  0x45,  0x56,  0x00,  0x90,  0x32,  0x58,  0x09,  0x90,  0x4E,  0x69,  0x61,  0xD0,  0x0C,  0x05,  0xD0,  0x19,  0x0A,  0xD0,  0x18,  0x04,  0xD0,  0x17,  0x0A,  0xD0,  0x18,
 0x05,  0xD0,  0x17,  0x09,  0xD0,  0x16,  0x05,  0xD0,  0x15,  0x07,  0x80,  0x42,  0x00,  0x03,  0xD0,  0x1B,  0x09,  0xD0,  0x19,  0x05,  0xD0,  0x16,  0x04,  0xD0,  0x14,  0x05,  0xD0,  0x10,  0x05,  0xD0,  0x0D,  0x05,  0xD0,  0x0C,  0x05,  0xD0,  0x0A,  0x05,  0xD0,  0x07,  0x04,  0xD0,  0x05,  0x05,  0xD0,  0x03,  0x04,  0xD0,  0x00,  0x21,  0x90,  0x42,  0x32,  0x23,  0x80,  0x45,  0x00,  0x01,  0x80,  0x26,  0x00,  0x06,  0x80,  0x42,  0x00,
 0x0D,  0x80,  0x32,  0x00,  0x0B,  0x80,  0x4E,  0x00,  0x52,  0x90,  0x45,  0x4C,  0x04,  0x90,  0x4A,  0x5F,  0x00,  0x90,  0x32,  0x52,  0x03,  0x90,  0x39,  0x50,  0x03,  0x90,  0x3E,  0x42,  0x00,  0x90,  0x4E,  0x5F,  0x0E,  0xB0,  0x40,  0x52,  0x05,  0xB0,  0x40,  0x1A,  0x05,  0xB0,  0x40,  0x00,  0x39,  0xB0,  0x40,  0x48,  0x04,  0xB0,  0x40,  0x7F,  0x01,  0xD0,  0x02,  0x04,  0xD0,  0x04,  0x06,  0xD0,  0x19,  0x03,  0xD0,  0x2D,  0x06,
 0xD0,  0x2C,  0x04,  0xD0,  0x2B,  0x18,  0xD0,  0x28,  0x12,  0xD0,  0x27,  0x05,  0xD0,  0x26,  0x05,  0xD0,  0x24,  0x05,  0xD0,  0x22,  0x05,  0xD0,  0x21,  0x06,  0xD0,  0x20,  0x03,  0xD0,  0x1E,  0x05,  0xD0,  0x1C,  0x05,  0xD0,  0x1A,  0x04,  0xD0,  0x18,  0x05,  0xD0,  0x15,  0x05,  0xD0,  0x11,  0x05,  0xD0,  0x0E,  0x04,  0xD0,  0x0C,  0x05,  0xD0,  0x09,  0x05,  0xD0,  0x07,  0x09,  0xD0,  0x05,  0x0A,  0xD0,  0x07,  0x01,  0x90,  0x51,
 0x5F,  0x03,  0xD0,  0x01,  0x0A,  0xD0,  0x03,  0x05,  0xD0,  0x04,  0x04,  0xD0,  0x05,  0x06,  0xD0,  0x06,  0x04,  0xD0,  0x08,  0x06,  0xD0,  0x09,  0x04,  0xD0,  0x0A,  0x04,  0xD0,  0x0B,  0x05,  0xD0,  0x0C,  0x05,  0xD0,  0x0D,  0x04,  0xD0,  0x0E,  0x06,  0xD0,  0x0F,  0x13,  0x80,  0x39,  0x00,  0x03,  0xD0,  0x16,  0x05,  0xD0,  0x15,  0x05,  0xD0,  0x06,  0x04,  0xD0,  0x00,  0x16,  0x80,  0x4A,  0x00,  0x06,  0x80,  0x4E,  0x00,  0x05,
 0x80,  0x3E,  0x00,  0x04,  0x80,  0x51,  0x00,  0x68,  0x90,  0x3E,  0x44,  0x03,  0x90,  0x39,  0x4A,  0x05,  0x90,  0x49,  0x61,  0x81,  0x17,  0x80,  0x3E,  0x00,  0x10,  0x80,  0x39,  0x00,  0x18,  0x80,  0x32,  0x00,  0x14,  0x80,  0x45,  0x00,  0x82,  0x05,  0x90,  0x45,  0x38,  0x03,  0x90,  0x26,  0x5C,  0x01,  0x90,  0x42,  0x4C,  0x01,  0x90,  0x3E,  0x58,  0x04,  0x90,  0x32,  0x5F,  0x08,  0xB0,  0x40,  0x5A,  0x05,  0xB0,  0x40,  0x1E,
 0x05,  0xB0,  0x40,  0x00,  0x19,  0xD0,  0x05,  0x06,  0xD0,  0x0C,  0x05,  0xD0,  0x14,  0x04,  0xD0,  0x1A,  0x05,  0xD0,  0x21,  0x04,  0xD0,  0x27,  0x05,  0xD0,  0x2C,  0x06,  0xD0,  0x32,  0x04,  0xD0,  0x36,  0x05,  0xD0,  0x3B,  0x05,  0xD0,  0x40,  0x04,  0xD0,  0x64,  0x06,  0xD0,  0x7F,  0x02,  0xB0,  0x40,  0x40,  0x05,  0xB0,  0x40,  0x7F,  0x0F,  0xD0,  0x7D,  0x05,  0xD0,  0x7A,  0x05,  0xD0,  0x78,  0x04,  0xD0,  0x73,  0x05,  0xD0,
 0x6E,  0x05,  0xD0,  0x69,  0x05,  0xD0,  0x63,  0x04,  0xD0,  0x5D,  0x05,  0xD0,  0x58,  0x05,  0xD0,  0x52,  0x04,  0xD0,  0x4D,  0x05,  0xD0,  0x46,  0x05,  0xD0,  0x3E,  0x05,  0xD0,  0x36,  0x05,  0xD0,  0x2E,  0x04,  0xD0,  0x28,  0x05,  0xD0,  0x1E,  0x05,  0xD0,  0x15,  0x05,  0xD0,  0x0C,  0x04,  0xD0,  0x02,  0x05,  0xD0,  0x00,  0x03,  0x80,  0x26,  0x00,  0x05,  0x80,  0x3E,  0x00,  0x07,  0x80,  0x42,  0x00,  0x02,  0x80,  0x45,  0x00,
 0x0A,  0x80,  0x32,  0x00,  0x16,  0x80,  0x49,  0x00,  0x81,  0x50,  0x90,  0x32,  0x52,  0x03,  0x90,  0x39,  0x54,  0x00,  0x90,  0x3E,  0x44,  0x05,  0x90,  0x42,  0x58,  0x00,  0x90,  0x49,  0x56,  0x02,  0x90,  0x45,  0x54,  0x57,  0x80,  0x49,  0x00,  0x23,  0x90,  0x49,  0x5A,  0x4C,  0x80,  0x49,  0x00,  0x26,  0x90,  0x49,  0x52,  0x4A,  0x80,  0x49,  0x00,  0x23,  0x90,  0x49,  0x56,  0x03,  0x80,  0x39,  0x00,  0x05,  0x80,  0x32,  0x00,
 0x10,  0x80,  0x42,  0x00,  0x0D,  0x80,  0x45,  0x00,  0x0C,  0x80,  0x3E,  0x00,  0x20,  0x80,  0x49,  0x00,  0x30,  0x90,  0x32,  0x50,  0x01,  0x90,  0x45,  0x5C,  0x01,  0x90,  0x49,  0x4C,  0x00,  0x90,  0x42,  0x5A,  0x04,  0x90,  0x3E,  0x3A,  0x01,  0x90,  0x39,  0x4C,  0x81,  0x63,  0x80,  0x42,  0x00,  0x08,  0x90,  0x47,  0x4C,  0x23,  0x80,  0x49,  0x00,  0x3D,  0x80,  0x39,  0x00,  0x20,  0x80,  0x3E,  0x00,  0x0E,  0x80,  0x32,  0x00,
 0x07,  0x80,  0x45,  0x00,  0x42,  0x80,  0x47,  0x00,  0x2A,  0x90,  0x32,  0x50,  0x00,  0x90,  0x45,  0x63,  0x04,  0x90,  0x3E,  0x42,  0x01,  0x90,  0x39,  0x4C,  0x01,  0x90,  0x40,  0x5C,  0x51,  0xD0,  0x02,  0x05,  0xD0,  0x04,  0x04,  0xD0,  0x06,  0x05,  0xD0,  0x18,  0x06,  0xD0,  0x27,  0x04,  0xD0,  0x25,  0x0E,  0xD0,  0x24,  0x0A,  0xD0,  0x25,  0x05,  0xD0,  0x24,  0x04,  0xD0,  0x25,  0x05,  0xD0,  0x24,  0x04,  0xD0,  0x21,  0x05,
 0xD0,  0x1F,  0x05,  0xD0,  0x1D,  0x05,  0xD0,  0x1C,  0x09,  0xD0,  0x19,  0x05,  0xD0,  0x16,  0x04,  0xD0,  0x14,  0x05,  0xD0,  0x12,  0x02,  0x80,  0x3E,  0x00,  0x03,  0xD0,  0x15,  0x05,  0xD0,  0x11,  0x05,  0xD0,  0x0C,  0x00,  0x80,  0x39,  0x00,  0x05,  0xD0,  0x0F,  0x04,  0xD0,  0x09,  0x05,  0xD0,  0x05,  0x04,  0xD0,  0x01,  0x05,  0xD0,  0x00,  0x00,  0x80,  0x40,  0x00,  0x05,  0xD0,  0x06,  0x09,  0x80,  0x32,  0x00,  0x01,  0xD0,
 0x0E,  0x05,  0x90,  0x47,  0x65,  0x02,  0xD0,  0x06,  0x0C,  0xD0,  0x07,  0x1B,  0xD0,  0x06,  0x0A,  0xD0,  0x05,  0x07,  0x80,  0x45,  0x00,  0x02,  0xD0,  0x0C,  0x06,  0xD0,  0x0B,  0x04,  0xD0,  0x0A,  0x04,  0xD0,  0x09,  0x05,  0xD0,  0x07,  0x05,  0xD0,  0x05,  0x04,  0xD0,  0x03,  0x05,  0xD0,  0x00,  0x14,  0x80,  0x47,  0x00,  0x06,  0x90,  0x49,  0x48,  0x79,  0x90,  0x3D,  0x54,  0x04,  0x90,  0x2D,  0x54,  0x02,  0x90,  0x40,  0x4C,
 0x01,  0x90,  0x39,  0x4C,  0x03,  0x90,  0x34,  0x50,  0x13,  0xB0,  0x40,  0x58,  0x05,  0xB0,  0x40,  0x1C,  0x05,  0xB0,  0x40,  0x00,  0x09,  0xD0,  0x01,  0x06,  0xD0,  0x07,  0x04,  0xD0,  0x0D,  0x04,  0xD0,  0x13,  0x05,  0xD0,  0x18,  0x05,  0xD0,  0x1D,  0x04,  0xD0,  0x22,  0x06,  0xD0,  0x26,  0x04,  0xD0,  0x2A,  0x04,  0xD0,  0x2E,  0x05,  0xD0,  0x31,  0x05,  0xD0,  0x35,  0x05,  0xD0,  0x50,  0x05,  0xD0,  0x68,  0x0E,  0xD0,  0x67,
 0x02,  0xB0,  0x40,  0x60,  0x02,  0xD0,  0x66,  0x02,  0xB0,  0x40,  0x7F,  0x04,  0xD0,  0x63,  0x04,  0xD0,  0x62,  0x01,  0xB0,  0x40,  0x68,  0x04,  0xD0,  0x60,  0x01,  0xB0,  0x40,  0x7F,  0x04,  0xD0,  0x5E,  0x04,  0xD0,  0x5D,  0x06,  0xD0,  0x5A,  0x04,  0xD0,  0x59,  0x05,  0xD0,  0x57,  0x04,  0xD0,  0x54,  0x05,  0xD0,  0x52,  0x06,  0xD0,  0x4E,  0x03,  0xD0,  0x4B,  0x05,  0xD0,  0x49,  0x05,  0xD0,  0x46,  0x04,  0xD0,  0x43,  0x05,
 0xD0,  0x3F,  0x05,  0xD0,  0x3C,  0x05,  0xD0,  0x38,  0x05,  0xD0,  0x35,  0x04,  0xD0,  0x30,  0x05,  0xD0,  0x2A,  0x05,  0xD0,  0x24,  0x04,  0xD0,  0x1C,  0x05,  0xD0,  0x13,  0x05,  0xD0,  0x0B,  0x07,  0xD0,  0x03,  0x03,  0xD0,  0x00,  0x11,  0x80,  0x40,  0x00,  0x04,  0x80,  0x3D,  0x00,  0x26,  0x80,  0x34,  0x00,  0x06,  0x80,  0x49,  0x00,  0x0A,  0x80,  0x2D,  0x00,  0x10,  0x80,  0x39,  0x00,  0x52,  0x90,  0x45,  0x5C,  0x02,  0x90,
 0x2D,  0x4E,  0x02,  0x90,  0x49,  0x5F,  0x00,  0x90,  0x39,  0x52,  0x02,  0x90,  0x40,  0x58,  0x04,  0x90,  0x34,  0x4C,  0x36,  0xD0,  0x04,  0x05,  0xD0,  0x07,  0x05,  0xD0,  0x09,  0x04,  0xD0,  0x0A,  0x05,  0xD0,  0x0C,  0x03,  0x80,  0x49,  0x00,  0x02,  0xD0,  0x14,  0x13,  0xD0,  0x10,  0x05,  0xD0,  0x0F,  0x07,  0x90,  0x49,  0x5C,  0x03,  0xD0,  0x08,  0x04,  0xD0,  0x09,  0x05,  0xD0,  0x0A,  0x09,  0xD0,  0x0B,  0x1C,  0xD0,  0x0A,
 0x05,  0xD0,  0x09,  0x05,  0xD0,  0x08,  0x06,  0xD0,  0x06,  0x03,  0xD0,  0x05,  0x03,  0x80,  0x49,  0x00,  0x02,  0xD0,  0x0B,  0x05,  0xD0,  0x09,  0x05,  0xD0,  0x07,  0x04,  0xD0,  0x06,  0x05,  0xD0,  0x04,  0x05,  0xD0,  0x00,  0x1C,  0x90,  0x49,  0x5C,  0x49,  0x80,  0x49,  0x00,  0x23,  0x80,  0x34,  0x00,  0x05,  0x90,  0x49,  0x4E,  0x10,  0x80,  0x2D,  0x00,  0x0D,  0x80,  0x40,  0x00,  0x0B,  0x80,  0x39,  0x00,  0x0F,  0x80,  0x45,
 0x00,  0x0C,  0x80,  0x49,  0x00,  0x34,  0x90,  0x45,  0x65,  0x03,  0x90,  0x2D,  0x4C,  0x00,  0x90,  0x49,  0x58,  0x03,  0x90,  0x40,  0x5A,  0x02,  0x90,  0x39,  0x50,  0x03,  0x90,  0x34,  0x4A,  0x10,  0xB0,  0x40,  0x60,  0x05,  0xB0,  0x40,  0x20,  0x05,  0xB0,  0x40,  0x00,  0x18,  0xB0,  0x40,  0x5C,  0x05,  0xB0,  0x40,  0x7F,  0x0B,  0xD0,  0x01,  0x04,  0xD0,  0x05,  0x05,  0xD0,  0x08,  0x06,  0xD0,  0x0C,  0x03,  0xD0,  0x0F,  0x05,
 0xD0,  0x11,  0x05,  0xD0,  0x14,  0x05,  0xD0,  0x29,  0x04,  0xD0,  0x3D,  0x05,  0xD0,  0x3E,  0x0A,  0xD0,  0x3B,  0x0E,  0xD0,  0x39,  0x05,  0xD0,  0x38,  0x13,  0xD0,  0x35,  0x04,  0xD0,  0x33,  0x05,  0xD0,  0x31,  0x05,  0xD0,  0x2E,  0x05,  0xD0,  0x2A,  0x05,  0xD0,  0x25,  0x05,  0xD0,  0x20,  0x05,  0xD0,  0x1A,  0x03,  0xD0,  0x12,  0x05,  0xD0,  0x08,  0x05,  0xD0,  0x00,  0x16,  0x90,  0x47,  0x5C,  0x08,  0x80,  0x49,  0x00,  0x1D,
 0x80,  0x34,  0x00,  0x0B,  0x80,  0x2D,  0x00,  0x24,  0x80,  0x39,  0x00,  0x01,  0x80,  0x45,  0x00,  0x10,  0x80,  0x40,  0x00,  0x2F,  0x80,  0x47,  0x00,  0x62,  0x90,  0x3E,  0x5A,  0x01,  0x90,  0x37,  0x52,  0x00,  0x90,  0x2B,  0x56,  0x00,  0x90,  0x45,  0x5A,  0x05,  0x90,  0x32,  0x5A,  0x00,  0x90,  0x3B,  0x54,  0x0F,  0xB0,  0x40,  0x56,  0x04,  0xB0,  0x40,  0x1A,  0x05,  0xB0,  0x40,  0x00,  0x3E,  0xD0,  0x02,  0x05,  0xD0,  0x04,
 0x05,  0xD0,  0x16,  0x05,  0xD0,  0x29,  0x05,  0xD0,  0x2A,  0x03,  0xD0,  0x29,  0x0B,  0xD0,  0x28,  0x01,  0xB0,  0x40,  0x54,  0x03,  0xD0,  0x25,  0x01,  0xB0,  0x40,  0x7F,  0x03,  0xD0,  0x23,  0x10,  0xD0,  0x21,  0x03,  0xD0,  0x1C,  0x05,  0xD0,  0x16,  0x05,  0xD0,  0x0F,  0x05,  0xD0,  0x0B,  0x04,  0xD0,  0x05,  0x06,  0xD0,  0x00,  0x10,  0x80,  0x3B,  0x00,  0x0C,  0x80,  0x37,  0x00,  0x0A,  0x80,  0x2B,  0x00,  0x08,  0x80,  0x3E,
 0x00,  0x06,  0x80,  0x32,  0x00,  0x04,  0x90,  0x47,  0x58,  0x09,  0x80,  0x45,  0x00,  0x19,  0x80,  0x47,  0x00,  0x58,  0x90,  0x45,  0x61,  0x71,  0x90,  0x3D,  0x5C,  0x05,  0x90,  0x42,  0x52,  0x02,  0x90,  0x36,  0x4C,  0x00,  0x90,  0x2A,  0x4C,  0x05,  0x90,  0x39,  0x48,  0x01,  0x90,  0x31,  0x4C,  0x0A,  0xB0,  0x40,  0x50,  0x04,  0xB0,  0x40,  0x18,  0x03,  0x80,  0x39,  0x00,  0x02,  0xB0,  0x40,  0x00,  0x1B,  0xD0,  0x05,  0x04,
 0xD0,  0x0A,  0x05,  0xD0,  0x0F,  0x05,  0xD0,  0x14,  0x05,  0xD0,  0x18,  0x05,  0xD0,  0x1C,  0x06,  0xD0,  0x20,  0x03,  0xD0,  0x23,  0x05,  0xD0,  0x26,  0x05,  0xD0,  0x29,  0x04,  0xD0,  0x43,  0x05,  0xD0,  0x5A,  0x13,  0xD0,  0x59,  0x09,  0xB0,  0x40,  0x5E,  0x04,  0xB0,  0x40,  0x7F,  0x07,  0xD0,  0x5A,  0x04,  0xD0,  0x58,  0x03,  0xB0,  0x40,  0x54,  0x01,  0xD0,  0x57,  0x04,  0xB0,  0x40,  0x7F,  0x01,  0xD0,  0x56,  0x05,  0xD0,
 0x55,  0x13,  0xD0,  0x53,  0x0A,  0xD0,  0x54,  0x04,  0xD0,  0x52,  0x04,  0xD0,  0x51,  0x0A,  0xD0,  0x4E,  0x06,  0xD0,  0x4B,  0x09,  0xD0,  0x48,  0x03,  0xD0,  0x44,  0x05,  0xD0,  0x41,  0x05,  0xD0,  0x3C,  0x05,  0xD0,  0x36,  0x05,  0xD0,  0x31,  0x04,  0xD0,  0x29,  0x05,  0xD0,  0x1F,  0x05,  0xD0,  0x15,  0x04,  0xD0,  0x09,  0x05,  0xD0,  0x00,  0x09,  0x80,  0x3D,  0x00,  0x10,  0x80,  0x42,  0x00,  0x0B,  0x80,  0x31,  0x00,  0x00,
 0x80,  0x45,  0x00,  0x05,  0x80,  0x2A,  0x00,  0x01,  0x80,  0x36,  0x00,  0x81,  0x2C,  0x90,  0x2A,  0x4C,  0x05,  0x90,  0x47,  0x5A,  0x02,  0x90,  0x36,  0x40,  0x00,  0x90,  0x42,  0x52,  0x01,  0x90,  0x31,  0x48,  0x01,  0x90,  0x45,  0x52,  0x73,  0x90,  0x49,  0x59,  0x41,  0x80,  0x49,  0x00,  0x2A,  0x90,  0x49,  0x51,  0x3D,  0x80,  0x42,  0x00,  0x08,  0x80,  0x49,  0x00,  0x1A,  0x80,  0x2A,  0x00,  0x13,  0x80,  0x31,  0x00,  0x03,
 0x90,  0x49,  0x49,  0x07,  0x80,  0x36,  0x00,  0x0D,  0x80,  0x45,  0x00,  0x17,  0x80,  0x47,  0x00,  0x1A,  0x80,  0x49,  0x00,  0x3F,  0x90,  0x2A,  0x4A,  0x01,  0x90,  0x36,  0x44,  0x01,  0x90,  0x49,  0x52,  0x00,  0x90,  0x45,  0x58,  0x01,  0x90,  0x42,  0x5A,  0x04,  0x90,  0x31,  0x46,  0x81,  0x4F,  0x90,  0x47,  0x58,  0x31,  0x80,  0x42,  0x00,  0x0A,  0x80,  0x49,  0x00,  0x0E,  0x80,  0x31,  0x00,  0x08,  0x80,  0x36,  0x00,  0x00,
 0x80,  0x2A,  0x00,  0x1A,  0x80,  0x45,  0x00,  0x02,  0x80,  0x47,  0x00,  0x81,  0x08,  0x90,  0x2A,  0x50,  0x01,  0x90,  0x36,  0x4A,  0x03,  0x90,  0x45,  0x5E,  0x01,  0x90,  0x49,  0x5F,  0x02,  0x90,  0x42,  0x5E,  0x00,  0x90,  0x31,  0x44,  0x12,  0xB0,  0x40,  0x5E,  0x04,  0xB0,  0x40,  0x20,  0x05,  0xB0,  0x40,  0x01,  0x06,  0xB0,  0x40,  0x00,  0x26,  0xB0,  0x40,  0x52,  0x04,  0xB0,  0x40,  0x7F,  0x06,  0xB0,  0x40,  0x3E,  0x03,
 0xB0,  0x40,  0x7F,  0x26,  0x80,  0x36,  0x00,  0x0A,  0x80,  0x31,  0x00,  0x1D,  0x80,  0x2A,  0x00,  0x28,  0x80,  0x42,  0x00,  0x18,  0x90,  0x34,  0x40,  0x00,  0x90,  0x4C,  0x61,  0x1E,  0x80,  0x49,  0x00,  0x1E,  0x80,  0x34,  0x00,  0x0F,  0x80,  0x45,  0x00,  0x0B,  0x80,  0x4C,  0x00,  0x81,  0x0E,  0x90,  0x42,  0x4A,  0x08,  0x90,  0x45,  0x5C,  0x04,  0x90,  0x32,  0x5F,  0x01,  0x90,  0x26,  0x56,  0x03,  0xB0,  0x40,  0x50,  0x00,
 0x90,  0x4E,  0x6B,  0x04,  0xB0,  0x40,  0x14,  0x05,  0xB0,  0x40,  0x00,  0x31,  0xD0,  0x05,  0x05,  0xD0,  0x0B,  0x05,  0xD0,  0x10,  0x05,  0xD0,  0x15,  0x05,  0xD0,  0x1A,  0x04,  0xD0,  0x1E,  0x05,  0xD0,  0x22,  0x05,  0xD0,  0x26,  0x05,  0xD0,  0x47,  0x04,  0xD0,  0x63,  0x05,  0xD0,  0x61,  0x06,  0xD0,  0x63,  0x04,  0xD0,  0x61,  0x0E,  0xD0,  0x60,  0x0A,  0xD0,  0x5E,  0x08,  0xB0,  0x40,  0x08,  0x01,  0xD0,  0x5D,  0x04,  0xB0,
 0x40,  0x66,  0x01,  0xD0,  0x5B,  0x04,  0xB0,  0x40,  0x7F,  0x01,  0xD0,  0x5A,  0x05,  0xD0,  0x58,  0x04,  0xD0,  0x57,  0x05,  0xD0,  0x55,  0x05,  0xD0,  0x52,  0x04,  0xD0,  0x4F,  0x06,  0xD0,  0x4B,  0x03,  0xD0,  0x46,  0x03,  0x80,  0x42,  0x00,  0x02,  0xD0,  0x4B,  0x05,  0xD0,  0x44,  0x05,  0xD0,  0x3C,  0x03,  0xD0,  0x34,  0x05,  0xD0,  0x2B,  0x04,  0xD0,  0x21,  0x06,  0xD0,  0x18,  0x03,  0xD0,  0x0D,  0x06,  0xD0,  0x03,  0x04,
 0xD0,  0x00,  0x1F,  0x90,  0x42,  0x1B,  0x0B,  0x80,  0x4E,  0x00,  0x06,  0x80,  0x45,  0x00,  0x10,  0x80,  0x42,  0x00,  0x21,  0x80,  0x32,  0x00,  0x81,  0x1E,  0x90,  0x42,  0x34,  0x0F,  0x90,  0x45,  0x5A,  0x05,  0x90,  0x32,  0x48,  0x05,  0x90,  0x4E,  0x69,  0x5F,  0xD0,  0x03,  0x04,  0xD0,  0x0F,  0x0A,  0xD0,  0x11,  0x06,  0xD0,  0x0F,  0x03,  0xD0,  0x10,  0x0A,  0xD0,  0x0F,  0x05,  0xD0,  0x0D,  0x08,  0x80,  0x42,  0x00,  0x01,
 0xD0,  0x17,  0x05,  0xD0,  0x14,  0x04,  0xD0,  0x10,  0x05,  0xD0,  0x0F,  0x0B,  0xD0,  0x0C,  0x04,  0xD0,  0x08,  0x04,  0xD0,  0x06,  0x05,  0xD0,  0x04,  0x05,  0xD0,  0x01,  0x04,  0xD0,  0x00,  0x03,  0x80,  0x26,  0x00,  0x02,  0xD0,  0x05,  0x05,  0xD0,  0x02,  0x05,  0xD0,  0x00,  0x13,  0x80,  0x32,  0x00,  0x12,  0x90,  0x42,  0x2C,  0x12,  0x80,  0x45,  0x00,  0x06,  0x80,  0x42,  0x00,  0x1A,  0x80,  0x4E,  0x00,  0x81,  0x2D,  0x90,
 0x45,  0x3E,  0x13,  0x90,  0x32,  0x46,  0x06,  0x90,  0x4A,  0x5F,  0x02,  0x90,  0x3E,  0x38,  0x01,  0x90,  0x39,  0x3A,  0x08,  0x90,  0x4E,  0x58,  0x81,  0x73,  0x90,  0x51,  0x65,  0x39,  0x80,  0x39,  0x00,  0x27,  0xD0,  0x03,  0x09,  0xD0,  0x00,  0x01,  0x80,  0x3E,  0x00,  0x04,  0xD0,  0x02,  0x04,  0xD0,  0x00,  0x3B,  0x80,  0x51,  0x00,  0x0D,  0x80,  0x4A,  0x00,  0x03,  0x80,  0x4E,  0x00,  0x34,  0x90,  0x3E,  0x2C,  0x0D,  0x90,
 0x39,  0x21,  0x43,  0x80,  0x45,  0x00,  0x3B,  0x90,  0x49,  0x42,  0x46,  0x80,  0x3E,  0x00,  0x09,  0x80,  0x39,  0x00,  0x20,  0x90,  0x47,  0x48,  0x01,  0x80,  0x49,  0x00,  0x01,  0x80,  0x32,  0x00,  0x81,  0x3D,  0x90,  0x3B,  0x2E,  0x0A,  0x90,  0x35,  0x30,  0x00,  0x90,  0x3E,  0x2C,  0x02,  0xB0,  0x40,  0x56,  0x02,  0x90,  0x41,  0x34,  0x02,  0x90,  0x29,  0x30,  0x01,  0xB0,  0x40,  0x1A,  0x05,  0xB0,  0x40,  0x00,  0x65,  0xB0,
 0x40,  0x01,  0x05,  0xB0,  0x40,  0x62,  0x04,  0xB0,  0x40,  0x7F,  0x7F,  0x80,  0x3B,  0x00,  0x01,  0x80,  0x35,  0x00,  0x08,  0x80,  0x41,  0x00,  0x0E,  0x80,  0x3E,  0x00,  0x81,  0x59,  0x90,  0x41,  0x1D,  0x00,  0x90,  0x3E,  0x1F,  0x2F,  0x90,  0x35,  0x1C,  0x04,  0x90,  0x3B,  0x1F,  0x81,  0x46,  0x80,  0x3E,  0x00,  0x07,  0x80,  0x41,  0x00,  0x24,  0x80,  0x3B,  0x00,  0x0D,  0x80,  0x35,  0x00,  0x81,  0x16,  0x90,  0x35,  0x20,
 0x04,  0x90,  0x3E,  0x1E,  0x0C,  0x90,  0x3B,  0x24,  0x02,  0x90,  0x41,  0x16,  0x38,  0x80,  0x3B,  0x00,  0x02,  0x80,  0x3E,  0x00,  0x04,  0x80,  0x41,  0x00,  0x55,  0x80,  0x35,  0x00,  0x4D,  0x90,  0x47,  0x38,  0x04,  0x80,  0x29,  0x00,  0x06,  0x80,  0x47,  0x00,  0x58,  0x80,  0x47,  0x00,  0x08,  0x90,  0x45,  0x40,  0x71,  0x80,  0x45,  0x00,  0x02,  0x90,  0x47,  0x32,  0x1B,  0x90,  0x49,  0x42,  0x02,  0x80,  0x47,  0x00,  0x31,
 0x80,  0x49,  0x00,  0x81,  0x22,  0x90,  0x45,  0x2F,  0x37,  0x80,  0x45,  0x00,  0x3E,  0x90,  0x45,  0x3E,  0x81,  0x19,  0x90,  0x39,  0x2F,  0x03,  0x90,  0x31,  0x28,  0x07,  0x90,  0x2A,  0x18,  0x15,  0xB0,  0x40,  0x64,  0x05,  0xB0,  0x40,  0x22,  0x05,  0xB0,  0x40,  0x01,  0x05,  0xB0,  0x40,  0x00,  0x5B,  0xB0,  0x40,  0x50,  0x05,  0xB0,  0x40,  0x7F,  0x0E,  0xB0,  0x40,  0x7A,  0x05,  0xB0,  0x40,  0x7F,  0x0E,  0x80,  0x45,  0x00,
 0x50,  0x80,  0x31,  0x00,  0x02,  0x90,  0x3D,  0x3C,  0x03,  0x80,  0x2A,  0x00,  0x6F,  0x80,  0x39,  0x00,  0x09,  0x80,  0x3D,  0x00,  0x79,  0x90,  0x3D,  0x40,  0x01,  0x90,  0x34,  0x22,  0x00,  0xB0,  0x40,  0x54,  0x03,  0x90,  0x2D,  0x28,  0x03,  0xB0,  0x40,  0x18,  0x04,  0xB0,  0x40,  0x00,  0x00,  0x90,  0x39,  0x36,  0x60,  0xB0,  0x40,  0x56,  0x05,  0xB0,  0x40,  0x7F,  0x81,  0x0B,  0x90,  0x40,  0x46,  0x04,  0x80,  0x3D,  0x00,
 0x03,  0x80,  0x34,  0x00,  0x03,  0x80,  0x2D,  0x00,  0x1A,  0x80,  0x39,  0x00,  0x4C,  0x90,  0x3D,  0x40,  0x1B,  0x80,  0x40,  0x00,  0x5A,  0x90,  0x31,  0x28,  0x06,  0x90,  0x2A,  0x1F,  0x05,  0x90,  0x39,  0x33,  0x05,  0xB0,  0x40,  0x30,  0x06,  0xB0,  0x40,  0x08,  0x03,  0xB0,  0x40,  0x00,  0x48,  0xB0,  0x40,  0x56,  0x06,  0xB0,  0x40,  0x7F,  0x0D,  0xB0,  0x40,  0x52,  0x05,  0xB0,  0x40,  0x7F,  0x45,  0x80,  0x3D,  0x00,  0x32,
 0x90,  0x3D,  0x36,  0x15,  0x80,  0x2A,  0x00,  0x02,  0x80,  0x31,  0x00,  0x1C,  0x80,  0x39,  0x00,  0x35,  0x80,  0x3D,  0x00,  0x81,  0x10,  0x90,  0x34,  0x2E,  0x06,  0x90,  0x2D,  0x26,  0x00,  0x90,  0x3D,  0x40,  0x08,  0x90,  0x39,  0x3C,  0x03,  0xB0,  0x40,  0x2E,  0x05,  0xB0,  0x40,  0x06,  0x05,  0xB0,  0x40,  0x00,  0x4C,  0xB0,  0x40,  0x1C,  0x05,  0xB0,  0x40,  0x70,  0x05,  0xB0,  0x40,  0x7F,  0x0A,  0xB0,  0x40,  0x50,  0x05,
 0xB0,  0x40,  0x7F,  0x67,  0x80,  0x34,  0x00,  0x17,  0x90,  0x40,  0x46,  0x01,  0x80,  0x2D,  0x00,  0x09,  0x80,  0x3D,  0x00,  0x1B,  0x80,  0x39,  0x00,  0x4F,  0x90,  0x3D,  0x38,  0x02,  0x80,  0x40,  0x00,  0x81,  0x00,  0x90,  0x25,  0x24,  0x03,  0x90,  0x2C,  0x24,  0x04,  0xB0,  0x40,  0x54,  0x03,  0x90,  0x34,  0x3E,  0x02,  0xB0,  0x40,  0x1A,  0x06,  0xB0,  0x40,  0x00,  0x42,  0xB0,  0x40,  0x10,  0x05,  0xB0,  0x40,  0x6A,  0x04,
 0xB0,  0x40,  0x7F,  0x0B,  0xB0,  0x40,  0x4E,  0x06,  0xB0,  0x40,  0x7F,  0x81,  0x0E,  0x80,  0x2C,  0x00,  0x04,  0x80,  0x34,  0x00,  0x02,  0x80,  0x25,  0x00,  0x23,  0x80,  0x3D,  0x00,  0x81,  0x03,  0x90,  0x3B,  0x42,  0x0A,  0x90,  0x36,  0x34,  0x01,  0xB0,  0x40,  0x50,  0x03,  0x90,  0x2D,  0x22,  0x02,  0xB0,  0x40,  0x18,  0x01,  0x90,  0x26,  0x24,  0x05,  0x90,  0x3D,  0x3A,  0x00,  0xB0,  0x40,  0x00,  0x54,  0x80,  0x3B,  0x00,
 0x1A,  0xB0,  0x40,  0x58,  0x04,  0xB0,  0x40,  0x7F,  0x18,  0xB0,  0x40,  0x58,  0x06,  0xB0,  0x40,  0x7F,  0x4D,  0x90,  0x3B,  0x3A,  0x18,  0x80,  0x3D,  0x00,  0x04,  0x80,  0x26,  0x00,  0x03,  0x80,  0x2D,  0x00,  0x10,  0x80,  0x36,  0x00,  0x10,  0x80,  0x3B,  0x00,  0x35,  0x90,  0x3B,  0x38,  0x81,  0x05,  0x90,  0x31,  0x26,  0x02,  0xB0,  0x40,  0x5E,  0x03,  0xB0,  0x40,  0x1E,  0x00,  0x90,  0x2A,  0x1E,  0x03,  0x90,  0x36,  0x2E,
 0x02,  0xB0,  0x40,  0x00,  0x06,  0x90,  0x39,  0x36,  0x04,  0x80,  0x3B,  0x00,  0x81,  0x3A,  0xB0,  0x40,  0x0A,  0x05,  0xB0,  0x40,  0x2A,  0x05,  0xB0,  0x40,  0x78,  0x05,  0xB0,  0x40,  0x7F,  0x30,  0xB0,  0x40,  0x6C,  0x05,  0xB0,  0x40,  0x30,  0x04,  0xB0,  0x40,  0x7A,  0x05,  0xB0,  0x40,  0x7F,  0x81,  0x67,  0x80,  0x31,  0x00,  0x02,  0x80,  0x2A,  0x00,  0x1C,  0x80,  0x36,  0x00,  0x01,  0x80,  0x39,  0x00,  0x83,  0x16,  0x90,
 0x31,  0x2A,  0x07,  0x90,  0x39,  0x30,  0x03,  0x90,  0x2A,  0x20,  0x16,  0xB0,  0x40,  0x70,  0x04,  0xB0,  0x40,  0x28,  0x07,  0xB0,  0x40,  0x02,  0x03,  0xB0,  0x40,  0x00,  0x4D,  0xB0,  0x40,  0x4A,  0x05,  0xB0,  0x40,  0x7F,  0x81,  0x0A,  0x80,  0x31,  0x00,  0x06,  0x80,  0x2A,  0x00,  0x11,  0x80,  0x39,  0x00,  0x81,  0x56,  0x90,  0x34,  0x2E,  0x05,  0x90,  0x3D,  0x2F,  0x03,  0x90,  0x2D,  0x28,  0x0F,  0xB0,  0x40,  0x52,  0x05,
 0xB0,  0x40,  0x18,  0x05,  0xB0,  0x40,  0x00,  0x56,  0xB0,  0x40,  0x16,  0x05,  0xB0,  0x40,  0x6E,  0x05,  0xB0,  0x40,  0x7F,  0x81,  0x29,  0x80,  0x34,  0x00,  0x03,  0x80,  0x3D,  0x00,  0x08,  0x80,  0x2D,  0x00,  0x81,  0x52,  0x90,  0x31,  0x2A,  0x03,  0xB0,  0x40,  0x5C,  0x03,  0x90,  0x39,  0x2E,  0x02,  0xB0,  0x40,  0x1C,  0x03,  0x90,  0x2A,  0x1E,  0x03,  0xB0,  0x40,  0x00,  0x48,  0xB0,  0x40,  0x0C,  0x04,  0xB0,  0x40,  0x68,
 0x05,  0xB0,  0x40,  0x42,  0x06,  0xB0,  0x40,  0x18,  0x04,  0xB0,  0x40,  0x01,  0x04,  0xB0,  0x40,  0x00,  0x05,  0xB0,  0x40,  0x5E,  0x05,  0xB0,  0x40,  0x7F,  0x81,  0x32,  0x80,  0x31,  0x00,  0x06,  0x80,  0x2A,  0x00,  0x27,  0x80,  0x39,  0x00,  0x53,  0x90,  0x34,  0x2C,  0x02,  0x90,  0x3D,  0x38,  0x03,  0x90,  0x2D,  0x26,  0x01,  0xB0,  0x40,  0x41,  0x06,  0xB0,  0x40,  0x10,  0x03,  0xB0,  0x40,  0x00,  0x45,  0xB0,  0x40,  0x06,
 0x03,  0xB0,  0x40,  0x66,  0x05,  0xB0,  0x40,  0x7F,  0x05,  0xB0,  0x40,  0x48,  0x05,  0xB0,  0x40,  0x1A,  0x06,  0xB0,  0x40,  0x01,  0x04,  0xB0,  0x40,  0x4C,  0x04,  0xB0,  0x40,  0x7F,  0x6B,  0x80,  0x34,  0x00,  0x24,  0x80,  0x3D,  0x00,  0x11,  0x80,  0x2D,  0x00,  0x81,  0x6B,  0x90,  0x34,  0x2C,  0x02,  0x90,  0x25,  0x24,  0x03,  0xB0,  0x40,  0x54,  0x00,  0x90,  0x2C,  0x1D,  0x04,  0xB0,  0x40,  0x18,  0x05,  0xB0,  0x40,  0x00,
 0x4C,  0xB0,  0x40,  0x60,  0x05,  0xB0,  0x40,  0x7F,  0x74,  0xB0,  0x40,  0x4C,  0x4F,  0x80,  0x2C,  0x00,  0x07,  0x80,  0x34,  0x00,  0x02,  0xB0,  0x40,  0x20,  0x03,  0x80,  0x25,  0x00,  0x44,  0xB0,  0x40,  0x72,  0x58,  0xB0,  0x40,  0x7F,  0x83,  0x50,  0xB0,  0x40,  0x52,  0x4F,  0xB0,  0x40,  0x18,  0x68,  0x90,  0x36,  0x29,  0x58,  0xB0,  0x40,  0x00,  0x39,  0x90,  0x2D,  0x1A,  0x81,  0x27,  0x80,  0x36,  0x00,  0x47,  0x80,  0x2D,
 0x00,  0x83,  0x41,  0xFF,  0x2F,  0x00, };

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static OS_STACKPTR int        _StackMain[1536/sizeof(int)];
static OS_TASK                _TCBMain;
static OS_STACKPTR int        _StackIsr[1276/sizeof(int)];
static OS_TASK                _TCBIsr;
static OS_STACKPTR int        _StackMelody[1276/sizeof(int)];
static OS_TASK                _TCBMelody;
static OS_STACKPTR int        _StackLyrics[1276/sizeof(int)];
static OS_TASK                _TCBLyrics;
static USBH_MIDI_HANDLE       _aMidiDevice[128];
static U8                     _aNoteOn[128];
static volatile int           _NoteChange;
static USBH_NOTIFICATION_HOOK _Hook;
static char                   _aLyric[128];
static U8                     _aManufacturerName [64];                     // Descriptor for manufacturer name
static unsigned               _ManufacturerNameLen;
static U8                     _aProductName      [64];                     // Descriptor for product name
static unsigned               _ProductNameLen;
static unsigned               _FirstDelta = 1;
static U64                    _OriginTime;
#if USE_GUI
static OS_STACKPTR int        _StackDisplay[1276/sizeof(int)];
static OS_TASK                _TCBDisplay;
static GUI_MEMDEV_Handle      _hMemDev;
#endif
/*********************************************************************
*
*       Static Code
*
**********************************************************************
*/
#if USE_GUI
/*********************************************************************
*
*       _GetStringA()
*
*  Function description
*    Extract US-ASCII string from UTF-16 string.
*
*  Parameters
*    pOutput     - Pointer to output string.
*    pOutputLen  - [IN]  Octet capacity of output string.
*                  [OUT] Octet lngth of output string.
*    pInput      - Pointer to UTF-16 LE octet string.
*    InputLen    - Octet length of the UTF-16 LE octet string.
*    Replacement - Character that replaces any non-US-ASCII character
*                  in the input string.
*/
static void _GetStringA(char *pOutput, unsigned *pOutputLen, U8 *pInput, unsigned InputLen, unsigned Replacement) {
  unsigned OutputLen;
  unsigned Char;
  //
  // First two bytes of descriptor are type and length.
  //
  if (InputLen > 2) {
    pInput   += 2;
    InputLen -= 2;
  }
  OutputLen = 0;
  while (OutputLen < *pOutputLen && InputLen > 1) {
    Char = USBH_LoadU16LE(pInput);
    if (Char >= 0x7E) {
      Char = Replacement;
    }
    if (Char) {
      *pOutput = Char & 0xFF;
      pOutput   += 1;
      OutputLen += 1;
    }
    pInput   += 2;
    InputLen -= 2;
  }
  //
  *pOutputLen = OutputLen;
}
#endif
/*********************************************************************
*
*       _MIDI_DeviceEvent()
*
*  Function description
*    Called when a device is added or removed.
*
*  Parameters
*    pContext - Pointer to user-provided context.
*    DevIndex - Zero-based device index.
*    Event    - Reason for callback.
*
*  Additional information
*    Called in the context of the USBH task.
*/
static void _MIDI_DeviceEvent(void * pContext, U8 DevIndex, USBH_DEVICE_EVENT Event) {
  USBH_MIDI_DEVICE_INFO DevInfo;
  USBH_STATUS           Status;
  unsigned              i;
  //
  (void)pContext;
  //
  switch (Event) {
  case USBH_DEVICE_EVENT_ADD:
    if (_aMidiDevice[DevIndex] != USBH_MIDI_INVALID_HANDLE) {
      USBH_Warnf_Application("MIDI device %d already open!", DevIndex);
    } else {
      _aMidiDevice[DevIndex] = USBH_MIDI_Open(DevIndex);
      if (_aMidiDevice[DevIndex] != USBH_MIDI_INVALID_HANDLE) {
        Status = USBH_MIDI_GetDeviceInfo(_aMidiDevice[DevIndex], &DevInfo);
        if (Status == USBH_STATUS_SUCCESS) {
          USBH_Logf_Application("MIDI device %d added: %u In cables, %u Out cables", DevIndex, DevInfo.NumInCables, DevInfo.NumOutCables);
        } else {
          USBH_Warnf_Application("Error getting MIDI device %d information!", DevIndex);
        }
      } else {
        USBH_Warnf_Application("Error opening MIDI device %d!", DevIndex);
      }
    }
    break;
    //
  case USBH_DEVICE_EVENT_REMOVE:
    if (_aMidiDevice[DevIndex] == USBH_MIDI_INVALID_HANDLE) {
      USBH_Warnf_Application("MIDI device %d already removed!", DevIndex);
    } else {
      (void)USBH_MIDI_Close(_aMidiDevice[DevIndex]);
      _aMidiDevice[DevIndex] = USBH_MIDI_INVALID_HANDLE;
      USBH_Logf_Application("MIDI device %d removed", DevIndex);
    }
    break;
    //
  default:
    break;
  }
  //
  // Now get the device and manufacturer name of the "primary" device.
  //
  if (Event == USBH_DEVICE_EVENT_ADD) {
    _ManufacturerNameLen = 0;
    _ProductNameLen      = 0;
    //
    for (i = 0; i < SEGGER_COUNTOF(_aMidiDevice); ++i) {
      if (_aMidiDevice[i] != USBH_MIDI_INVALID_HANDLE) {
        USBH_INTERFACE_HANDLE hInterface;
        U8                    aDesc[18];
        unsigned              DescLen;
        U8                    iManufacturer;
        U8                    iProduct;
        //
        USBH_MIDI_GetDeviceInfo(_aMidiDevice[i], &DevInfo);
        Status = USBH_OpenInterface(DevInfo.InterfaceId, 0, &hInterface);
        if (Status == USBH_STATUS_SUCCESS) {
          Status = USBH_GetDeviceDescriptor(hInterface, aDesc, &DescLen);
        }
        if (Status == USBH_STATUS_SUCCESS) {
          iManufacturer = aDesc[14];
          iProduct      = aDesc[15];
          _ManufacturerNameLen = sizeof(_aManufacturerName);
          Status = USBH_GetStringDescriptor(hInterface, iManufacturer, 0, _aManufacturerName, &_ManufacturerNameLen);
          if (Status != USBH_STATUS_SUCCESS) {
            _ManufacturerNameLen = 0;
          }
          _ProductNameLen = sizeof(_aProductName);
          Status = USBH_GetStringDescriptor(hInterface, iProduct, 0, _aProductName, &_ProductNameLen);
          if (Status != USBH_STATUS_SUCCESS) {
            _ProductNameLen = 0;
          }
        }
        USBH_CloseInterface(hInterface);
      }
    }
    //
    _NoteChange = 1;
  }
}

#if USE_GUI
/*********************************************************************
*
*       _GetWhiteRect()
*
*  Function description
*    Get rectangle corresponding to a white key.
*
*  Parameters
*    pRect - Pointer to the object that receives the rectangle.
*    Pos   - White key position.
*/
static void _GetWhiteRect(GUI_RECT *pRect, unsigned Pos) {
  pRect->x0 = MARGIN_LEFT + Pos * WHITE_KEY_WIDTH;
  pRect->y0 = MARGIN_TOP;
  pRect->x1 = pRect->x0 + WHITE_KEY_WIDTH;  // Overlap white key rectangles by one pixel
  pRect->y1 = pRect->y0 + WHITE_KEY_LENGTH - 1;
}

/*********************************************************************
*
*       _GetBlackRect()
*
*  Function description
*    Get rectangle corresponding to a black key.
*
*  Parameters
*    pRect - Pointer to the object that receives the rectangle.
*    Pos   - Black key position.
*/
static void _GetBlackRect(GUI_RECT *pRect, unsigned Pos) {
  pRect->x0 = MARGIN_LEFT + Pos * WHITE_KEY_WIDTH + WHITE_KEY_WIDTH - BLACK_KEY_WIDTH/2;
  pRect->y0 = MARGIN_TOP;
  pRect->x1 = pRect->x0 + BLACK_KEY_WIDTH  - 1;
  pRect->y1 = pRect->y0 + BLACK_KEY_LENGTH - 1;
}

/*********************************************************************
*
*       _DrawWhiteKey()
*
*  Function description
*    Draw white key.
*
*  Parameters
*    Pos      - White key position.
*    pPalette - Pointer to key palette colors.
*    sLabel   - Optional label for key.
*/
static void _DrawWhiteKey(unsigned Pos, const KEY_PALETTE *pPalette, const char *sLabel) {
  GUI_RECT Rect;
  //
  _GetWhiteRect(&Rect, Pos);
  GUI_SetColor(pPalette->KeyColor);
  GUI_FillRectEx(&Rect);
  GUI_SetColor(pPalette->KeyBorder);
  GUI_DrawRectEx(&Rect);
  //
  if (sLabel != NULL) {
    GUI_SetFont(&GUI_Font13B_ASCII);
    GUI_SetBkColor(pPalette->KeyColor);
    GUI_SetColor(pPalette->LegendColor);
    GUI_DispStringHCenterAt(sLabel, Rect.x0 + WHITE_KEY_WIDTH/2 + 2, Rect.y1-15);
  }
}

/*********************************************************************
*
*       _DrawBlackKey()
*
*  Function description
*    Draw black key.
*
*  Parameters
*    Pos      - Black key position.
*    pPalette - Pointer to key palette colors.
*/
static void _DrawBlackKey(unsigned Pos, const KEY_PALETTE *pPalette) {
  GUI_RECT Rect;
  //
  _GetBlackRect(&Rect, Pos);
  //
  GUI_SetColor(pPalette->KeyColor);
  GUI_FillRectEx(&Rect);
  GUI_SetColor(pPalette->KeyBorder);
  GUI_DrawRectEx(&Rect);
}

/*********************************************************************
*
*       _CalcKeyBase()
*
*  Function description
*    Convert MIDI note number to virtual keyboard position.
*
*  Parameters
*    Note - MIDI note number, 0...127.
*
*  Return value
*    Left x-coordinate of key.
*/
static int _CalcKeyBase(unsigned Note) {
  int Octave;
  int Semitone;
  //
  Octave   = (Note - 9) / 12;
  Semitone = (Note + 3) % 12;
  //
  return Octave*7 + _aKeyMapping[Semitone].Index;
}

/*********************************************************************
*
*       _DrawKey()
*
*  Function description
*    Draw key corresponding to a MIDI note.
*
*  Parameters
*    Note  - MIDI note number, 0...127.
*    Color - Whote or black indication.
*/
static void _DrawKey(int Note, unsigned Color) {
  int  Semitone;
  int  XPos;
  char aNoteName[5];
  //
  MIDI_DecodeNote(Note, aNoteName);
  Semitone = (Note + 3) % 12;
  XPos     = _CalcKeyBase(Note) - _CalcKeyBase(KEYBED_LOW_NOTE);
  //
  if (_aKeyMapping[Semitone].Color == Color) {
    if (Color == WHITE) {
      if (Semitone == 3) {
        _DrawWhiteKey(XPos, _aNoteOn[Note] ? &WhiteKeyOn : &WhiteKeyOff, aNoteName);  // Legend always shown for C
      } else {
        _DrawWhiteKey(XPos, _aNoteOn[Note] ? &WhiteKeyOn : &WhiteKeyOff, _aNoteOn[Note] ? aNoteName : NULL);  // Show legend if key pressed
      }
    } else {
      _DrawBlackKey(XPos, _aNoteOn[Note] ? &BlackKeyOn : &BlackKeyOff);
    }
  }
}

/*********************************************************************
*
*       _DrawKeyboard()
*
*  Function description
*    Draw entire keyboard with note down indication.
*/
static void _DrawKeyboard(void) {
  unsigned Key;
  char     aTitle[64];
  unsigned T1Len;
  unsigned T2Len;
  //
  GUI_MEMDEV_Select(_hMemDev);
  //
  GUI_SetColor(COLOR_BACKGROUND);
  GUI_FillRect(0, 0, LCD_GetXSize()-1, LCD_GetYSize()-1);
  //
  T1Len = sizeof(aTitle);
  _GetStringA(&aTitle[0], &T1Len, _aManufacturerName, _ManufacturerNameLen, 0);
  aTitle[T1Len++] = ' ';
  _GetStringA(&aTitle[T1Len], &T2Len, _aProductName, _ProductNameLen, 0);
  aTitle[T1Len + T2Len] = 0;
  //
  GUI_SetFont(&GUI_Font20B_ASCII);
  GUI_SetBkColor(COLOR_BACKGROUND);
  GUI_SetColor(GUI_WHITE);
  GUI_DispStringHCenterAt(aTitle, LCD_GetXSize()/2, MARGIN_TOP - 10 - GUI_GetYSizeOfFont(&GUI_Font20B_ASCII));
  //
  GUI_DispStringAt(_aLyric, MARGIN_LEFT, MARGIN_TOP + WHITE_KEY_LENGTH + 20);
  //
  for (Key = KEYBED_LOW_NOTE; Key < KEYBED_HIGH_NOTE; ++Key) {
    _DrawKey(Key, WHITE);
  }
  for (Key = KEYBED_LOW_NOTE; Key < KEYBED_HIGH_NOTE; ++Key) {
    _DrawKey(Key, BLACK);
  }
  //
  GUI_MEMDEV_Select(0);
  GUI_MEMDEV_CopyToLCD(_hMemDev);
}

/*********************************************************************
*
*       _DisplayTask()
*
*  Function description
*    Virtual keyboard display task.
*
*  Additional information
*    Waits for note changed events and redisplays the virtual
*    keyboard through the GUI.
*/
static void _DisplayTask(void) {
  _DrawKeyboard();
  for (;;) {
    if (_NoteChange) {
      _NoteChange = 0;
      _DrawKeyboard();
    }
    OS_Delay(50);
  }
}
#endif

/*********************************************************************
*
*       _PlayAPI_Delta()
*
*  Function description
*    Process MIDI file Delta.
*
*  Parameters
*    pContext - Pointer to user context.
*    Delta    - Number of delta-ticks before next event.
*
*  Additional information
*    This function waits until the note should be sounded or the lyric
*    displayed using a 64-bit 1us high-precision timer that is readily
*    available in embOS.  It does so without using a busy-loop on
*    a timer.
*/
static void _PlayAPI_Delta(void *pContext, U32 Delta) {
  PLAY_CONTEXT * pSelf;
  U64            AbsMicroseconds;
  I64            DeltaMicroseconds;
  //
  pSelf = (PLAY_CONTEXT *)pContext;
  //
  if (_FirstDelta) {
    _OriginTime = OS_GetTime_us64();
    _FirstDelta = 0;
  }
  if (pSelf->CurTrackId == pSelf->PlayTrackId) {
    pSelf->PlayHead += Delta;
    AbsMicroseconds  = pSelf->PlayHead;
    AbsMicroseconds *= pSelf->Tempo;      // Tempo is microseconds per quarter note
    AbsMicroseconds /= pSelf->Division;   // Division is delta-ticks per quarter note
    AbsMicroseconds += _OriginTime;       // Number of microseconds since play commenced
    //
    // Compute microseconds until the event fires.
    //
    DeltaMicroseconds = (I64)(AbsMicroseconds - OS_GetTime_us64());
    //
    // Coarse millisecond-based delay.  We delay until there is approximately
    // 3ms until the event expires, leaving us some margin because we don't
    // exactly know when the system tick will happen: could be immediately,
    // or could be in 999us.
    //
    if (DeltaMicroseconds > 3000) {
      OS_Delay((unsigned)(DeltaMicroseconds / 1000));
    }
    //
    // The event fires within 3ms, or 3,000 microseconds which
    // we can pass to embOS to sort out for us.  It does this
    // precisely.
    //
    DeltaMicroseconds = (I64)(AbsMicroseconds - OS_GetTime_us64());
    if (DeltaMicroseconds > 0) {
      OS_Delayus((unsigned)DeltaMicroseconds);
    }
  }
}

/*********************************************************************
*
*       _PlayAPI_MidiEvent()
*
*  Function description
*    Send MIDI event to all attached devices.
*
*  Parameters
*    pContext - Pointer to user context.
*    pData    - Pointer to MIDI message octet string.
*    DataLen  - Octet length of the MIDI message octet string.
*
*  Additional information
*    This function also updates the state of the virtual MIDI
*    keyboard and signals to the display task that a redraw
*    is necessary.
*/
static void _PlayAPI_MidiEvent(void *pContext, const U8 *pData, unsigned DataLen) {
  PLAY_CONTEXT * pSelf;
  U32            Event;
  unsigned       i;
  //
  pSelf = (PLAY_CONTEXT *)pContext;
  //
  if (pSelf->CurTrackId == pSelf->PlayTrackId) {
    Event = MIDI_ConstructEvent(pData, DataLen);
    Event |= ((Event & 0x00F00000) << 4);  // Copy status bits to CID
    //
    // Update keyboard with event
    //
    if ((MIDI_STATUS(Event) & 0xF0) == 0x90) {  // Note On
      _aNoteOn[MIDI_DATA2(Event)] = MIDI_DATA3(Event) != 0;  // Velocity 0 is note off
      _NoteChange = 1;
    } else if ((MIDI_STATUS(Event) & 0xF0) == 0x80) {  // Note Off
      _aNoteOn[MIDI_DATA2(Event)] = 0;
      _NoteChange = 1;
    }
    //
    for (i = 0; i < SEGGER_COUNTOF(_aMidiDevice); ++i) {
      if (_aMidiDevice[i] != USBH_MIDI_INVALID_HANDLE) {
        USBH_MIDI_WrEvent(_aMidiDevice[i], Event);
      }
    }
  }
}

/*********************************************************************
*
*       _PlayAPI_Header()
*
*  Function description
*    Process Header event.
*
*  Parameters
*    pContext - Pointer to user context.
*    Format   - Format of MIDI file.
*    TrackCnt - Number of tracks in MIDI file.
*    Division - Number of delta-ticks per quarter note.
*/
static void _PlayAPI_Header(void *pContext, unsigned Format, unsigned TrackCnt, unsigned Division) {
  PLAY_CONTEXT * pSelf;
  //
  (void)Format;
  (void)TrackCnt;
  //
  pSelf = (PLAY_CONTEXT *)pContext;
  pSelf->Division = Division;
}

/*********************************************************************
*
*       _PlayAPI_BeginTrack()
*
*  Function description
*    Process Begin Track event.
*
*  Parameters
*    pContext - Pointer to user context.
*    TrackId  - Zero-based track index.
*/
static void _PlayAPI_BeginTrack(void *pContext, unsigned TrackId) {
  PLAY_CONTEXT * pSelf;
  //
  pSelf = (PLAY_CONTEXT *)pContext;
  pSelf->CurTrackId = TrackId;
}

/*********************************************************************
*
*       _PlayAPI_SetTempo()
*
*  Function description
*    Process Set Tempo event.
*
*  Parameters
*    pContext - Pointer to user context.
*    Tempo    - New tempo, microseconds per quarter note.
*/
static void _PlayAPI_SetTempo(void *pContext, U32 Tempo) {
  PLAY_CONTEXT * pSelf;
  //
  pSelf = (PLAY_CONTEXT *)pContext;
  pSelf->Tempo = Tempo;
}

/*********************************************************************
*
*       _PlayAPI_Lyric()
*
*  Function description
*    Process Lyric event.
*
*  Parameters
*    pContext - Pointer to user context.
*    pText    - Pointer to lyric octet string.
*    TextLen  - Octet length of the lyric octet string.
*/
static void _PlayAPI_Lyric(void *pContext, const U8 *pText, unsigned TextLen) {
  PLAY_CONTEXT * pSelf;
  unsigned       i;
  //
  pSelf = (PLAY_CONTEXT *)pContext;
  //
  if (pSelf->CurTrackId == pSelf->PlayTrackId) {
    for (i = 0; i < TextLen; ++i) {
      if (pText[i] == '\r' || pText[i] == '\n') {
        pSelf->LyricPos = 0;
      } else if (pText[i] == '_') {
        pSelf->aLyric[pSelf->LyricPos++] = ' ';
        pSelf->aLyric[pSelf->LyricPos] = 0;
      } else {
        pSelf->aLyric[pSelf->LyricPos++] = pText[i];
        pSelf->aLyric[pSelf->LyricPos] = 0;
      }
    }
    OS_IncDI();
    memcpy(_aLyric, pSelf->aLyric, 128);
    OS_DecRI();
    _NoteChange = 1;
  }
}

/*********************************************************************
*
*       _WaitingForMIDIDevice()
*
*  Function description
*    Query if MIDI device is plugged in.
*
*  Return value
*    == 0 - No longer waiting, device ready.
*    != 0 - Still waiting.
*/
static int _WaitingForMIDIDevice(void) {
  unsigned i;
  //
  for (i = 0; i < SEGGER_COUNTOF(_aMidiDevice); ++i) {
    if (_aMidiDevice[i] != USBH_MIDI_INVALID_HANDLE) {
      return 0;
    }
  }
  return 1;
}

/*********************************************************************
*
*       _MelodyTask()
*
*  Function description
*    Play the melody track and exit.
*/
static void _MelodyTask(void) {
  PLAY_CONTEXT Context;
  MIDI_FILE    F;
  //
  Context.PlayHead    = 0;
  Context.CurTrackId  = 0;
  Context.Division    = 0;
  Context.PlayTrackId = MELODY_TRACK;
  Context.Tempo       = 0;
  Context.aLyric[0]   = 0;
  Context.LyricPos    = 0;
  //
  while (_WaitingForMIDIDevice()) {
    OS_Delay(100);
  }
  //
  MIDI_PrepareFile     (&F, _aMidiFile, sizeof(_aMidiFile));
  MIDI_ProcessAllTracks(&F, &_PlayAPI, &Context);
  //
  OS_Terminate(0);
}

/*********************************************************************
*
*       _LyricsTask()
*
*  Function description
*    Display the lyrics in time with the music and exit.
*/
static void _LyricsTask(void) {
  PLAY_CONTEXT Context;
  MIDI_FILE    F;
  //
  Context.PlayHead    = 0;
  Context.CurTrackId  = 0;
  Context.Division    = 0;
  Context.PlayTrackId = LYRIC_TRACK;
  Context.Tempo       = 0;
  Context.aLyric[0]   = 0;
  Context.LyricPos    = 0;
  //
  strcpy(_aLyric, "SEGGER emUSB-Host MIDI and emLib-MIDI demo");
  _NoteChange = 1;
  //
  while (_WaitingForMIDIDevice()) {
    OS_Delay(100);
  }
  //
  MIDI_PrepareFile     (&F, _aMidiFile, sizeof(_aMidiFile));
  MIDI_ProcessAllTracks(&F, &_LyricAPI, &Context);
  //
  OS_Terminate(0);
}

/*********************************************************************
*
*       Public code
*
**********************************************************************
*/

#ifdef __cplusplus
extern "C" {     /* Make sure we have C-declarations in C++ programs */
#endif
void MainTask(void);
#ifdef __cplusplus
}
#endif

/*********************************************************************
*
*       MainTask()
*
*  Function description
*    Start all tasks.
*/
void MainTask(void) {
  //
  USBH_Init();
  USBH_MIDI_Init();
  USBH_MIDI_AddNotification(&_Hook, _MIDI_DeviceEvent, NULL);
#if USE_GUI
  //                                                                                            // Tasks using emUSB-Host API should always have a lower priority than emUSB-Host main and ISR tasks.
  _hMemDev = GUI_MEMDEV_Create(0, 0, LCD_GetXSize(), LCD_GetYSize());
  if (_hMemDev == 0) {
    GUI_X_Warn("Cannot create memdev!");
    return;
  }
  OS_CREATETASK(&_TCBDisplay, "Display",   _DisplayTask, TASK_PRIO_KB_DISP,   _StackDisplay);   // Start task managing keyboard display  }
#endif
  //
  OS_SetPriority(OS_GetTaskID(), TASK_PRIO_APP);                                                // This task has the lowest prio for real-time application.
  OS_CREATETASK(&_TCBMain,    "USBH_Task", USBH_Task,    TASK_PRIO_USBH_MAIN, _StackMain);      // Start USBH main task
  OS_CREATETASK(&_TCBIsr ,    "USBH_ISR",  USBH_ISRTask, TASK_PRIO_USBH_ISR,  _StackIsr);       // Start USBH ISR task
  OS_CREATETASK(&_TCBMelody,  "Melody",    _MelodyTask,  TASK_PRIO_KB_PLAY,   _StackMelody);    // Start task managing playing melody on channel #1
  OS_CREATETASK(&_TCBLyrics,  "Lyrics",    _LyricsTask,  TASK_PRIO_KB_PLAY,   _StackLyrics);    // Start task managing displaying lyrics
  //
  for (;;) {
    OS_Delay(250);
    BSP_ToggleLED(1);
  }
}

/*************************** End of file ****************************/
